import bpy
import json
from bpy.types import Scene, Operator, PropertyGroup
from bpy.props import (CollectionProperty,
                       BoolProperty,
                       FloatVectorProperty,
                       StringProperty,
                       FloatProperty)

from ..utils.utils import (
    refresh, get_fcurve_selected_keys, get_selected_keys,
    get_fcurve_nice_name, get_handle_length, sub_v2_v2v2,
    normalize_v2_length, add_v2_v2v2,
    get_prefs
)

from mathutils import Vector,Euler
from dataclasses import dataclass, field, asdict
from typing import List, Optional, Any

@dataclass
class Keyframe:
    time: float = 0.0
    value: float = 0.0
    handleLeft: List[float] = field(default_factory=list)
    handleRight: List[float] = field(default_factory=list)
    handleLeftType: str = "ALIGNED"
    handleRightType: str = "ALIGNED"
    interpolation: str = "LINEAR"

@dataclass

class FCurve:
    name: str = "FCurve"
    data_path: str = ""
    extrapolation: str = "LINEAR"
    keyframes: List[Keyframe] = field(default_factory=list)

@dataclass
class Composition:
    name: str = "Blender Scene"
    type: str = "After Effects Composition"
    duration: int = 0
    fps: int = 12
    width: int = 1920
    height: int = 1080
    frame_start: int = 1
    frame_end: int = 72
    aspect: float = 1.777777777777778
    size: float = 100.0
    bgColor: List[float] = field(default_factory=list)
    fcurves: List[FCurve] = field(default_factory=list)

debug_mode = False

def get_left_length(keyframe):
    return get_handle_length(keyframe.co[0], keyframe.handle_left[0],
        keyframe.co[1], keyframe.handle_left[1])

def get_right_length(keyframe):
    return get_handle_length(keyframe.co[0], keyframe.handle_right[0],
        keyframe.co[1], keyframe.handle_right[1])

def set_left_length(sk,preserve_types,handle_length):
    if not preserve_types:
        handle_types = ["AUTO_CLAMPED","ALIGNED","AUTO","VECTOR"]
        if sk.handle_left_type in handle_types:
            sk.handle_left_type = "FREE"

    handle_direction = sub_v2_v2v2(sk.handle_left, sk.co)
    normalize_v2_length(handle_direction, handle_length)
    handle_direction = add_v2_v2v2(sk.handle_left,sk.co,handle_direction)

def set_right_length(sk,preserve_types,handle_length):
    if not preserve_types:
        handle_types = ["AUTO_CLAMPED","ALIGNED","AUTO","VECTOR"]
        if sk.handle_right_type in handle_types:
            sk.handle_right_type = "FREE"

    handle_direction = sub_v2_v2v2(sk.handle_right, sk.co)
    normalize_v2_length(handle_direction, handle_length)
    add_v2_v2v2(sk.handle_right,sk.co,handle_direction)

def set_key_ease(p, keyframe, preserve_types=True, paste_angle=True, invert=None, relative_time=1, relative_val=1):
    p.interpolation = keyframe.interpolation
    p.handle_left_type = keyframe.handle_left_type
    p.handle_right_type = keyframe.handle_right_type

    if paste_angle == True:
        p.handle_left[0] = p.co[0] + ((keyframe["handle_left"][0] - keyframe["co"][0])*relative_time)
        p.handle_right[0] = p.co[0] + ((keyframe["handle_right"][0] - keyframe["co"][0])*relative_time)

        if invert == False:
            p.handle_left[1] = p.co[1] - ((keyframe["handle_left"][1] - keyframe["co"][1])*relative_val)
            p.handle_right[1] = p.co[1] - ((keyframe["handle_right"][1] - keyframe["co"][1])*relative_val)
        else:
            p.handle_left[1] = p.co[1] + ((keyframe["handle_left"][1] - keyframe["co"][1])*relative_val)
            p.handle_right[1] = p.co[1] + ((keyframe["handle_right"][1] - keyframe["co"][1])*relative_val)

    else:
        set_left_length(p,preserve_types,keyframe.handle_left_length)
        set_right_length(p,preserve_types,keyframe.handle_right_length)

def set_key_value(p, keyframe):
    p.co[1] = keyframe.co[1]

def reset_keyframe(keyframe,start_frame):
    keyframe.amplitude = 0.0
    keyframe.back = 0.0
    keyframe.easing = 'AUTO'
    keyframe.handle_left_type = 'FREE'
    keyframe.handle_right_type = 'FREE'
    keyframe.co[0] = start_frame
    keyframe.handle_left[0] = start_frame-1.0
    keyframe.handle_right[0] = start_frame+1.0
    keyframe.handle_left[1] = keyframe.co[1]
    keyframe.handle_right[1] = keyframe.co[1]
    keyframe.period = 0.0
    keyframe.type = 'KEYFRAME'
    return start_frame

def delete_keyframe_by_frame_number(keyframes,frame_num):
    for k in reversed(keyframes[:]):
        if k.co[0] == frame_num:
            keyframes.remove(k, fast=True)

def check_clipboard(extra=""):
    # Calculate the range of the F-Curve values
    fcurve_min = min(k.co[1] for k in bpy.context.scene.fcurve_clipboard)
    fcurve_max = max(k.co[1] for k in bpy.context.scene.fcurve_clipboard)
    fcurve_range = fcurve_max - fcurve_min

    frames = []
    for x in bpy.context.scene.fcurve_clipboard:
        frames.append(f"handle_left: {round(x.handle_left[0],2)} "\
            f"frame: {x.co[0]} handle_right: {round(x.handle_right[0],2)}")
    print("Clipboard:")
    print(f"Clipboard Value Range: {fcurve_range}")
    for f in frames:
        print(f"{f}",extra)


def compare_lists(list1, list2):
    result = []
    tgt_lst = list1
    if len(list2) < len(list1):
        tgt_lst = list2
    for i in range(len(tgt_lst)):
        if i == 0:
            try:
                sign1 = (list1[i+1] - list1[i]) > 0
            except IndexError:
                sign1 = None
            try:
                sign2 = (list2[i+1] - list2[i]) > 0
            except IndexError:
                sign2 = None
        else:
            try:
                sign1 = (list1[i] - list1[i-1]) > 0
            except IndexError:
                sign1 = None

            try:
                sign2 = (list2[i] - list2[i-1]) > 0
            except IndexError:
                sign2 = None

        result.append(sign1 == sign2)
    return result

def format_fcurve_clipboard_to_js(context,fcurve):
    """
    Formats the fcurve_clipboard data into a JSON-like string that can be pasted in After Effects.
    """
    scene = context.scene
    clipboard_data = scene.fcurve_clipboard

    scene = context.scene
    duration = scene.frame_end+1-scene.frame_start
    # Get the file path of the current Blender file
    file_path = bpy.data.filepath
    # Extract the file name from the file path
    file_name = bpy.path.display_name_from_filepath(file_path)

    try:
        value = bpy.data.worlds["World"].node_tree.nodes["Background"].inputs[0].default_value
        background_color = [value[0],value[1],value[2]]
    except:
        background_color = [1.0,1.0,1.0]

    comp = Composition(name = file_name,
                        type = "F-Curve",
                        duration = duration,
                        fps = scene.render.fps,
                        width = scene.render.resolution_x,
                        height = scene.render.resolution_y,
                        frame_start= scene.frame_start,
                        frame_end= scene.frame_end,
                        bgColor=background_color)

    fcurve_name = get_fcurve_nice_name(fcurve)
    # Prepare a list to store formatted keyframe data
    fcurve = FCurve(name=fcurve_name,data_path=fcurve.data_path,extrapolation=fcurve.extrapolation,keyframes=[])

    # Loop through the fcurve_clipboard and convert to JavaScript-like format
    for keyframe in clipboard_data:
        # Extract keyframe data
        time = round(keyframe.co[0], 2)
        value = round(keyframe.co[1], 4)
        handle_left = [round(keyframe.handle_left[0], 2), round(keyframe.handle_left[1], 4)]
        handle_right = [round(keyframe.handle_right[0], 2), round(keyframe.handle_right[1], 4)]
        
        # Create Keyframe object
        kf = Keyframe(
            time=time,
            value=value,
            handleLeft=handle_left,
            handleRight=handle_right,
            handleLeftType=keyframe.handle_left_type,
            handleRightType=keyframe.handle_right_type,
            interpolation=keyframe.interpolation
        )
        
        # Append to the FCurve's keyframes list
        fcurve.keyframes.append(kf)
    
    # Append the FCurve to the Composition's fcurves list
    comp.fcurves.append(fcurve)

    # Convert Composition to JSON-like string
    comp_json_str = json.dumps(asdict(comp), indent=4)

    # Copy to OS clipboard
    bpy.context.window_manager.clipboard = comp_json_str

    # Debug output to console
    print("Copied keyframe data to clipboard:")
    print(comp_json_str)

class ANIMCOPY_OT_copy_fcurves(Operator):
    """Copies the values, timing, and ease of the active F-Curve's keyframes"""
    bl_idname = "graph.copy_fcurves"
    bl_label = "Copy F-Curves"

    use_selected_keys: BoolProperty(
        name="Use Selected Keyframes",
        description="Only copy data from selected keyframes, otherwise copy all keyframes",
        default=True
    )

    def execute(self, context):
        scene = context.scene
        scene.fcurve_clipboard.clear()

        prefs = get_prefs()
        global debug_mode
        if prefs:
            debug_mode = prefs.debug_mode
        else:
            debug_mode = True            

        if debug_mode:
            print("scene.fcurve_clipboard.clear():",len(scene.fcurve_clipboard))

        fcurve,selected = get_fcurve_selected_keys()

        if self.use_selected_keys and len(selected) > 0:
            keyframes_to_copy = selected
        else:
            keyframes_to_copy = fcurve.keyframe_points

        for keyframe in keyframes_to_copy:
            clip_key = context.scene.fcurve_clipboard.add()
            clip_key.interpolation = keyframe.interpolation
            clip_key.co[0] = keyframe.co[0]
            clip_key.co[1] = keyframe.co[1]
            clip_key.handle_left_type = keyframe.handle_left_type
            clip_key.handle_right_type = keyframe.handle_right_type
            clip_key.handle_left[0] = keyframe.handle_left[0]
            clip_key.handle_left[1] = keyframe.handle_left[1]
            clip_key.handle_right[0] = keyframe.handle_right[0]
            clip_key.handle_right[1] = keyframe.handle_right[1]
            clip_key.handle_left_length = get_left_length(keyframe)
            clip_key.handle_right_length = get_right_length(keyframe)
            clip_key.extrapolation = fcurve.extrapolation  # Copy the extrapolation type

        if debug_mode:
            for x in scene.fcurve_clipboard:
                print("*"*20, "Keyframe:", x.co[0], x.co[1], "*"*20)
                print("Handle Left Type:", x.handle_left_type)
                print("Handle Right Type:", x.handle_right_type)
                print("Handle Left Frame:", x.handle_left[0])
                print("Handle Left Value:", x.handle_left[1])
                print("Handle Right Frame:", x.handle_right[0])
                print("Handle Right Value:", x.handle_right[1])
                print("Extrapolation:", x.extrapolation)  # Debug extrapolation
            print("len(scene.fcurve_clipboard)", len(scene.fcurve_clipboard))

        fcurve_name = get_fcurve_nice_name(fcurve)

        format_fcurve_clipboard_to_js(context,fcurve)

        status_text = "%s keyframes copied from %s." % (len(scene.fcurve_clipboard),
                                                        fcurve_name)

        context.workspace.status_text_set(status_text)

        return {'FINISHED'}

class ANIMCOPY_OT_Paste_Ease(Operator):
    """Pastes the Ease from the copied F-Curve keyframes to the selected F-Curves"""
    bl_idname = "graph.paste_ease"
    bl_label = "Paste Ease"
    bl_options = { "REGISTER", "UNDO" }

    absolute_handles: BoolProperty(
        name= "Absolute Handles",
        description= "Paste ease with exact handle values as the copied F-Curve",
        default = False
    )

    preserve_handle_types: BoolProperty(
        name= "Preserve Handle Types",
        description= "Preserve the handle types of the copied keyframes",
        default = True
    )

    handle_angles: BoolProperty(
        name= "Handle Angles",
        description= "Paste the handle angles of the copied keyframes",
        default = True
    )

    invert_handles: BoolProperty(
        name= "Invert Handles",
        description= "Paste the inverse handle angles when the F-Curve's values are opposite of the source F-Curve",
        default = True
    )

    paste_extrapolation: BoolProperty(
        name="Paste Extrapolation",
        description="Paste the extrapolation type of the copied F-Curve",
        default=True
    )

    def execute(self, context):
        scene = context.scene
        fcurves = context.selected_editable_fcurves

        if len(scene.fcurve_clipboard) == 0:
            return {'FINISHED'}

        global debug_mode

        if debug_mode:
            print("*"*30)

        for fcurve in fcurves:
            clip_len = len(scene.fcurve_clipboard)
            selected_keys = get_selected_keys(fcurve)
            keys_len = len(selected_keys)

            clip_fcurve_time_min = min(k.co[0] for k in scene.fcurve_clipboard)
            clip_fcurve_time_max = max(k.co[0] for k in scene.fcurve_clipboard)

            clip_fcurve_value_min = min(k.co[1] for k in scene.fcurve_clipboard)
            clip_fcurve_value_max = max(k.co[1] for k in scene.fcurve_clipboard)

            clipboard_time_range = clip_fcurve_time_max - clip_fcurve_time_min
            clipboard_value_range = clip_fcurve_value_max - clip_fcurve_value_min

            relative_time_difference = 1
            relative_value_difference = 1

            if keys_len and not self.absolute_handles:
                fcurve_time_min = min(k.co[0] for k in selected_keys)
                fcurve_time_max = max(k.co[0] for k in selected_keys)
                selected_time_range = fcurve_time_max - fcurve_time_min

                fcurve_val_min = min(k.co[1] for k in selected_keys)
                fcurve_val_max = max(k.co[1] for k in selected_keys)
                selected_value_range = fcurve_val_max - fcurve_val_min
                if clipboard_time_range:
                    relative_time_difference = selected_time_range/clipboard_time_range
                if clipboard_value_range:
                    relative_value_difference = selected_value_range/clipboard_value_range

            if debug_mode:
                check_clipboard()
                print("Clipboard Length",clip_len,"keys_len",keys_len)
                print(f"Time Percentage: {relative_time_difference}")
                print(f"Value Percentage: {relative_value_difference}")
                pass
            
            invert_list = []
            if self.invert_handles:
                list1 = [k.co[1] for k in scene.fcurve_clipboard]
                list2 = [p.co[1] for p in selected_keys]
                invert_list = compare_lists(list1, list2)

            if clip_len > keys_len:
                #if more copied keys than F-Curve keys
                if debug_mode: print("Case 1 - clip_len: %s > keys_len: %s"%(clip_len,keys_len))
                for i in range(keys_len):
                    keyframe = scene.fcurve_clipboard[i]
                    p = selected_keys[i]

                    if self.invert_handles:
                        set_key_ease(p, keyframe, self.preserve_handle_types, self.handle_angles,
                                     invert=invert_list[i], relative_time=relative_time_difference,
                                     relative_val=relative_value_difference)
                    else:
                        set_key_ease(p, keyframe, self.preserve_handle_types, self.handle_angles,
                                     relative_time=relative_time_difference, relative_val=relative_value_difference)                        

            elif (keys_len > clip_len and clip_len > 1) or keys_len==clip_len:
                #if more F-Curve keys than copied keys
                if debug_mode: print("Case 2 - keys_len: %s > clip_len: %s"%(clip_len,keys_len))
                for i,keyframe in enumerate(scene.fcurve_clipboard):
                    p = selected_keys[i]

                    if self.invert_handles:
                        set_key_ease(p, keyframe, self.preserve_handle_types, self.handle_angles, invert=invert_list[i],
                                     relative_time=relative_time_difference, relative_val=relative_value_difference)
                    else:
                        set_key_ease(p, keyframe, self.preserve_handle_types, self.handle_angles,
                                     relative_time=relative_time_difference, relative_val=relative_value_difference)
            elif clip_len == 1:
                if debug_mode: print("Case 3 - clip_len: %s == 1"%(clip_len))
                keyframe = scene.fcurve_clipboard[0]
                for kf in selected_keys:
                    if kf != keyframe:
                        if self.invert_handles:
                            set_key_ease(kf, keyframe, self.preserve_handle_types, self.handle_angles,
                                         invert=invert_list[i], relative_time=relative_time_difference,
                                         relative_val=relative_value_difference)
                        else:
                            set_key_ease(kf, keyframe, self.preserve_handle_types, self.handle_angles,
                                         relative_time=relative_time_difference, relative_val=relative_value_difference)

            current_keys = [keyframe.co[0] for keyframe in fcurve.keyframe_points[:]]
            if debug_mode: print("current_keys",current_keys)

            # Set the extrapolation type
            if self.paste_extrapolation:
                fcurve.extrapolation = scene.fcurve_clipboard[0].extrapolation

        refresh()
        return {'FINISHED'}

class ANIMCOPY_OT_Paste_Timing(Operator):
    bl_idname = "graph.paste_timing"
    bl_label = "Paste Timing"
    bl_options = { "REGISTER", "UNDO" }

    replace: BoolProperty(
        name= "Replace",
        description= "Replace existing keyframes",
        default = False
    )

    preserve_handle_types: BoolProperty(
        name= "Preserve Handle Types",
        description= "Preserve the handle types of the copied keyframes",
        default = True
    )

    handle_lengths: BoolProperty(
        name= "Handle Lengths",
        description= "Paste the handle lengths of the selected keyframes",
        default = True
    )

    handle_angles: BoolProperty(
        name= "Handle Angles",
        description= "Paste the handle angles of the selected keyframes",
        default = False
    )
    
    paste_values: BoolProperty(
        name= "Paste Values",
        description= "Paste the values of the selected keyframes",
        default = False
    )

    paste_extrapolation: BoolProperty(
        name="Paste Extrapolation",
        description="Paste the extrapolation type of the copied F-Curve",
        default=True
    )

    @classmethod
    def description(cls,context,properties) -> str:
        if properties.handle_angles and properties.handle_lengths and properties.paste_values:
           return "Pastes all keyframe data from the stored F-Curve keyframes to the selected F-Curves"
        elif properties.handle_angles and properties.handle_lengths:
           return "Pastes the timing and ease from the stored F-Curve keyframes to the selected F-Curves"
        elif properties.paste_values:
           return "Pastes the timing and values from the stored F-Curve keyframes to the selected F-Curves"
        return f"Pastes keyframe data to the selected F-Curves"

    def execute(self, context):
        scene = context.scene
        fcurves = context.selected_editable_fcurves

        if len(scene.fcurve_clipboard) == 0:
            return {'FINISHED'}

        #frames = [ x.co[0] for x in bpy.context.scene.fcurve_clipboard]
        for fcurve in fcurves:
            reset_frame = None
            if self.replace:
                while len(fcurve.keyframe_points) > 1:
                    fcurve.keyframe_points.remove(fcurve.keyframe_points[-1], fast=True)
                reset_frame = reset_keyframe(fcurve.keyframe_points[0],-100)
            for keyframe in scene.fcurve_clipboard:
                if self.paste_values:
                    current_value = keyframe.co[1]
                else:
                    current_value = fcurve.evaluate(keyframe.co[0])
                new_key = fcurve.keyframe_points.insert(keyframe.co[0], value=current_value, options={"FAST"})
                if self.handle_lengths:
                    set_key_ease(new_key, keyframe, self.preserve_handle_types, self.handle_angles, self.paste_values)
            if self.replace and reset_frame:
                delete_keyframe_by_frame_number(fcurve.keyframe_points,reset_frame)

            # Set the extrapolation type
            if self.paste_extrapolation:
                fcurve.extrapolation = scene.fcurve_clipboard[0].extrapolation

        refresh()
        return {'FINISHED'}

preview_collections = {}

class Clipboard_Keyframe(PropertyGroup):
    co: FloatVectorProperty(name="Keyframe Coordinates")
    interpolation: StringProperty(name="Interpolation", default='CONSTANT')
    handle_left: FloatVectorProperty(name="Handle Left Coordinates")
    handle_left_type: StringProperty(name="Handle Left Type", default='FREE')
    handle_right: FloatVectorProperty(name="Handle Right Coordinates")
    handle_right_type: StringProperty(name="Handle Right Type", default='FREE')
    handle_left_length: FloatProperty(name="Handle Left Length")
    handle_right_length: FloatProperty(name="Handle Right Length")
    extrapolation: StringProperty(name="Extrapolation", default='CONSTANT')  # Add extrapolation property

classes = (Clipboard_Keyframe,
           ANIMCOPY_OT_copy_fcurves,
           ANIMCOPY_OT_Paste_Timing,
           ANIMCOPY_OT_Paste_Ease)

def register():
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)
    Scene.fcurve_clipboard = CollectionProperty(type=Clipboard_Keyframe)

def unregister():
    from bpy.utils import unregister_class
    for cls in reversed(classes):
        unregister_class(cls)

    del Scene.fcurve_clipboard
    
if __name__ == "__main__":
    register()
