import bpy
from bpy.types import Operator
from mathutils import Vector
from bpy.props import (BoolProperty,EnumProperty,PointerProperty)
from ..utils.utils import (
    refresh,
    get_fcurve_nice_name,
    fcurves_active_object,
    has_animation,
    get_selected_no_active_obj,
    get_selected_keys,
    get_prefs,
    get_new_fcurve,
    copy_fcurve_modifiers
)
from ..icons.custom_icons import get_icon

debug_mode = False

def get_fcurve_array_index(fcurve, invert_quaternion_w, remap_x, remap_y, remap_z):
    fc_array_index = fcurve.array_index

    if "rotation_quaternion" in fcurve.data_path:
        invert = 1
        if fcurve.array_index == 0 and invert_quaternion_w:
            invert = -1
        if fcurve.array_index == 1 and remap_x.startswith("-"):
            invert = -1
            remap_x = remap_x.replace("-","")
        if fcurve.array_index == 2 and remap_y.startswith("-"):
            invert = -1
            remap_y = remap_y.replace("-","")
        if fcurve.array_index == 3 and remap_z.startswith("-"):
            invert = -1
            remap_z = remap_z.replace("-","")

        dimensions = ["W","X","Y","Z"]
        if fcurve.array_index == 0:
            fc_array_index = 0
        elif fcurve.array_index == 1:
            fc_array_index = dimensions.index(remap_x)
        elif fcurve.array_index == 2:
            fc_array_index = dimensions.index(remap_y)
        elif fcurve.array_index == 3:
            fc_array_index = dimensions.index(remap_z)
    else:
        invert = 1
        if fcurve.array_index == 0 and remap_x.startswith("-"):
            invert = -1
            remap_x = remap_x.replace("-","")
        if fcurve.array_index == 1 and remap_y.startswith("-"):
            invert = -1
            remap_y = remap_y.replace("-","")
        if fcurve.array_index == 2 and remap_z.startswith("-"):
            invert = -1
            remap_z = remap_z.replace("-","")
        dimensions = ["X","Y","Z"]
        if fcurve.array_index == 0:
            fc_array_index = dimensions.index(remap_x)
        elif fcurve.array_index == 1:
            fc_array_index = dimensions.index(remap_y)
        elif fcurve.array_index == 2:
            fc_array_index = dimensions.index(remap_z)

    return fc_array_index,invert

def copy_to_selected(self, context, anim_data, copy_modifiers, selection_mode,
                     replace, invert_quaternion_w, remap_x, remap_y, remap_z, paste_extrapolation, data=False):
    # Debugging: Print the animation data if debug mode is enabled
    if debug_mode:
        print(f"anim_data: {anim_data}")
        
    # Get the selected objects, excluding the active object
    selected = get_selected_no_active_obj(context)

    # Check if the active object is in POSE mode
    if context.object.mode == "POSE":
        obj = context.object
        action = obj.animation_data.action

        # Get the active bone and selected bones
        active_bone = context.active_pose_bone
        selected_bones = context.selected_pose_bones

        if not active_bone or len(selected_bones) < 2:
            self.report({"WARNING"}, "No active bone or not enough selected bones.")
            return

        active_bone_name = active_bone.name

        # Iterate over each F-Curve in the action
        for fcurve in action.fcurves:
            # Check if the F-Curve is related to the active bone
            if f'pose.bones["{active_bone_name}"]' in fcurve.data_path:
                # Calculate the proper index and invert value based on the F-Curve
                fc_array_index, invert = get_fcurve_array_index(fcurve, invert_quaternion_w, remap_x, remap_y, remap_z)

                # Iterate over the selected bones (excluding the active bone)
                for bone in selected_bones:
                    if bone.name != active_bone_name:
                        # Replace the active bone's name with the selected bone's name
                        new_data_path = fcurve.data_path.replace(f'pose.bones["{active_bone_name}"]', f'pose.bones["{bone.name}"]')

                        # Try to find an existing F-Curve with the new data path and index
                        new_fcurve = action.fcurves.find(new_data_path, index=fcurve.array_index)
                        
                        # If the F-Curve exists and the replace argument is True, remove the existing F-Curve
                        if new_fcurve and replace:
                            action.fcurves.remove(new_fcurve)
                            new_fcurve = None

                        # Create a new F-Curve if no existing F-Curve is found
                        if not new_fcurve:
                            new_fcurve = action.fcurves.new(data_path=new_data_path, index=fcurve.array_index, action_group=bone.name)

                        # Copy keyframe points from the original F-Curve to the new F-Curve
                        for kf in fcurve.keyframe_points:
                            if selection_mode == "SELECTED_KEYS" and not kf.select_control_point:
                                continue
                            new_key = new_fcurve.keyframe_points.insert(kf.co[0], kf.co[1] * invert, options={"FAST"})
                            new_key.interpolation = kf.interpolation
                            new_key.amplitude = kf.amplitude
                            new_key.back = kf.back
                            new_key.easing = kf.easing
                            new_key.period = kf.period
                            new_key.type = kf.type
                            new_key.handle_left_type = kf.handle_left_type
                            new_key.handle_left = Vector((kf.handle_left[0], kf.handle_left[1] * invert))
                            new_key.handle_right_type = kf.handle_right_type
                            new_key.handle_right = Vector((kf.handle_right[0], kf.handle_right[1] * invert))
                            new_key.select_control_point = kf.select_control_point
                            new_key.select_left_handle = kf.select_left_handle
                            new_key.select_right_handle = kf.select_right_handle

                # Copy F-curve modifiers if the option is enabled
                if copy_modifiers:
                    copy_fcurve_modifiers(fcurve, new_fcurve, replace)

                # Match the color and color mode of the new F-curve to the original
                new_fcurve.color = fcurve.color
                new_fcurve.color_mode = fcurve.color_mode

                # Set the extrapolation of the new F-curve if the option is enabled
                if paste_extrapolation:
                    new_fcurve.extrapolation = fcurve.extrapolation

                # Update the new F-curve to apply changes
                new_fcurve.update()
                self.report({"INFO"}, "F-curves copied to selected bones.")

    else:
        # If the object is not in POSE mode, handle the animation data differently
        for selected_obj in selected:
            active_obj_fcurves = fcurves_active_object(context, selection_mode)

            # Debugging: Print the data path and array index of each F-curve
            if debug_mode:
                print([(fcurve.data_path, fcurve.array_index) for fcurve in active_obj_fcurves])

            # If no F-curves are found, display a warning and exit
            if not active_obj_fcurves:
                if selection_mode == "SELECTED_FCURVES":
                    self.report({"WARNING"}, "Active object has no selected F-Curves.")
                else:
                    self.report({"WARNING"}, "Active object has no animation.")
                return

            # Iterate over each F-curve in the active object
            for fcurve in active_obj_fcurves:
                fc_array_index = fcurve.array_index
                fc_array_index, invert = get_fcurve_array_index(fcurve, invert_quaternion_w, remap_x, remap_y, remap_z)

                fc_dpath = fcurve.data_path
                
                # Determine if the data path should reference object data (e.g., mesh data) or the object itself
                if data:
                    try:
                        # Split the data path into parts and dynamically access attributes of selected_obj.data
                        path_parts = fc_dpath.split('.')
                        attr = selected_obj.data
                        for part in path_parts:
                            attr = getattr(attr, part, None)
                            if attr is None:
                                raise AttributeError
                        valid = attr is not None
                    except AttributeError:
                        # If the attribute doesn't exist, display a warning and skip this F-curve
                        self.report({"WARNING"}, f"'data.{fc_dpath}' does not exist for {selected_obj.name}")
                        continue
                else:
                    try:
                        # Split the data path into parts and dynamically access attributes of selected_obj
                        path_parts = fc_dpath.split('.')
                        attr = selected_obj
                        for part in path_parts:
                            attr = getattr(attr, part, None)
                            if attr is None:
                                raise AttributeError
                        valid = attr is not None
                    except AttributeError:
                        # If the attribute doesn't exist, display a warning and skip this F-curve
                        self.report({"WARNING"}, f"'{fc_dpath}' does not exist for {selected_obj.name}")
                        continue

                # Debugging: Print whether the data path is valid
                if debug_mode:
                    print(f"valid data_path: {valid}")

                # Skip this F-curve if the selection mode is "SELECTED_KEYS" and no keys are selected
                if selection_mode == "SELECTED_KEYS" and not len(get_selected_keys(fcurve)):
                    continue

                fcurve_grp_name = ""

                # Get the group name of the F-curve if it has one
                if fcurve.group:
                    fcurve_grp_name = fcurve.group.name

                # Attempt to retrieve or create an action group for the F-curve's group name
                so_grp = anim_data.action.groups.get(fcurve_grp_name)
                if not so_grp:
                    so_grp = anim_data.action.groups.new(fcurve_grp_name)

                # If the selected object has animation data, create a new F-curve
                if selected_obj.animation_data and selected_obj.animation_data.action:
                    new_fcurve = get_new_fcurve(selected_obj.animation_data, fc_dpath, fc_array_index, fcurve_grp_name, replace)

                    # Debugging: Print the new F-curve information
                    if debug_mode:
                        print(f"new_fcurve: {new_fcurve}")

                    # Copy keyframe points from the original F-curve to the new F-curve
                    for kf in fcurve.keyframe_points:
                        if selection_mode == "SELECTED_KEYS" and not kf.select_control_point:
                            continue
                        new_key = new_fcurve.keyframe_points.insert(kf.co[0], kf.co[1] * invert, options={"FAST"})
                        new_key.interpolation = kf.interpolation
                        new_key.amplitude = kf.amplitude
                        new_key.back = kf.back
                        new_key.easing = kf.easing
                        new_key.period = kf.period
                        new_key.type = kf.type
                        new_key.handle_left_type = kf.handle_left_type
                        new_key.handle_left = Vector((kf.handle_left[0], kf.handle_left[1] * invert))
                        new_key.handle_right_type = kf.handle_right_type
                        new_key.handle_right = Vector((kf.handle_right[0], kf.handle_right[1] * invert))
                        new_key.select_control_point = kf.select_control_point
                        new_key.select_left_handle = kf.select_left_handle
                        new_key.select_right_handle = kf.select_right_handle

                    # Set the extrapolation of the new F-curve if the option is enabled
                    if paste_extrapolation:
                        new_fcurve.extrapolation = fcurve.extrapolation

                else:
                    # If the selected object doesn't have animation data, insert keyframes directly
                    for kf in fcurve.keyframe_points:
                        if selection_mode == "SELECTED_KEYS" and not kf.select_control_point:
                            continue

                        if data:
                            tgt = selected_obj.data
                        else:
                            tgt = selected_obj

                        # Insert keyframe into the target object or its data
                        if fcurve_grp_name != "":
                            tgt.keyframe_insert(data_path=fc_dpath,
                                                index=fc_array_index,
                                                frame=kf.co[0],
                                                group=fcurve_grp_name)
                        else:
                            tgt.keyframe_insert(data_path=fc_dpath, frame=kf.co[0])

                        # Retrieve the newly created F-curve
                        new_fcurve = tgt.animation_data.action.fcurves.find(fc_dpath, index=fc_array_index)

                        # Find the corresponding keyframe in the new F-curve and copy its properties
                        new_key = next(keyframe for keyframe in new_fcurve.keyframe_points if keyframe.co[0] == kf.co[0])
                        new_key.co[1] = kf.co[1] * invert
                        new_key.interpolation = kf.interpolation
                        new_key.amplitude = kf.amplitude
                        new_key.back = kf.back
                        new_key.easing = kf.easing
                        new_key.period = kf.period
                        new_key.type = kf.type
                        new_key.handle_left_type = kf.handle_left_type
                        new_key.handle_left = Vector((kf.handle_left[0], kf.handle_left[1] * invert))
                        new_key.handle_right_type = kf.handle_right_type
                        new_key.handle_right = Vector((kf.handle_right[0], kf.handle_right[1] * invert))
                        new_key.select_control_point = kf.select_control_point
                        new_key.select_left_handle = kf.select_left_handle
                        new_key.select_right_handle = kf.select_right_handle

                # Display a status message in the workspace after copying an F-curve
                status_text = f"{get_fcurve_nice_name(new_fcurve)} copied."
                context.workspace.status_text_set(status_text)
                
                # Copy F-curve modifiers if the option is enabled
                if copy_modifiers:
                    copy_fcurve_modifiers(fcurve, new_fcurve, replace)
                
                # Match the color and color mode of the new F-curve to the original
                new_fcurve.color = fcurve.color
                new_fcurve.color_mode = fcurve.color_mode
                
                # Update the new F-curve to apply changes
                new_fcurve.update()
    
    # Refresh the context to apply all changes
    refresh()


class ANIMCOPY_OT_copy_to_selected(Operator):
    """Copies the specified animation of the active object to the corresponding channels of other selected objects"""
    bl_idname = "graph.copy_to_selected"
    bl_label = "Copy to Selected"
    bl_options = { "REGISTER", "UNDO" }

    selection_mode: EnumProperty(
        name= "Target",
        default="ALL",
        description= "Selection mode to use for Copy to Selected",
        items= [
                ("ALL","All F-Curves","Copies all F-Curves to selected objects","RADIOBUT_ON",0),
                ("TRANSFORMS","Transform F-Curves","Copies only Transform F-Curves (e.g. Location, Scale, Rotation)","ORIENTATION_LOCAL",1),
                ("SELECTED_FCURVES","Selected F-Curves","Copies all frames of selected F-Curves","IPO_EASE_IN_OUT",2),
                ("SELECTED_KEYS","Selected Keyframes","Copies only selected keyframes","KEYFRAME_HLT",3),
                ]
    )

    copy_modifiers: BoolProperty(
        name = "Copy Modifiers",
        default = True,
        description = "Copy F-Curve modifiers of the active object to selected objects",
    )

    replace: BoolProperty(
        name = "Replace",
        default = True,
        description = "If an F-Curve exists, remove it and replace it with the active object's F-Curve",
    )

    invert_quaternion_w: BoolProperty(
        name = "Invert W (Quaternion)",
        default = False,
        description = "If using Quaternion rotation mode, invert the W rotation",
    )
    
    remap_x: EnumProperty(
        name= "Remap X",
        default="X",
        description= "Remap X to another dimension",
        items= [
                ("X","X","Keeps X values the same","",0),
                ("Y","Y","Maps X dimension values to Y","",1),
                ("Z","Z","Maps X dimension values to Z","",2),
                ("-X","-X","Maps X dimension values to -X","",3),
                ("-Y","-Y","Maps X dimension values to -Y","",4),
                ("-Z","-Z","Maps X dimension values to -Z","",5),
        ]
    )

    remap_y: EnumProperty(
        name= "Remap Y",
        default="Y",
        description= "Remap Y to another dimension",
        items= [
                ("X","X","Maps Y dimension values to X","",0),
                ("Y","Y","Keeps Y values the same","",1),
                ("Z","Z","Maps Y dimension values to Z","",2),
                ("-X","-X","Maps Y dimension values to -X","",3),
                ("-Y","-Y","Maps Y dimension values to -Y","",4),
                ("-Z","-Z","Maps Y dimension values to -Z","",5),
        ]
    )

    remap_z: EnumProperty(
        name= "Remap Z",
        default="Z",
        description= "Remap Z to another dimension",
        items= [
                ("X","X","Maps Z dimension values to X","",0),
                ("Y","Y","Maps Z dimension values to Y","",1),
                ("Z","Z","Keeps Z values the same","",2),
                ("-X","-X","Maps Z dimension values to -X","",3),
                ("-Y","-Y","Maps Z dimension values to -Y","",4),
                ("-Z","-Z","Maps Z dimension values to -Z","",5),
        ]
    )

    paste_extrapolation: BoolProperty(
        name="Paste Extrapolation",
        description="Paste the extrapolation type of the copied F-Curve",
        default=True
    )

    @classmethod
    def poll(cls, context):
        sel_objs = context.selected_objects
        if context.object.mode == "POSE":
            sel_objs = context.selected_pose_bones
        return len(sel_objs) > 1

    def execute(self, context):
        prefs = get_prefs()
        global debug_mode
        if prefs:
            debug_mode = prefs.debug_mode
        else:
            debug_mode = True

        anim_types = has_animation(context.object)

        if anim_types == "OBJECT":
            copy_to_selected(self, context, context.object.animation_data,
                             self.copy_modifiers, self.selection_mode, self.replace,
                             self.invert_quaternion_w, self.remap_x, self.remap_y, self.remap_z, self.paste_extrapolation)
        elif anim_types == "DATA":
            copy_to_selected(self, context, context.object.data.animation_data,
                             self.copy_modifiers, self.selection_mode, self.replace,
                             self.invert_quaternion_w, self.remap_x, self.remap_y, self.remap_z, self.paste_extrapolation, True)
        elif anim_types == "BOTH":
            copy_to_selected(self, context, context.object.animation_data,
                             self.copy_modifiers, self.selection_mode, self.replace, 
                             self.invert_quaternion_w, self.remap_x, self.remap_y, self.remap_z, self.paste_extrapolation)
            copy_to_selected(self, context, context.object.data.animation_data,
                             self.copy_modifiers, self.selection_mode, self.replace, 
                             self.invert_quaternion_w, self.remap_x, self.remap_y, self.remap_z, self.paste_extrapolation, True)
        elif not anim_types:
            self.report({"WARNING"},"Active object has no animation.")
            return

        return {'FINISHED'}

classes = (ANIMCOPY_OT_copy_to_selected,)

def register():
    for cls in classes:
        bpy.utils.register_class(cls)

def unregister():
    for cls in reversed(classes):
        bpy.utils.unregister_class(cls)

if __name__ == "__main__":
    register()
