import bpy, math

def refresh():
    # Updates all animation related UI
    for area in bpy.context.screen.areas:
        if area.type in {'VIEW_3D', 'DOPESHEET_EDITOR', 'GRAPH_EDITOR', 'NLA_EDITOR', 'OUTLINER', 'PROPERTIES'}:
            for region in area.regions:
                if (area.type == 'VIEW_3D' and region.type == 'UI') or \
                   (area.type != 'VIEW_3D' and region.type in {'WINDOW', 'CHANNELS'}):
                    region.tag_redraw()

def is_transform(fcurve_data_path):
    prefs = get_prefs()
    debug_mode = None
    if prefs:
        debug_mode = get_prefs().debug_mode
    else:
        debug_mode = True
    transforms = ["location", "rotation_euler", "rotation_quaternion", "scale",
        "delta_location", "delta_rotation_euler", "delta_rotation_quaternion",
        "delta_scale"
        ]
    if debug_mode:
        print("is_transform",fcurve_data_path,fcurve_data_path in transforms)
    return fcurve_data_path in transforms

def get_modifier(modifiers,mtype):
    for m in modifiers:
        if m.type == mtype:
            return m

def get_new_fcurve(anim_data,fc_dpath,fc_array_index,fcurve_grp_name,replace):
    new_fcurve = anim_data.action.fcurves.find(fc_dpath, index=fc_array_index)
    prefs = get_prefs()
    debug_mode = None
    if prefs:
        debug_mode = get_prefs().debug_mode
    else:
        debug_mode = True
    if new_fcurve and replace:
        anim_data.action.fcurves.remove(new_fcurve)
        if fcurve_grp_name != "":
            new_fcurve = anim_data.action.fcurves.new(fc_dpath,
                                                    index=fc_array_index,
                                                    action_group=fcurve_grp_name)
        else:
            new_fcurve = anim_data.action.fcurves.new(fc_dpath, index=fc_array_index)

    elif not new_fcurve:
        if fcurve_grp_name != "":
            new_fcurve = anim_data.action.fcurves.new(fc_dpath,
                                                    index=fc_array_index,
                                                    action_group=fcurve_grp_name)
        else:
            new_fcurve = anim_data.action.fcurves.new(fc_dpath, index=fc_array_index)
   
    if debug_mode:
        print(new_fcurve)

    return new_fcurve

def copy_modifier_values(modifier,new_modifier):
    for name, prop in modifier.rna_type.properties.items():
        if prop and prop.type != "POINTER" and not prop.is_readonly:
            mod_prop_value = getattr(modifier,name)
            setattr(new_modifier,name,mod_prop_value)

def copy_fcurve_modifiers(fcurve, new_fcurve, replace):
    prefs = get_prefs()
    debug_mode = prefs.debug_mode if prefs else True

    # Separate Cycles modifiers and other modifiers
    cycles_mods = [m for m in fcurve.modifiers if m.type == "CYCLES"]
    other_mods = [m for m in fcurve.modifiers if m.type != "CYCLES"]

    # Add the Cycles modifier first
    for cycles_mod in cycles_mods:
        # Remove existing Cycles modifier on the target if replacing
        new_modifier = get_modifier(new_fcurve.modifiers, "CYCLES")
        if new_modifier and replace:
            if debug_mode:
                print(f"Removing existing Cycles modifier: {new_modifier}")
            new_fcurve.modifiers.remove(new_modifier)
            new_modifier = None

        # Create and copy the Cycles modifier
        if not new_modifier:
            try:
                new_modifier = new_fcurve.modifiers.new("CYCLES")
                if debug_mode:
                    print(f"Created Cycles modifier on {new_fcurve.data_path}")
                copy_modifier_values(cycles_mod, new_modifier)
            except Exception as e:
                print(f"Failed to create Cycles modifier: {e}")
                continue

    # **Add this loop to handle other modifiers**
    for modifier in other_mods:
        # Remove existing modifier of the same type if replacing
        new_modifier = get_modifier(new_fcurve.modifiers, modifier.type)
        if new_modifier and replace:
            if debug_mode:
                print(f"Removing existing modifier: {new_modifier}")
            new_fcurve.modifiers.remove(new_modifier)
            new_modifier = None

        # Create and copy the modifier
        if not new_modifier:
            try:
                new_modifier = new_fcurve.modifiers.new(modifier.type)
                if debug_mode:
                    print(f"Created new modifier: {new_modifier.type} on {new_fcurve.data_path}")
                copy_modifier_values(modifier, new_modifier)
            except Exception as e:
                print(f"Failed to create new modifier of type {modifier.type}: {e}")
                continue


def get_prefs():
    # Attempt to determine the add-on name dynamically from available addons
    addon_name = 'animcopy'  # Replace with the unique part of your add-on's name
    addons = bpy.context.preferences.addons
    
    # Loop through all add-ons to find an exact match with the last part of each key
    for key in addons.keys():
        # Split the key by '.' and check the last element
        if key.split('.')[-1].replace(' ', '_').lower() == addon_name.lower():
            return addons[key].preferences  # Return preferences if exact match is found
    
    return None  # Return None if no exact match is found

def rename(dp_name):
    dp_name = "euler_rotation" if dp_name == "rotation_euler" else dp_name
    dp_name = "quaternion_rotation" if dp_name == "rotation_quaternion" else dp_name
    dp_name = "delta_euler_rotation" if dp_name == "delta_rotation_euler" else dp_name
    dp_name = "delta_quaternion_rotation" if dp_name == "delta_rotation_quaternion" else dp_name
    new_name = dp_name.replace("_"," ")
    new_name = new_name.title()
    return new_name

def get_fcurve_nice_name(fcurve):
    print("fcurve.array_index", fcurve.array_index, "fcurve.data_path", fcurve.data_path)
    name = ""
    axis_name = ["X", "Y", "Z"]
    quat_name = ["W", "X", "Y", "Z"]

    # Extract property name and object name (if any) from data path
    data_path = fcurve.data_path
    object_name = ""

    if 'pose.bones[' in data_path:
        # Example: 'pose.bones["bone_name"].rotation_euler'
        start_idx = data_path.index('["') + 2
        end_idx = data_path.index('"]')
        object_name = data_path[start_idx:end_idx]
        property_name = data_path[end_idx + 3:]
    else:
        property_name = data_path

    # Use rename function to format the property name
    formatted_name = rename(property_name)

    # Determine axis name or quaternion component
    if "quaternion" in property_name:
        axis = quat_name[fcurve.array_index]
    elif "rotation_euler" in property_name or "location" in property_name or "scale" in property_name or "delta" in property_name:
        axis = axis_name[fcurve.array_index]
    else:
        # Non-transform properties (use the property name as is)
        return f"{property_name}"

    # Format the final name
    if object_name:
        name = f"{axis} {formatted_name} ({object_name})"
    else:
        name = f"{axis} {formatted_name}"

    return name

def get_selected_no_active_obj(context):
    prefs = get_prefs()
    debug_mode = None
    if prefs:
        debug_mode = prefs.debug_mode
    else:
        debug_mode = True

    active_obj = context.object
    if active_obj.mode == "POSE":
        active_bone = context.active_pose_bone
        sel_bones = context.selected_pose_bones[:]
        for pose_bone in sel_bones:
            if pose_bone == active_bone:
                sel_bones.remove(pose_bone)
                print("active bone is",pose_bone)
                return sel_bones
    else:
        sel_objs = context.selected_objects[:]
        for obj in sel_objs:
            if obj == active_obj:
                sel_objs.remove(obj)
                if debug_mode:
                    print("active obj is",obj)
                return sel_objs

def has_animation(obj):
    types = []
    anim_data = obj.animation_data
    if anim_data and anim_data.action:
        if len(anim_data.action.fcurves):
            types.append("OBJECT")
    if obj.data:
        anim_data = obj.data.animation_data
        if anim_data and anim_data.action:
            if len(anim_data.action.fcurves):
                types.append("DATA")
    if len(types) > 2:
        return "BOTH"
    elif len(types) == 1:
        return types[0]
    else:
        return
   

def fcurves_active_object(context, selection_mode):
    prefs = get_prefs()
    debug_mode = None
    if prefs:
        debug_mode = prefs.debug_mode
    active_obj = context.object
    active_fcurves = []
    if active_obj.mode == "POSE":
        fcurves = active_obj.animation_data.action.fcurves
        if context.active_pose_bone:
            bone_name = context.active_pose_bone.name
            for fcurve in fcurves:
                if fcurve.data_path.startswith("pose.bones"):
                    if fcurve.data_path.split('"')[1] == bone_name:
                        if selection_mode == "SELECTED_FCURVES":
                            if fcurve.select:
                                active_fcurves.append(fcurve)
                        elif selection_mode == "TRANSFORMS":
                            if is_transform(fcurve.data_path):
                                active_fcurves.append(fcurve)
                        else:
                            active_fcurves.append(fcurve)
                else:
                    if debug_mode:
                        print("active_obj",active_obj)
                    has_anim = has_animation(active_obj)

                    anim_datas = []
                    if has_anim == "OBJECT":
                        anim_datas.append(active_obj.animation_data)
                    if has_anim == "DATA":
                        anim_datas.append(active_obj.data.animation_data)
                    if has_anim == "BOTH":
                        anim_datas.append(active_obj.animation_data)
                        anim_datas.append(active_obj.data.animation_data)
                    if not has_anim:
                        return active_fcurves

                    for anim_data in anim_datas:
                        if not len(anim_data.action.fcurves):
                            anim_data = active_obj.data.animation_data
                            if debug_mode:
                                print("No anim_data. Trying object data. anim_data.data fcurve count:",len(anim_data.action.fcurves))
                        fcurves = anim_data.action.fcurves
                        for fcurve in fcurves:
                            if debug_mode:
                                print(get_fcurve_nice_name(fcurve),"selected:",selection_mode=="SELECTED_FCURVES","fcurve.select:",fcurve.select)
                            if selection_mode=="SELECTED_FCURVES":
                                if fcurve.select:
                                    active_fcurves.append(fcurve)
                            elif selection_mode=="TRANSFORMS":
                                if is_transform(fcurve.data_path):
                                    active_fcurves.append(fcurve)
                            else:
                                active_fcurves.append(fcurve)     
    else:
        if debug_mode:
            print("active_obj",active_obj)
        has_anim = has_animation(active_obj)

        anim_datas = []
        if has_anim == "OBJECT":
            anim_datas.append(active_obj.animation_data)
        if has_anim == "DATA":
            anim_datas.append(active_obj.data.animation_data)
        if has_anim == "BOTH":
            anim_datas.append(active_obj.animation_data)
            anim_datas.append(active_obj.data.animation_data)
        if not has_anim:
            return active_fcurves

        for anim_data in anim_datas:
            if not len(anim_data.action.fcurves):
                anim_data = active_obj.data.animation_data
                if debug_mode:
                    print("No anim_data. Trying object data. anim_data.data fcurve count:",len(anim_data.action.fcurves))
            fcurves = anim_data.action.fcurves
            for fcurve in fcurves:
                if debug_mode:
                    print(get_fcurve_nice_name(fcurve),"selected:",selection_mode=="SELECTED_FCURVES","fcurve.select:",fcurve.select)
                if selection_mode=="SELECTED_FCURVES":
                    if fcurve.select:
                        active_fcurves.append(fcurve)
                elif selection_mode=="TRANSFORMS":
                    if is_transform(fcurve.data_path):
                        active_fcurves.append(fcurve)
                else:
                    active_fcurves.append(fcurve)                
    return active_fcurves

def get_handle_length(x1, x2, y1, y2):
    a = x1 - x2
    b = y1 - y2

    c = math.sqrt( a*a + b*b )
    return c

def sub_v2_v2v2(a, b):
    r = [0,0]
    r[0] = a[0] - b[0]
    r[1] = a[1] - b[1]
    return r

def dot_v2v2(a, b):
    return a[0] * b[0] + a[1] * b[1]

def mul_v2_v2fl(r, a, f):
    r[0] = a[0] * f
    r[1] = a[1] * f

def zero_v2(r):
    r[0] = 0.0
    r[1] = 0.0

def normalize_v2_v2_length(r, a, unit_length):
    d = dot_v2v2(a, a)
    if (d > 1.0e-35):
        d = math.sqrt(d)
        mul_v2_v2fl(r, a, unit_length / d)
    else:
        # Either the vector is small or one of it's values contained `nan`.
        zero_v2(r)
        d = 0.0
    return d

def normalize_v2_length(n, unit_length):
    return normalize_v2_v2_length(n, n, unit_length)

def add_v2_v2v2(r, a, b):
    r[0] = a[0] + b[0]
    r[1] = a[1] + b[1]
    return r

def get_fcurve_selected_keys():
    selected_keys = []
    fcurves = bpy.context.selected_editable_fcurves[:]

    for keyframe in fcurves[0].keyframe_points:
        if keyframe.select_control_point:
            selected_keys.append(keyframe)

    return fcurves[0],selected_keys

def get_selected_keys(fcurve):
    selected_keys = []

    for keyframe in fcurve.keyframe_points:
        if keyframe.select_control_point:
            selected_keys.append(keyframe)

    return selected_keys
