import bpy, os, math
from pathlib import Path
from ..utils.utils import refresh, max_diff, get_prefs
from bpy.types import Operator
from bpy.props import (EnumProperty,BoolProperty,IntProperty,FloatProperty)
from ..icons.custom_icons import get_icon

class GRAPH_OT_Delete_Static_Channels(Operator):
    """Delete F-Curve channels based on each F-Curve's number of keyframes and """\
    """the difference between its keyframes' values"""
    bl_idname = "graph.delete_static_channels"
    bl_label = "Delete Static Channels"
    bl_options = { "REGISTER", "UNDO" }

    filters: EnumProperty(
        name= "Channel Filter",
        description= "Filter which channels can get deleted.",
        default='VIS_FCURVE',
        items= [("ALL","All","Delete all of a scene's static channels.","RADIOBUT_ON",0),
                ("SEL_OBJS","Selected Objects/Bones","Only delete static channels of selected objects or pose bones.","OBJECT_DATA",1),
                ("SEL_FCURVE","Selected F-Curves","Only delete static channels of selected F-Curves.","RESTRICT_SELECT_OFF",2),
                ("VIS_FCURVE","Visible F-Curves","Only delete static channels of visible F-Curves.","HIDE_OFF",3),
        ]
    )

    keep_modifiers: BoolProperty(
        name = "Keep Channels with F-Curve Modifiers",
        description = "Saves channels with F-Curve modifiers from deletion",
        default = True)

    keep_constraints: BoolProperty(
        name = "Keep Constraint Channels",
        description = "Saves Constraint channels with F-Curves from deletion",
        default = True)

    min_key_count: IntProperty(
        name = "Min. Key Count",
        description = "Minimum number of keys required to avoid deletion",
        default = 2,
        soft_min = 0, soft_max = 20)

    min_difference: FloatProperty(
        name = "Min. Difference",
        description = "Minimum difference of an F-Curves keyframe values",
        default = 0.001,
        soft_min = 0.0001, soft_max = 10)

    def print_removed_fcurve(self,action,fcurve):
        print("Action %s, F-Curve %s[%s] was removed."%(action.name,fcurve.data_path,fcurve.array_index))

    def delete_selected_channels(self, action, fcurve, debug_mode=False):
        rotations = ["rotation_euler", "rotation_quaternion", "delta_rotation_euler", "delta_rotation_quaternion"]
        is_constraint_channel = "constraints[" in fcurve.data_path

        if (self.keep_constraints and is_constraint_channel) or fcurve.lock:
            return 0  # Skip deletion for constraint channels

        if (len(fcurve.keyframe_points) < self.min_key_count):
            if len(fcurve.modifiers.items()) == 0 or self.keep_modifiers == False:
                if debug_mode:
                    self.print_removed_fcurve(action, fcurve)
                action.fcurves.remove(fcurve)
                return 1
        else:
            values = [math.degrees(k.co[1]) if fcurve.data_path in rotations else k.co[1] for k in fcurve.keyframe_points]
            if max_diff(values) < self.min_difference:
                if debug_mode:
                    self.print_removed_fcurve(action, fcurve)
                action.fcurves.remove(fcurve)
                return 1
        return 0

    # @classmethod
    # def poll(self,context):
    #     return context.visible_fcurves

    def execute(self, context):
        bpy.ops.ed.undo_push(message=self.__class__.bl_label)
        vis_curves = None

        prefs = get_prefs()
        debug_mode = None
        if prefs:
            debug_mode = get_prefs().debug_mode
        else:
            debug_mode = True

        if self.filters == "VIS_FCURVE":
            vis_curves = context.visible_fcurves[:]

        if not context.object:
            print("Please make sure an object is active.")
            self.report({"WARNING"}, "Please make sure an object is active.")
            return {"CANCELLED"}

        current_mode = context.object.mode

        total_curves = 0
        delete_count = 0
        
        if current_mode == "POSE" and self.filters == "SEL_OBJS":
            action = context.active_object.animation_data.action
            sel_bone_names = list()

            for pb in context.selected_pose_bones:
                sel_bone_names.append(pb.bone.name)

            for fcurve in action.fcurves:
                tmp = fcurve.data_path.split("[", maxsplit=1)[1].split("]", maxsplit=1)
                bone_name = tmp[0][1:-1]
                if bone_name in sel_bone_names:
                    delete_count += self.delete_selected_channels(action,fcurve,debug_mode)
                    total_curves += 1
        else:
            for action in bpy.data.actions:
                if self.filters == "SEL_OBJS":
                    for action in bpy.data.actions:
                        for o in bpy.data.objects:
                            if o.animation_data and o.animation_data.action is action:
                                if o.select_get():
                                    for fcurve in action.fcurves:
                                        total_curves += 1
                                        delete_count += self.delete_selected_channels(action,fcurve,debug_mode)
                elif self.filters == "SEL_FCURVE":
                    for fcurve in action.fcurves:
                        if fcurve.select:
                            total_curves += 1
                            delete_count += self.delete_selected_channels(action,fcurve,debug_mode)
                elif self.filters == "VIS_FCURVE":
                    for fcurve in action.fcurves:
                        if fcurve in vis_curves:
                            total_curves += 1
                            delete_count += self.delete_selected_channels(action,fcurve,debug_mode)
                else:
                    for fcurve in action.fcurves:
                        total_curves += 1
                        delete_count += self.delete_selected_channels(action,fcurve,debug_mode)

        if delete_count == 1:
            context.workspace.status_text_set("%s static channel of %s channels deleted."%(delete_count,total_curves))
        elif delete_count > 1:
            context.workspace.status_text_set("%s static channels of %s channels deleted."%(delete_count,total_curves))
        else:
            context.workspace.status_text_set("No static channels of %s channels deleted."%total_curves)

        for area in context.screen.areas:
            if area.type == 'VIEW_3D':
                for region in area.regions:
                    if region.type == 'UI':
                        region.tag_redraw()
            elif area.type == 'DOPESHEET_EDITOR':
                for region in area.regions:
                    if region.type == 'WINDOW':
                        region.tag_redraw()
                    if region.type == 'CHANNELS':
                        region.tag_redraw()
            elif area.type == 'GRAPH_EDITOR':
                for region in area.regions:
                    if region.type == 'WINDOW':
                        region.tag_redraw()
                    if region.type == 'CHANNELS':
                        region.tag_redraw()
            elif area.type == 'NLA_EDITOR':
                for region in area.regions:
                    if region.type == 'WINDOW':
                        region.tag_redraw()
                    if region.type == 'CHANNELS':
                        region.tag_redraw()
            elif area.type == 'OUTLINER':
                for region in area.regions:
                    if region.type == 'WINDOW':
                        region.tag_redraw()
            elif area.type == 'PROPERTIES':
                for region in area.regions:
                    if region.type == 'WINDOW':
                        region.tag_redraw()
        return {'FINISHED'}

def dsc_menu_func(self, context):
    self.layout.operator("graph.delete_static_channels", icon_value=get_icon("Delete_Static_Channels"))

def register():
    bpy.utils.register_class(GRAPH_OT_Delete_Static_Channels)
    bpy.types.GRAPH_MT_channel.append(dsc_menu_func)
    bpy.types.DOPESHEET_MT_channel.append(dsc_menu_func)

def unregister():
    bpy.types.GRAPH_MT_channel.remove(dsc_menu_func)
    bpy.types.DOPESHEET_MT_channel.remove(dsc_menu_func)
    bpy.utils.unregister_class(GRAPH_OT_Delete_Static_Channels)
