import bpy

bad_handle_types = ["ALIGNED", "AUTO","AUTO_CLAMPED"]

def copy_key(keyframe, key_index=None):
    new_key = Keyframe(
        [keyframe.co[0], keyframe.co[1]],
        [keyframe.handle_left[0], keyframe.handle_left[1]],
        [keyframe.handle_right[0], keyframe.handle_right[1]],
        key_index=key_index
    )
    new_key.interpolation = keyframe.interpolation
    new_key.amplitude = keyframe.amplitude
    new_key.back = keyframe.back
    new_key.easing = keyframe.easing
    new_key.period = keyframe.period
    new_key.type = keyframe.type
    new_key.handle_left_type = keyframe.handle_left_type
    new_key.handle_right_type = keyframe.handle_right_type
    return new_key

def get_fcurve_selected_keys(fcurve, no_limits=False):
    selected_keys = []

    for keyframe in fcurve.keyframe_points:
        if keyframe.select_control_point:
            selected_keys.append(keyframe)

    if no_limits:
        selected_keys = selected_keys[1:-1]

    return selected_keys

def get_selected_keys(fcurve, mode, exclude_limits=False):
    selection = []

    if mode == "SELECTED_FCURVE":
        selection = fcurve.keyframe_points

    elif mode == "SELECTED_KEYS":
        selection = get_fcurve_selected_keys(fcurve)

    elif mode == "PLAYBACK_RANGE":
        context = bpy.context
        scene = context.scene

        # Set the keyframe at scene.frame_start
        value_start = fcurve.evaluate(scene.frame_start)
        fcurve.keyframe_points.insert(scene.frame_start, value_start)

        # Set the keyframe at scene.frame_end
        value_end = fcurve.evaluate(scene.frame_end)
        fcurve.keyframe_points.insert(scene.frame_end, value_end)

        # Iterate over the keyframe points in the F-Curve
        for keyframe in fcurve.keyframe_points:
            # Check if the keyframe is within the playback frame range
            if scene.frame_start <= keyframe.co.x <= scene.frame_end:
                selection.append(keyframe)

    if exclude_limits:
        selection = selection[1:-1]

    return selection

class Loop:
    def __init__(self, keys=[], fcurve=None, selection_mode=None, mirrored=False):
        self.keys = keys
        self.mirrored = mirrored
        if fcurve:
            self.add_keys_from_fcurve(fcurve, selection_mode)

    def add_key(self, key):
        self.keys.append(key)

    def add_keys_from_fcurve(self, fcurve, selection_mode):
        selection = None
        self.keys = []
        selection = get_selected_keys(fcurve, selection_mode)
        first_key = selection[0]
        last_key = selection[-1]

        if first_key.handle_right_type in bad_handle_types:
            first_key.handle_left_type = "FREE"
            first_key.handle_right_type = "FREE"
        if last_key.handle_left_type in bad_handle_types:
            last_key.handle_left_type = "FREE"
            last_key.handle_right_type = "FREE"

        for ki, keyframe in enumerate(selection):
            new_key = copy_key(keyframe, key_index=ki)
            self.keys.append(new_key)

    @property
    def frame_length(self):
        return self.keys[-1].co[0] - self.keys[0].co[0]

    @property
    def value_change(self):
        return self.keys[-1].co[1] - self.keys[0].co[1]

    def shift_frames(self, frame_shift):
        for x in self.keys:
            x.co[0] += frame_shift
            x.handle_left[0] += frame_shift
            x.handle_right[0] += frame_shift

    def shift_values(self, value_shift):
        for x in self.keys:
            x.co[1] += value_shift
            x.handle_left[1] += value_shift
            x.handle_right[1] += value_shift

    def mirror(self):
        dup_keys = []
        dup_keys = [k.duplicate() for k in self.keys]
        pivot = dup_keys[0].co[0]
        for k in dup_keys:
            key_distance = pivot - k.co[0]
            hl_distance = pivot - k.handle_left[0]
            hr_distance = pivot - k.handle_right[0]
            hl_val = float(k.handle_left[1])
            hr_val = float(k.handle_right[1])
            k.co[0] = pivot - (key_distance * -1)
            k.handle_left[0] = pivot - (hr_distance * -1)
            k.handle_left[1] = hr_val
            k.handle_right[0] = pivot - (hl_distance * -1)
            k.handle_right[1] = hl_val
        dup_keys.reverse()
        return Loop(dup_keys, mirrored=True)

    def duplicate(self):
        dup_keys = []
        dup_keys = [k.duplicate() for k in self.keys]
        return Loop(dup_keys)

    def __repr__(self):
        loop_str = ""
        for x in self.keys:
            loop_str += str(x) + "\n"
        return loop_str

class FCurve:
    def __init__(self, data_path, array_index, keys=[], action=None):
        self.data_path = data_path
        self.array_index = array_index
        self.keys = keys
        self.action = action

    def get_obj(self):
        fcurve = self.get_fcurve()
        return next(o for o in bpy.data.objects if o.animation_data and o.animation_data.action is fcurve.id_data)

    def get_fcurve(self):
        for action in bpy.data.actions:
            for fcu in action.fcurves:
                if action.name == self.action and fcu.data_path == self.data_path and fcu.array_index == self.array_index:
                    return fcu

    def nice_fcurve_name(self):
        axis_name = ["X", "Y", "Z"]
        quat_name = ["W", "X", "Y", "Z"]
        data_path = self.data_path
        array_index = self.array_index

        if data_path in ["location", "rotation_euler", "rotation_quaternion", "scale"]:
            if "quaternion" in data_path:
                return f"{quat_name[array_index]} {data_path}"
            else:
                return f"{axis_name[array_index]} {data_path}"
        else:
            return data_path

    def __repr__(self):
        return f"Action: {self.action}, F-Curve: {self.nice_fcurve_name()}"

class Keyframe:
    def __init__(
        self,
        co=[],
        handle_left=[],
        handle_right=[],
        interpolation="CONSTANT",
        amplitude=0.0,
        back=0.0,
        easing="AUTO",
        period=0.0,
        type="KEYFRAME",
        handle_left_type="FREE",
        handle_right_type="FREE",
        key_index=None,
        select_control_point=False,
        select_left_handle=False,
        select_right_handle=False,
    ):
        self.co = co
        self.handle_left = handle_left
        self.handle_right = handle_right
        self.interpolation = interpolation
        self.amplitude = amplitude
        self.back = back
        self.easing = easing
        self.period = period
        self.type = type
        self.handle_left_type = handle_left_type
        self.handle_right_type = handle_right_type
        self.index = key_index
        self.select_control_point = select_control_point
        self.select_left_handle = select_left_handle
        self.select_right_handle = select_right_handle

    def duplicate(self):
        return Keyframe(
            list(self.co),
            list(self.handle_left),
            list(self.handle_right),
            str(self.interpolation),
            float(self.amplitude),
            float(self.back),
            str(self.easing),
            float(self.period),
            str(self.type),
            str(self.handle_left_type),
            str(self.handle_right_type),
            int(self.index),
            str(self.select_control_point),
            str(self.select_left_handle),
            str(self.select_right_handle),
        )

    def __repr__(self):
        return f"Keyframe\nindex: {self.index}\n\tco: ({self.co[0]}, {self.co[1]}),\n\thandle_left: "\
            f"({self.handle_left[0]}, {self.handle_left[1]})\n\thandle_right: ({self.handle_right[0]}, {self.handle_right[1]})"
