import bpy
from bpy.types import Operator
from ..utils.utils import get_prefs
from bpy.props import EnumProperty, FloatProperty

prefs = None
debug_mode = False

def get_selected_keys(fcurve, selection_mode):
    if selection_mode == 'SELECTED_FCURVE':
        return fcurve.keyframe_points
    elif selection_mode == 'SELECTED_KEYS':
        return [k for k in fcurve.keyframe_points if k.select_control_point]
    elif selection_mode == 'PLAYBACK_RANGE':
        start_frame = bpy.context.scene.frame_start
        end_frame = bpy.context.scene.frame_end
        return [k for k in fcurve.keyframe_points if start_frame <= k.co[0] <= end_frame]

class GRAPHKIT_OT_align(Operator):
    """Aligns keyframe points of selected F-Curves"""
    bl_idname = "graph.align"
    bl_label = "Align"
    bl_options = {"REGISTER", "UNDO"}

    selection_mode: EnumProperty(
        name="Target",
        default="SELECTED_FCURVE",
        description="Amplify target to use",
        items=[
            ("SELECTED_FCURVE", "Selected F-Curves", "Align all frames of selected F-Curves.", "IPO_EASE_IN_OUT", 0),
            ("SELECTED_KEYS", "Selected Keyframes", "Align selected keyframes.", "KEYFRAME_HLT", 1),
            ("PLAYBACK_RANGE", "Playback Range", "Align the visible F-Curves in the Playback Range.", "TIME", 2),
        ],
    )

    align_mode: EnumProperty(
        name="Mode",
        description="Repeating mode to use before the first keyframe",
        default="MIN",
        items=[
            ("MIN", "Minimum", "Aligns keyframe points of selected F-Curves to the minimum keyframe point."),
            ("CENTER", "Center", "Aligns keyframe points of selected F-Curves to the centermost keyframe point."),
            ("MAX", "Maximum", "Aligns keyframe points of selected F-Curves to the maximum keyframe point."),
        ],
    )

    amount: FloatProperty(
        name="Amount",
        description="Amount by which to align selected keyframes, as a percentage of the distance to the target frame.",
        default=100.0,
        soft_min=-100,
        soft_max=100,
    )

    dimension: EnumProperty(
        name="Dimension",
        description="Dimension to align (X or Y)",
        default="X",
        items=[
            ("X", "X", "Align along X dimension"),
            ("Y", "Y", "Align along Y dimension"),
        ],
    )

    @classmethod
    def poll(cls, context):
        return context.selected_visible_fcurves

    def get_target(self, fcurves):
        all_keyframes = []
        for fcurve in fcurves:
            selected_keyframes = get_selected_keys(fcurve, self.selection_mode)
            all_keyframes.extend((kf, fcurve) for kf in selected_keyframes)
        
        if not all_keyframes:
            return None, None
        
        if self.align_mode == 'MIN':
            target_keyframe, target_fcurve = min(all_keyframes, key=lambda kf: kf[0].co[0] if self.dimension == "X" else kf[0].co[1])
        elif self.align_mode == 'MAX':
            target_keyframe, target_fcurve = max(all_keyframes, key=lambda kf: kf[0].co[0] if self.dimension == "X" else kf[0].co[1])
        elif self.align_mode == 'CENTER':
            frames = [kf[0].co[0] if self.dimension == "X" else kf[0].co[1] for kf in all_keyframes]
            target_value = sum(frames) / len(frames)
            target_keyframe, target_fcurve = min(all_keyframes, key=lambda kf: abs((kf[0].co[0] if self.dimension == "X" else kf[0].co[1]) - target_value))
        
        target_value = target_keyframe.co[0] if self.dimension == "X" else target_keyframe.co[1]
        return target_value, target_fcurve

    def align_keyframes(self, fcurve, keyframes, target_value):
        values = [kf.co[0] if self.dimension == "X" else kf.co[1] for kf in keyframes]
        fcurve_target = None
        if self.align_mode == 'MIN':
            fcurve_target = min(values)
        elif self.align_mode == 'MAX':
            fcurve_target = max(values)
        elif self.align_mode == 'CENTER':
            fcurve_target = sum(values) / len(values)

        for kf in keyframes:
            if self.dimension == "X":
                difference = fcurve_target - target_value
                handle_left_frame_offset = kf.co[0] - kf.handle_left[0]
                handle_right_frame_offset = kf.co[0] - kf.handle_right[0]
                adjustment = difference * (self.amount / 100.0)
                delta = kf.co[0] - adjustment
                kf.co[0] = delta
                kf.handle_left[0] = delta - handle_left_frame_offset
                kf.handle_right[0] = delta - handle_right_frame_offset
            elif self.dimension == "Y":
                difference = fcurve_target - target_value
                handle_left_value_offset = kf.co[1] - kf.handle_left[1]
                handle_right_value_offset = kf.co[1] - kf.handle_right[1]
                adjustment = difference * (self.amount / 100.0)
                delta = kf.co[1] - adjustment
                kf.co[1] = delta
                kf.handle_left[1] = delta - handle_left_value_offset
                kf.handle_right[1] = delta - handle_right_value_offset

    def execute(self, context):
        fcurves = context.selected_editable_fcurves[:]

        global prefs, debug_mode
        prefs = get_prefs()
        if prefs:
            debug_mode = prefs.debug_mode
        else:
            debug_mode = True
        
        target_value, target_fcurve = self.get_target(fcurves)
        if target_value is None:
            return {"CANCELLED"}
        
        for fcurve in fcurves:
            if fcurve == target_fcurve and self.align_mode != "CENTER":
                continue  # Skip this F-Curve as it contains the target value

            if debug_mode:
                print("-" * 40, f"fcurve {fcurve.data_path}[{fcurve.array_index}]", "-" * 40)
            
            selected_keyframes = get_selected_keys(fcurve, self.selection_mode)
            if not selected_keyframes:
                continue

            self.align_keyframes(fcurve, selected_keyframes, target_value)

        self.report({"INFO"},f"{len(fcurves)} aligned in {self.dimension} to {self.align_mode} by {self.amount}.")

        return {"FINISHED"}

def register():
    bpy.utils.register_class(GRAPHKIT_OT_align)

def unregister():
    bpy.utils.unregister_class(GRAPHKIT_OT_align)
