import bpy
from bpy.types import Operator
from ..utils.utils import get_prefs, get_selected_keys
from bpy.props import EnumProperty, FloatProperty

prefs = None
debug_mode = False

class GRAPHKIT_OT_amplify(Operator):
    """Adjust the intensity of the selected F-Curves or keyframe range"""
    bl_idname = "graph.amplify"
    bl_label = "Amplify"
    bl_options = {"REGISTER", "UNDO"}

    selection_mode: EnumProperty(
        name="Target",
        default="SELECTED_FCURVE",
        description="Amplify target to use",
        items=[
            ("SELECTED_FCURVE", "Selected F-Curves", "Amplify all frames of selected F-Curves.", "IPO_EASE_IN_OUT", 0),
            ("SELECTED_KEYS", "Selected Keyframes", "Amplify selected keyframes.", "KEYFRAME_HLT", 1),
            ("PLAYBACK_RANGE", "Playback Range", "Amplify the visible F-Curves in the Playback Range.", "TIME", 2),
        ],
    )

    pivot_point: EnumProperty(
        name="Pivot Point",
        description="Repeating mode to use before the first keyframe",
        default="CENTER",
        items=[
            ("CENTER", "Center", "Amplify keys from each F-Curve's center point."),
            ("FIRST", "First", "Amplify keys from each F-Curve's first keyframe."),
            ("LAST", "Last", "Amplify keys from each F-Curve's last keyframe."),
            ("MAX", "Max", "Amplify keys from each F-Curve's keyframe with maximum value."),
            ("MIN", "Min", "Amplify keys from each F-Curve's keyframe with minimum value."),
        ],
    )

    dimension: EnumProperty(
        name="Dimension",
        default="Y",
        description="Graph dimension to use",
        items=[
            ("X", "X", "Amplify keyframes on the X Dimension."),
            ("Y", "Y", "Amplify keyframes on the Y Dimension."),
        ],
    )

    amount: FloatProperty(
        name="Amount",
        description="Amount by which to amplify selected F-Curves.",
        default=1.0,
        soft_min=-100,
        soft_max=100,
    )

    def get_bounding_box(self, fcurve):
        points = list()
        if self.selection_mode == "SELECTED_KEYS":
            points = [x for x in fcurve.keyframe_points if x.select_control_point == True]
            if len(points) == 0:
                return None
        elif self.selection_mode == "PLAYBACK_RANGE":
            points = get_selected_keys(fcurve, self.selection_mode)
            if len(points) == 0:
                return None
        else:
            points = fcurve.keyframe_points

        if self.dimension == "X":
            key_values = list(map(lambda x: x.co[0], points))
        else:
            key_values = list(map(lambda x: x.co[1], points))

        maximum = max(key_values)
        minimum = min(key_values)

        if self.pivot_point == "CENTER":
            return (maximum + minimum) / 2
        elif self.pivot_point == "FIRST":
            return key_values[0]
        elif self.pivot_point == "LAST":
            return key_values[-1]
        elif self.pivot_point == "MAX":
            return maximum
        elif self.pivot_point == "MIN":
            return minimum
        return None

    def amplify(self, pivot, k):
        if self.dimension == "X":
            key_distance = pivot - k.co[0]
            hl_distance = pivot - k.handle_left[0]
            hr_distance = pivot - k.handle_right[0]
            k.co[0] = pivot - (key_distance * self.amount)
            k.handle_left[0] = pivot - (hl_distance * self.amount)
            k.handle_right[0] = pivot - (hr_distance * self.amount)
        else:
            key_distance = pivot - k.co[1]
            hl_distance = pivot - k.handle_left[1]
            hr_distance = pivot - k.handle_right[1]
            k.co[1] = pivot - (key_distance * self.amount)
            k.handle_left[1] = pivot - (hl_distance * self.amount)
            k.handle_right[1] = pivot - (hr_distance * self.amount)

        global debug_mode
        if debug_mode:
            if self.dimension == "X":
                print(
                    f" key_distance: {round(key_distance, 2)}\n",
                    f"hl_distance: {round(hl_distance, 2)}\n",
                    f"hr_distance: {round(hr_distance, 2)}\n",
                    f"k.co[0]: {round(k.co[0], 2)}\n",
                    f"k.handle_left[0]: {round(k.handle_left[0], 2)}\n",
                    f"k.handle_right[0]: {round(k.handle_right[0], 2)}\n",
                )
            else:
                print(
                    f" key_distance: {round(key_distance, 2)}\n",
                    f"hl_distance: {round(hl_distance, 2)}\n",
                    f"hr_distance: {round(hr_distance, 2)}\n",
                    f"k.co[1]: {round(k.co[1], 2)}\n",
                    f"k.handle_left[1]: {round(k.handle_left[1], 2)}\n",
                    f"k.handle_right[1]: {round(k.handle_right[1], 2)}\n",
                )

    @classmethod
    def poll(cls, context):
        return context.selected_visible_fcurves

    def execute(self, context):
        fcurves = context.selected_editable_fcurves[:]

        global prefs, debug_mode
        prefs = get_prefs()
        if prefs:
            debug_mode = prefs.debug_mode
        else:
            debug_mode = True            

        for fcurve in fcurves:
            print("-" * 40,
                  f"fcurve {fcurve.data_path}[{fcurve.array_index}]",
                  "-" * 40)
            pivot = self.get_bounding_box(fcurve)
            if debug_mode:
                print(f" pivot: {round(pivot, 2)}")
            if pivot is not None:
                selection = get_selected_keys(fcurve, self.selection_mode)
                for k in selection:
                    if debug_mode:
                        print("*" * 20, f"keyframe {k.co[0] if self.dimension == 'X' else k.co[1]}", "*" * 20)
                    self.amplify(pivot, k)

        return {"FINISHED"}


def register():
    bpy.utils.register_class(GRAPHKIT_OT_amplify)


def unregister():
    bpy.utils.unregister_class(GRAPHKIT_OT_amplify)
