import bpy
from bpy.types import Operator
from ..utils.utils import (refresh,get_selected_keys,
    get_prefs,update_fcurves,snap)
from bpy.props import (EnumProperty,
                       BoolProperty,
                       FloatProperty)


prefs = None
debug_mode = False
class GRAPHKIT_OT_distribute(Operator):
    """Distribute Keyframes by a specified frame amount"""
    bl_idname = "graph.distribute"
    bl_label = "Distribute"
    bl_options = { "REGISTER", "UNDO" }

    selection_mode: EnumProperty(
        name= "Target",
        default="SELECTED_FCURVE",
        description= "Selection to use for repeating",
        items= [
            ("SELECTED_FCURVE","Selected F-Curves","Distribute all frames of selected F-Curves.","IPO_EASE_IN_OUT",0),
            ("SELECTED_KEYS","Selected Keyframes","Distribute selected keyframes.","KEYFRAME_HLT",1),
            ("PLAYBACK_RANGE", "Playback Range", "Distribute the keyframes in the Playback Range.", "TIME", 2),
        ]
    )

    distribute_mode: EnumProperty(
        name= "Mode",
        default="FRAME",
        description= "Distribute by frame or note",
        items= [("FRAME","Frames","Distribute frames by a frame number."),
                ("NOTE","Notes","Distribute frames to notes based on Scene Frame Rate.")
                ]
    )

    distribute_frames: FloatProperty(
        name= "Frames",
        step=100.0,
        default=2.0,
        description= "Number of frames/notes by which to distribute",
        soft_min = 1.0, soft_max = 100.0
    )

    snap: BoolProperty(
        name= "Snap to Frame",
        default=True,
        description= "Snap keyframes to nearest frame after distribution.",
    )

    bpm: FloatProperty(
        name= "BPM",
        default= 120.0,
        description= "Beats per minute",
        soft_min = 1.0, soft_max = 300.0
    )

    note_type: EnumProperty(
        name= "Notes",
        default="1/4",
        description= "Selection to use for repeating",
        items= [("1/1","1/1","Distribute to whole notes."),
                ("1/2","1/2","Distribute to half notes."),
                ("1/4","1/4","Distribute to quarter notes."),
                ("1/8","1/8","Distribute to eighth notes."),
                ("1/16","1/16","Distribute to sixteenth notes."),
                ("1/32","1/32","Distribute to thirth-second notes.")
                ]
    )

    @classmethod
    def poll(self,context):
        return context.selected_visible_fcurves

    def distribute(self,context,keyframes):
        if self.bpm <= 0:
            return

        global debug_mode
        if debug_mode:
            print("DISTRIBUTE MODE: ",self.distribute_mode)

        first_keyframe = 1

        if self.distribute_mode == "FRAME":
            if self.snap:
                snap(keyframes[0])
            first_keyframe = keyframes[0].co[0]
            if debug_mode:
                print("first_keyframe",first_keyframe)
        elif self.distribute_mode == "NOTE":

            if self.selection_mode in ("SELECTED_KEYS","PLAYBACK_RANGE"):
                first_keyframe = keyframes[0].co[0]
                if debug_mode:
                    print("first_keyframe",first_keyframe)

            #frames per second
            fps = context.scene.render.fps
            #frames per minute
            fpm = fps * 60
            if self.bpm:
                quarter_note = fpm/self.bpm
            else:
                quarter_note = 0                
            if debug_mode:
                print("fps",fps,"fpm",fpm,"quarter_note",quarter_note)

        amt = 0

        for i,keyframe in enumerate(keyframes):
            if i == 0:
                continue

            if self.distribute_mode == "FRAME":
                amt = first_keyframe + (i*self.distribute_frames)
            elif self.distribute_mode == "NOTE":
                if self.note_type == "1/1":
                    amt = first_keyframe + i*(quarter_note*4)
                elif self.note_type == "1/2":
                    amt = first_keyframe + i*(quarter_note*2)
                elif self.note_type == "1/4":
                    amt = first_keyframe + i*(quarter_note)
                elif self.note_type == "1/8":
                    amt = first_keyframe + i*(quarter_note*0.5)
                elif self.note_type == "1/16":
                    amt = first_keyframe + i*(quarter_note*0.25)
                elif self.note_type == "1/32":
                    amt = first_keyframe + i*(quarter_note*0.125)

            if debug_mode:
                print("self.note_type",self.note_type,"amt",amt)

            handle_left_frame_offset = keyframe.co[0] - keyframe.handle_left[0]
            handle_right_frame_offset = keyframe.co[0] - keyframe.handle_right[0]

            keyframe.co[0] = amt
            keyframe.handle_left[0] = amt - handle_left_frame_offset
            keyframe.handle_right[0] = amt - handle_right_frame_offset
            if debug_mode:
                print("keyframe.co[0]",keyframe.co[0],"keyframe.handle_left[0]",
                    keyframe.handle_left[0],"keyframe.handle_right[0]",
                    keyframe.handle_right[0])

            if self.snap:
                snap(keyframe,False)

    def invoke(self, context, event):
        fcurves = context.selected_visible_fcurves
        avg_dist = []

        for fcurve in fcurves:
            frames = [k.co[0] for k in fcurve.keyframe_points]
            diffs = [j-i for i, j in zip(frames[:-1], frames[1:])]
            if len(diffs):
                avg_dist.append(sum(diffs)/len(diffs))

        if len(avg_dist):
            self.distribute_frames = round(sum(avg_dist)/len(avg_dist))
        if len(frames) < 2:
            self.report({"WARNING"},"Please select F-Curves with more than one keyframe before running Distribute.")
        
        return self.execute(context)

    def execute(self, context):
        global prefs,debug_mode
        prefs = get_prefs()
        if prefs:
            debug_mode = prefs.debug_mode
        else:
            debug_mode = True

        fcurves = context.selected_editable_fcurves[:]
        for fcurve in fcurves:
            selection = get_selected_keys(fcurve,self.selection_mode)
            self.distribute(context,selection)

        update_fcurves()
        refresh()
        return {'FINISHED'}

def register():
    bpy.utils.register_class(GRAPHKIT_OT_distribute)

def unregister():
    bpy.utils.unregister_class(GRAPHKIT_OT_distribute)
