import bpy
import numpy as np
from bpy.types import Operator
from bpy.props import (IntProperty,BoolProperty,EnumProperty)
from ..utils.utils import (refresh,get_selected_keys,
    get_prefs,update_fcurves,snap,get_keyframe_at_frame)

prefs = None
debug_mode = False

class GRAPHKIT_OT_divide(Operator):
    bl_idname = "graph.divide"
    bl_label = "Divide"
    bl_options = {'REGISTER', 'UNDO'}

    selection_mode: EnumProperty(
        name= "Target",
        default="SELECTED_KEYS",
        description= "Selection to use for repeating",
        items= [
            ("SELECTED_FCURVE","Selected F-Curves","Distribute all frames of selected F-Curves.","IPO_EASE_IN_OUT",0),
            ("SELECTED_KEYS","Selected Keyframes","Distribute selected keyframes.","KEYFRAME_HLT",1)
        ]
    )

    divisions: IntProperty(name="Divisions", default=2, min=1)
    hold_frames: IntProperty(name="Hold Duration", default=4, min=0)

    snap: BoolProperty(
        name= "Snap to Frame",
        default=True,
        description= "Snap keyframes to nearest frame after division.",
    )

    @staticmethod
    def delete_keyframes_between_min_max_selected(fcurve,min_keyframe,max_keyframe):
        keyframes_to_remove = [
            keyframe
            for keyframe in fcurve.keyframe_points
            if min_keyframe.co[0] < keyframe.co[0] < max_keyframe.co[0]
        ]

        for keyframe in keyframes_to_remove:
            fcurve.keyframe_points.remove(keyframe)

    def duplicate_with_hold_frames(self,data):
        """
        Duplicates each tuple in a list except the first and last, adding hold_frames to their first element.

        Args:
        data: A list of tuples.
        hold_frames: The number of frames to hold for each duplicate.

        Returns:
        A new list of tuples with duplicates and hold frames added.
        """

        new_data = []

        for i, item in enumerate(data):
            if i == 0:
                new_data.append(item)
            elif i > 0 and i < len(data) - 1:
                new_data.append((item[0] + (self.hold_frames*(i-1)), item[1]))
                new_data.append((item[0] + (self.hold_frames*i), item[1]))
            else:
                new_data.append((item[0] + (self.hold_frames*self.divisions), item[1]))

        return new_data

    @classmethod
    def poll(cls, context):
        return context.selected_visible_fcurves

    def execute(self, context):
        global prefs,debug_mode
        prefs = get_prefs()
        if prefs:
            debug_mode = prefs.debug_mode
        else:
            debug_mode = True

        fcurve_count = 0

        for fcurve in context.selected_visible_fcurves:
            selected_keyframes = get_selected_keys(fcurve,self.selection_mode)
            if len(selected_keyframes) < 2:
                continue

            # Sort keyframes by frame number
            sorted_keyframes = sorted(selected_keyframes, key=lambda k: k.co.x)
            self.delete_keyframes_between_min_max_selected(fcurve,sorted_keyframes[0],sorted_keyframes[1])

            selected_keyframes = get_selected_keys(fcurve,self.selection_mode)
            sorted_keyframes = sorted(selected_keyframes, key=lambda k: k.co.x)
            min_keyframe = sorted_keyframes[0]
            min_interpolation = min_keyframe.interpolation
            max_keyframe = sorted_keyframes[-1]

            # Extract start and end points
            start_time = min_keyframe.co[0]
            start_value = min_keyframe.co[1]
            end_time = max_keyframe.co[0]
            end_value = max_keyframe.co[1]

            # Calculate the step size for time and value
            time_step = (end_time - start_time) / (self.divisions + 1)
            new_keyframes = [(start_time,start_value)]
            total_hold_time = 0

            # Inside the execute function
            for i in range(1, self.divisions + 1):
                interpolated_time = start_time + i * time_step + total_hold_time
                interpolated_value = fcurve.evaluate(interpolated_time)
                new_keyframes.append((interpolated_time, interpolated_value))

            new_keyframes.append((end_time,end_value))

            new_keyframes = self.duplicate_with_hold_frames(new_keyframes)

            if debug_mode:
                for k in new_keyframes:
                    print(f"i: {new_keyframes.index(k)}, frame: {k[0]} value: {k[1]}")

            # Remove the original end keyframe and add a new one at the adjusted time
            fcurve.keyframe_points.remove(get_keyframe_at_frame(fcurve, end_time))

            keyframes_added = False

            for i,keyframe in enumerate(new_keyframes):
                if i == 0:
                    continue
                frame = keyframe[0]
                value = keyframe[1]
                new_keyframe = fcurve.keyframe_points.insert(frame, value, options={'FAST'})
                new_keyframe.interpolation = min_interpolation
                keyframes_added = True
                if self.snap:
                    snap(new_keyframe,False)

            if keyframes_added:
                fcurve_count += 1

        update_fcurves()
        refresh()
        self.report({"INFO"},f"{fcurve_count * (self.divisions * 2)} keyframes added.")
        return {'FINISHED'}

def register():
    bpy.utils.register_class(GRAPHKIT_OT_divide)

def unregister():
    bpy.utils.unregister_class(GRAPHKIT_OT_divide)