import bpy
import math
from bpy.types import Operator
from ..utils.utils import (
    get_all_selected_keys,
    get_prefs,
    get_linear_y_value,
    line_length
)
from ..utils.eases import *
from bpy.props import (
    BoolProperty,
    EnumProperty,
    IntProperty,
    FloatProperty
)

prefs = None
debug_mode = False

class GRAPHKIT_OT_fade(Operator):
    """Fade selected keyframes or specified keyframe count in or out"""
    bl_idname = "graph.fade"
    bl_label = "Fade"
    bl_options = {"REGISTER", "UNDO"}

    eases = [
        ("", "Linear", ""),
        ("linear", "Linear", "Linear ease in", "IPO_LINEAR", 0),
        ("", "Ease In", ""),
        ("easeInSine", "Sine In", "Sine ease in", "IPO_SINE", 1),
        ("easeInQuad", "Quad In", "Quad ease in", "IPO_QUAD", 2),
        ("easeInCubic", "Cubic In", "Cubic ease in", "IPO_CUBIC", 3),
        ("easeInQuart", "Quart In", "Quart ease in", "IPO_QUART", 4),
        ("easeInQuint", "Quint In", "Quint ease in", "IPO_QUINT", 5),
        ("easeInCirc", "Circ In", "Circ ease in", "IPO_CIRC", 6),
        ("easeInExpo", "Expo In", "Expo ease in", "IPO_EXPO", 7),
        ("", "Ease Out", ""),
        ("easeOutSine", "Sine Out", "Sine ease out", "IPO_SINE", 8),
        ("easeOutQuad", "Quad Out", "Quad ease out", "IPO_QUAD", 9),
        ("easeOutCubic", "Cubic Out", "Cubic ease out", "IPO_CUBIC", 10),
        ("easeOutQuart", "Quart Out", "Quart ease out", "IPO_QUART", 11),
        ("easeOutQuint", "Quint Out", "Quint ease out", "IPO_QUINT", 12),
        ("easeOutCirc", "Circ Out", "Circ ease out", "IPO_CIRC", 13),
        ("easeOutExpo", "Expo Out", "Expo ease out", "IPO_EXPO", 14),
    ]

    subframes: BoolProperty(
        name="Subframes",
        description="Fade keyframes on frames not equal to whole numbers.",
        default=False
    )

    fade_handles: BoolProperty(
        name="Fade Handles",
        description="Fade handles with Keyframes.",
        default=True
    )

    fade_type: EnumProperty(
        name="Fade",
        default="out",
        description="Specify the Fade type.",
        items=[
            ("in", "In", "Fade keyframes in."),
            ("out", "Out", "Fade keyframes out."),
            ("both", "Both", "Fade keyframes in and out."),
        ]
    )

    fade_target: EnumProperty(
        name = "Target",
        description = "Fade target for specified keyframes",
        default = "CENTER",
        items= [
            ("CENTER","Center","Fade keyframes to the average value of the F-Curve."),
            ("ZERO","Zero","Fade keyframes to zero."),
            ("TREND","Trend Line","Fade keyframes to the trend line between the first and last keyframe."),
        ]
    )
    fade_mode: EnumProperty(
        name= "Mode",
        default="keycount",
        description= "Repeating mode to use before the first keyframe.",
        items= [
            ("keycount","Key Count","Specify number of keyframes to fade in."),
            ("selected","Selected","Fade in selected keyframes"),
        ]
    )
    fade_key_count: IntProperty(
        name= "Key Count",
        default=5,
        description= "Number of keyframes to fade in when in Key Count mode.",
        soft_min = 0, soft_max = 50
    )
    fade_x_amount: FloatProperty(
        name = "X Amt",
        description = "Speed at which the F-Curve loses its timing.",
        default = 0.0,
        soft_min = 0, soft_max = 100
    )
    fade_y_amount: FloatProperty(
        name = "Y Amt",
        description = "Speed at which the F-Curve loses its power.",
        default = 100,
        soft_min = 0, soft_max = 100
    )
    fade_ease_mode: EnumProperty(
        name= "Ease",
        default="linear",
        description= "Ease function to use for fading.",
        items= eases
    )
    fade_ease_axis: EnumProperty(
        name= "Ease Axis",
        default="both",
        description= "Axis on which to ease the fade.",
        items= [
            ("x","X","Ease the X axis."),
            ("y","Y","Ease the Y axis."),
            ("both","Both","Ease both axes."),
        ]
    )


    def fade_x(self,type,x1,x2,y1,y2,keyframes):
        d = [x2-x1,y2-y1]
        if type == "in":
            keyframes.reverse()
            d = [x1-x2,y1-y2]

        first_x1 = keyframes[0].co[0]
        first_y1 = keyframes[0].co[1]

        global debug_mode
        if debug_mode:
            print("first_x1",first_x1,"first_y1",first_y1)

        #https://codepen.io/blastframe/pen/RwjYpEb
        for keyframe in keyframes:
            fade_x_amount = self.fade_x_amount/100

            cx = keyframe.co[0] - first_x1
            ratio = 0
            if d[0] != 0:
                ratio = cx / d[0]

            x = keyframe.co[0]
            y = keyframe.co[1]
            handle_left_frame_offset = keyframe.co[0] - keyframe.handle_left[0]
            handle_left_value_offset = keyframe.co[1] - keyframe.handle_left[1]
            handle_right_frame_offset = keyframe.co[0] - keyframe.handle_right[0]
            handle_right_value_offset = keyframe.co[1] - keyframe.handle_right[1]

            ease = get_ease_amt("linear",ratio)

            if self.fade_ease_axis == "x" or self.fade_ease_axis == "both":
                ease = get_ease_amt(self.fade_ease_mode,ratio)

            if debug_mode:
                print("ease",ease)

            # don't scale y, but _reposition_ it along the trend line instead.

            x_target = first_x1 + ease * d[0] * (1-fade_x_amount)
            t = first_y1 + ratio * d[1]
            h = y - t
            y_target = first_y1 + h + ease * d[1] * (1-fade_x_amount)

            diff_x = x_target-x
            diff_y = y_target-y

            ease_amt_x = (x+(diff_x*fade_x_amount))
            ease_amt_y = (y+(diff_y*fade_x_amount))

            if self.subframes == False:
                ease_amt_x = round(ease_amt_x)

            keyframe.co[0] = ease_amt_x
            keyframe.co[1] = ease_amt_y
            handle_fade = 1

            if self.fade_handles:
                handle_fade = 1-(fade_x_amount * ease)

            keyframe.handle_left[0] = ease_amt_x - handle_left_frame_offset*handle_fade
            keyframe.handle_left[1] = ease_amt_y - handle_left_value_offset
            keyframe.handle_right[0] = ease_amt_x - handle_right_frame_offset*handle_fade
            keyframe.handle_right[1] = ease_amt_y - handle_right_value_offset

            if debug_mode:
                print("keyframe.co[0]",keyframe.co[0],"keyframe.co[1]",keyframe.co[1],
                "keyframe.handle_left[0]",keyframe.handle_left[0],
                "keyframe.handle_left[1]",keyframe.handle_left[1],
                "keyframe.handle_right[0]",keyframe.handle_right[0],
                "keyframe.handle_right[1]",keyframe.handle_right[1])

    def fade_y(self,type,x1,x2,y1,y2,keyframes):
        global debug_mode
        d = [x2-x1,y2-y1]

        if type == "in":
            keyframes.reverse()

        avg = sum([keyframe.co[1] for keyframe in keyframes])/len(keyframes)
        if debug_mode:
            if self.fade_target == "CENTER":
                print("Average Y Value:",avg)

        for keyframe in keyframes:
            fade_y_amount = self.fade_y_amount/100

            x = keyframe.co[0]
            y = keyframe.co[1]

            first = keyframes[0].co
            last = keyframes[-1].co

            handle_left_value_offset = keyframe.co[1] - keyframe.handle_left[1]
            handle_right_value_offset = keyframe.co[1] - keyframe.handle_right[1]

            if self.fade_target == "CENTER":
                y_target = get_linear_y_value(x1,avg,x2,avg,x)
            elif self.fade_target == "TREND":
                y_target = get_linear_y_value(x1,y1,x2,y2,x)
            else:
                y_target = get_linear_y_value(x1,0,x2,0,x)                

            distance = line_length(first, last)
            y_dist = line_length(first, [x, y_target])
            perc = 0
            if distance != 0:
                perc = y_dist / distance

            difference = y_target - y
            amt_y = y + difference * perc

            ease_amt_y = amt_y
            ease = 0

            if self.fade_ease_axis == "y" or self.fade_ease_axis == "both":
                ease = get_ease_amt(self.fade_ease_mode,perc) * fade_y_amount
                if debug_mode:
                    print("ease",ease)
                difference_ease = (y_target - y) * ease
                ease_amt_y = amt_y + difference_ease

                linear_sign = math.copysign(1, amt_y-y_target)
                ease_sign = math.copysign(1, ease_amt_y-y_target)

                if linear_sign!=ease_sign:
                  ease_amt_y = y_target

                ease_amt_y = y + difference * ease


            handle_fade = 1
            if self.fade_handles:
                handle_fade *= (1-ease)
                #handle_fade = fade_y_amount * 1-ease

            keyframe.co[1] = ease_amt_y
            keyframe.handle_left[1] = ease_amt_y - handle_left_value_offset*handle_fade
            keyframe.handle_right[1] = ease_amt_y - handle_right_value_offset*handle_fade

            if debug_mode:
                print("keyframe.co[0]",keyframe.co[0],"keyframe.co[1]",keyframe.co[1],
                "keyframe.handle_left[1]",keyframe.handle_left[1],
                "keyframe.handle_right[1]",keyframe.handle_right[1])

            """
            e2 = [x-x1,y-y1]
            dot_p = dot(d,e2)
            len2 = d[0] * d[0] + d[1] * d[1]

            v = [x1 + (dot_p*d[0])/ len2,y1 + (dot_p*d[1])/ len2]
            """

    def fade(self,fcurve):
        global prefs, debug_mode

        if self.fade_mode == "selected":
            keyframes = get_all_selected_keys()
        else:
            keyframes = fcurve.keyframe_points

        x1 = keyframes[0].co[0]
        x2 = keyframes[-1].co[0]
        y1 = keyframes[0].co[1]
        y2 = keyframes[-1].co[1]

        if debug_mode:
            print("*"*40)
            print(f"{fcurve.data_path}[{fcurve.array_index}]")
            print("x1",x1,"x2",x2,"y1",y1,"y2",y2)

        key_length = len(keyframes)-1
        if debug_mode:
            print("key_length",key_length,"self.fade_type",self.fade_type)

        if self.fade_type == "in":
            if self.fade_x_amount != 0:
                if self.fade_mode == "keycount" and self.fade_key_count > 1:
                    self.fade_x("in",x1,x2,y1,y2,keyframes[:self.fade_key_count])
                elif self.fade_mode == "selected":
                    self.fade_x("in",x1,x2,y1,y2,keyframes)
            if self.fade_y_amount != 0:
                if self.fade_mode == "keycount" and self.fade_key_count > 1:
                    self.fade_y("in",x1,x2,y1,y2,keyframes[:self.fade_key_count])
                elif self.fade_mode == "selected":
                    self.fade_y("in",x1,x2,y1,y2,keyframes)
        elif self.fade_type == "out":
            if self.fade_x_amount != 0:
                if self.fade_mode == "keycount" and self.fade_key_count > 1:
                    self.fade_x("out",x1,x2,y1,y2,keyframes[-self.fade_key_count:])
                elif self.fade_mode == "selected":
                    self.fade_x("out",x1,x2,y1,y2,keyframes)
            if self.fade_y_amount != 0:
                if self.fade_mode == "keycount" and self.fade_key_count > 1:
                    self.fade_y("out",x1,x2,y1,y2,keyframes[-self.fade_key_count:])
                elif self.fade_mode == "selected":
                    self.fade_y("out",x1,x2,y1,y2,keyframes)
        elif self.fade_type == "both":
            if self.fade_x_amount != 0:
                if self.fade_mode == "keycount" and self.fade_key_count > 1:
                    if self.fade_key_count > len(keyframes)/2:
                        self.fade_key_count = int(round(len(keyframes)/2))
                    self.fade_x("in",x1,x2,y1,y2,keyframes[:self.fade_key_count])
                    self.fade_x("out",x1,x2,y1,y2,keyframes[-self.fade_key_count:])
                elif self.fade_mode == "selected":
                    if self.fade_key_count > len(keyframes)/2:
                        self.fade_key_count = int(round(len(keyframes)/2))
                    self.fade_x("in",x1,x2,y1,y2,keyframes)
                    self.fade_x("out",x1,x2,y1,y2,keyframes)
            if self.fade_y_amount != 0:
                if self.fade_mode == "keycount" and self.fade_key_count > 1:
                    if self.fade_key_count > len(keyframes)/2:
                        self.fade_key_count = int(round(len(keyframes)/2))
                    self.fade_y("in",x1,x2,y1,y2,keyframes[:self.fade_key_count])
                    self.fade_y("out",x1,x2,y1,y2,keyframes[-self.fade_key_count:])
                elif self.fade_mode == "selected":
                    if self.fade_key_count > len(keyframes)/2:
                        self.fade_key_count = int(round(len(keyframes)/2))
                    self.fade_y("in",x1,x2,y1,y2,keyframes)
                    self.fade_y("out",x1,x2,y1,y2,keyframes)

    @classmethod
    def poll(self,context):
        return context.selected_visible_fcurves

    def execute(self, context):
        global prefs,debug_mode
        prefs = get_prefs()
        debug_mode = prefs.debug_mode

        fcurves = context.selected_editable_fcurves[:]
        for fcurve in fcurves:
            self.fade(fcurve)
        if self.fade_x_amount != 0:
            bpy.ops.graph.clean(channels=False)
        return {'FINISHED'}

def register():
    bpy.utils.register_class(GRAPHKIT_OT_fade)

def unregister():
    bpy.utils.unregister_class(GRAPHKIT_OT_fade)