import bpy
from bpy.types import Operator
from mathutils import Vector
import copy
from bpy.props import (IntProperty,BoolProperty,EnumProperty)
from ..utils.utils import (refresh,get_selected_keys,
    get_prefs,update_fcurves,snap,get_keyframe_at_frame)

prefs = None
debug_mode = False

class GRAPHKIT_OT_hold(Operator):
    bl_idname = "graph.hold"
    bl_label = "Hold"
    bl_options = {'REGISTER', 'UNDO'}

    selection_mode: EnumProperty(
        name= "Target",
        default="SELECTED_FCURVE",
        description= "Selection to use for flattening holds.",
        items= [("SELECTED_FCURVE","Selected F-Curves","Holds all keyframes of selected F-Curves.","IPO_EASE_IN_OUT",0),
                ("SELECTED_KEYS","Selected Keyframes","Holds selected keyframes.","KEYFRAME_HLT",1),
                ("PLAYBACK_RANGE", "Playback Range", "Holds keyframes in the Playback Range.", "TIME", 2)
                ]
    )

    hold_frames: IntProperty(
        name="Hold Frames",
        description="Number of frames to hold",
        default=10,
        min=0,
        max=100
    )

    @classmethod
    def poll(self,context):
        return context.selected_visible_fcurves

    def execute(self, context):
        fcurves = context.selected_editable_fcurves[:]

        global prefs,debug_mode
        prefs = get_prefs()
        debug_mode = prefs.debug_mode

        for fi,fcurve in enumerate(fcurves):
            selected_keys = get_selected_keys(fcurve,self.selection_mode)[:]

            # Creating a list of the original keyframes data
            original_keyframes = [{
                    "co":Vector((pt.co.x,pt.co.y)),
                    "handle_left":Vector((pt.handle_left.x,pt.handle_left.y)),
                    "handle_right":Vector((pt.handle_right.x,pt.handle_right.y)),
                    "interpolation" : pt.interpolation,
                    "amplitude" : pt.amplitude,
                    "back" : pt.back,
                    "easing" : pt.easing,
                    "period" : pt.period,
                    "type" : pt.type,
                    "handle_left_type" : pt.handle_left_type,
                    "handle_right_type" : pt.handle_right_type,
                    "select_control_point" : pt.select_control_point,
                    "select_left_handle" : pt.select_left_handle,
                    "select_right_handle" : pt.select_right_handle,
                } for pt in selected_keys]

            # Create a combined list with deep copies for duplication
            combined_keyframes = []
            for o in original_keyframes:
                combined_keyframes.append(o)
                duplicated = copy.deepcopy(o)
                combined_keyframes.append(duplicated)

            combined_keyframes.sort(key=lambda k: k["co"].x)

            if self.hold_frames > 0:
                for i,k in enumerate(combined_keyframes):
                    if i % 2 == 1:
                        k["handle_left_type"] = "VECTOR"
                    else:
                        k["handle_right_type"] = "VECTOR"

            # Apply the hold_frames * i distribution
            for i, o in enumerate(combined_keyframes):
                o["co"].x += self.hold_frames * i
                o["handle_left"].x += self.hold_frames * i
                o["handle_right"].x += self.hold_frames * i

            # Clear existing keyframes and add new ones
            for sk in reversed(selected_keys):
                fcurve.keyframe_points.remove(sk)
            
            for o in combined_keyframes:
                keyframe = fcurve.keyframe_points.insert(o["co"].x, o["co"].y)
                keyframe.handle_left = o["handle_left"]
                keyframe.handle_right = o["handle_right"]
                keyframe.interpolation = o["interpolation"]
                keyframe.amplitude = o["amplitude"]
                keyframe.back = o["back"]
                keyframe.easing = o["easing"]
                keyframe.period = o["period"]
                keyframe.type = o["type"]
                keyframe.handle_left_type = o["handle_left_type"]
                keyframe.handle_right_type = o["handle_right_type"]
                keyframe.select_control_point = o["select_control_point"]
                keyframe.select_left_handle = o["select_left_handle"]
                keyframe.select_right_handle = o["select_right_handle"]
 
            fcurve.update()

        return {'FINISHED'}

classes = (GRAPHKIT_OT_hold,)

def register():
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)

def unregister():
    from bpy.utils import unregister_class
    for cls in reversed(classes):
        unregister_class(cls)