import bpy
from bpy.types import Operator
from ..data_classes.animation_data import (FCurve, Keyframe)
from bpy.props import (EnumProperty, FloatProperty, BoolProperty)
from ..utils.utils import (refresh, get_selected_keys, snap,
                           convert_to_aligned, get_prefs, get_fcurve_unselected_keys)


class GRAPHKIT_OT_overlap(Operator):
    """Loops the selected F-Curves keyframe range"""
    bl_idname = "graph.overlap"
    bl_label = "Overlap"
    bl_options = {"REGISTER", "UNDO"}

    selection_mode: EnumProperty(
        name="Target",
        default="SELECTED_FCURVE",
        description="Selection mode to use for Overlap",
        items=[
            ("SELECTED_FCURVE", "Selected F-Curves", "Overlaps all frames of selected F-Curves", "IPO_EASE_IN_OUT", 0),
            ("SELECTED_KEYS", "Selected Keyframes", "Overlaps selected keyframes.", "KEYFRAME_HLT", 1),
            ("PLAYBACK_RANGE", "Playback Range", "Overlaps the keyframes in the Playback Range.", "TIME", 2),
        ]
    )

    overlap_mode: EnumProperty(
        name="Overlap Mode",
        default="FCURVE",
        description="Overlap mode for either disabling, overlapping selected objects, or selected F-Curves",
        items=[
            ("NONE", "None", "No overlap.", "RADIOBUT_OFF", 0),
            ("OBJECT", "Object Selection Order", "Overlaps based on object selection order.", "OBJECT_DATA", 1),
            ("FCURVE", "F-Curve Selection Order", "Overlaps based on F-Curve selection order.", "IPO_EASE_IN_OUT", 2),
            ("NAME", "Name Order", "Overlaps based on object or pose bone name order.", "SORTALPHA", 3),
        ]
    )

    frame_overlap: FloatProperty(
        name="Frame Overlap",
        default=1.0,
        description="Amount by which to overlap the selected F-Curves.",
        soft_min=-500, soft_max=500
    )

    snap: BoolProperty(
        name="Snap to Frame",
        default=False,
        description="Snap keyframes to nearest frame after overlap.",
    )

    loop: BoolProperty(
        name="Loop",
        default=False,
        description="Loop selection when overlapping.",
    )

    keep_loop_length: BoolProperty(
        name="Original Loop Length",
        default=True,
        description="Remove keyframes added outside of original keyframe range.",
    )

    def __init__(self):
        self.prefs = get_prefs()
        self.debug_mode = self.prefs.debug_mode

    def copy_key(self, kf: Keyframe, frame: float, value: float, hl: list[float], hr: list[float]) -> Keyframe:
        """Creates a copy of a keyframe with adjusted parameters."""
        new_key = Keyframe([frame, value], hl, hr)
        for attr in ('interpolation', 'amplitude', 'back', 'easing', 'period', 'type',
                     'handle_left_type', 'handle_right_type', 'select_control_point',
                     'select_left_handle', 'select_right_handle'):
            setattr(new_key, attr, getattr(kf, attr))
        return new_key

    def build_loops(self, fcurve):
        """Builds two identical loops on either side of the F-Curve based on selection."""
        new_fcurve = FCurve(fcurve.data_path, fcurve.array_index, [], action=fcurve.id_data.name)
        selection = get_selected_keys(fcurve, self.selection_mode)
        if not selection:
            return None

        start_frame = selection[0].co[0]
        end_frame = selection[-1].co[0]
        frame_shift = end_frame - start_frame

        for before in (True, False):  # Handle before and after loop creation
            loop = [self.copy_key(kf,
                                  kf.co[0] - frame_shift if before else kf.co[0] + frame_shift,
                                  kf.co[1],
                                  [kf.handle_left[0] - frame_shift if before else kf.handle_left[0] + frame_shift, kf.handle_left[1]],
                                  [kf.handle_right[0] - frame_shift if before else kf.handle_right[0] + frame_shift, kf.handle_right[1]])
                    for kf in selection]
            if before:
                loop.reverse()
            new_fcurve.keys.append(loop)

        return new_fcurve

    def overlap_fcurve(self, fcurve, loop_length=0, index=None):
        """Overlaps F-Curve keyframes by the specified amount."""
        amount = self.frame_overlap if index is None else self.frame_overlap * index

        if loop_length and self.overlap_mode not in {"OBJECT", "FCURVE"}:
            amount %= loop_length

        if self.debug_mode:
            print(f"fcurve {fcurve.data_path}[{fcurve.array_index}] amount {amount}")

        selection = get_selected_keys(fcurve, self.selection_mode)
        for keyframe in selection:
            keyframe.co[0] += amount
            keyframe.handle_left[0] += amount
            keyframe.handle_right[0] += amount
            if self.snap:
                snap(keyframe)

    def process_fcurves(self, fcurves, object_index=1):
        """Processes all F-Curves for overlapping and looping."""
        new_fcurves = []
        for i, fcurve in enumerate(fcurves):
            if len(fcurve.keyframe_points) < 3:
                if self.loop:
                    self.report({"WARNING"}, "Looping an F-Curve requires at least 3 keyframes.")
                # Overlap even if there are less than 3 keyframes
                self.overlap_fcurve(fcurve, index=i if self.overlap_mode == "FCURVE" else object_index)
                continue

            if self.loop:
                new_fcurve = self.build_loops(fcurve)
                if new_fcurve:
                    new_fcurves.append((i, new_fcurve))
            else:
                self.overlap_fcurve(fcurve, index=i if self.overlap_mode == "FCURVE" else object_index)

        if self.loop:
            self.apply_loops_and_overlap(new_fcurves, object_index)

    def apply_loops_and_overlap(self, new_fcurves, object_index):
        """Applies the loops and overlaps to the new F-Curves."""
        for j, (i, new_fcurve) in enumerate(new_fcurves):
            fcurve = new_fcurve.get_fcurve()
            selection = get_selected_keys(fcurve, self.selection_mode)
            if not selection:
                continue

            first_keyframe = selection[0].co[0]
            last_keyframe = selection[-1].co[0]
            loop_length = last_keyframe - first_keyframe

            self.insert_new_keys(new_fcurve, fcurve)
            self.overlap_fcurve(fcurve, loop_length, index=j if self.overlap_mode == "FCURVE" else object_index)

            if self.keep_loop_length:
                self.trim_keyframes(fcurve, first_keyframe, last_keyframe)

    def insert_new_keys(self, new_fcurve, fcurve):
        """Inserts new keys from the loops into the F-Curve."""
        for loop in new_fcurve.keys:
            for keyframe in loop:
                new_key = fcurve.keyframe_points.insert(keyframe.co[0], keyframe.co[1], options={"FAST"})
                self.copy_keyframe_data(new_key, keyframe)

    def copy_keyframe_data(self, new_key, keyframe):
        """Copies all necessary keyframe data from old to new keyframe."""
        for attr in ('interpolation', 'amplitude', 'back', 'easing', 'period', 'type', 'handle_left_type', 'handle_right_type'):
            setattr(new_key, attr, getattr(keyframe, attr))

        new_key.handle_left = keyframe.handle_left.copy()
        new_key.handle_right = keyframe.handle_right.copy()
        new_key.select_control_point = keyframe.select_control_point
        new_key.select_left_handle = keyframe.select_left_handle
        new_key.select_right_handle = keyframe.select_right_handle

    def trim_keyframes(self, fcurve, first_keyframe, last_keyframe):
        """Trims keyframes outside the original loop length."""
        frames = {keyframe.co[0] for keyframe in fcurve.keyframe_points}
        if first_keyframe not in frames:
            fcurve.keyframe_points.insert(first_keyframe, fcurve.evaluate(first_keyframe), options={"FAST"})
        if last_keyframe not in frames:
            fcurve.keyframe_points.insert(last_keyframe, fcurve.evaluate(last_keyframe), options={"FAST"})

        selected_keys = set(get_selected_keys(fcurve, self.selection_mode))
        unselected_frames = {kf.co[0] for kf in fcurve.keyframe_points if kf not in selected_keys}

        for keyframe in reversed(fcurve.keyframe_points):
            if (keyframe.co[0] < first_keyframe or keyframe.co[0] > last_keyframe) and keyframe.co[0] not in unselected_frames:
                fcurve.keyframe_points.remove(keyframe, fast=True)

        self.fix_handles(fcurve)

    def fix_handles(self, fcurve):
        """Converts handles of keyframes to aligned if needed."""
        bad_types = {"AUTO", "AUTO_CLAMPED"}
        if fcurve.keyframe_points:
            if fcurve.keyframe_points[0].handle_right_type in bad_types:
                convert_to_aligned(fcurve.keyframe_points[0])
            if fcurve.keyframe_points[-1].handle_left_type in bad_types:
                convert_to_aligned(fcurve.keyframe_points[-1])
        fcurve.update()

    def get_all_fcurves(self, context, o):
        """Fetches all F-Curves associated with the given object or pose bone."""
        fcurves = []
        obj = bpy.data.objects.get(o["name"]) if o["type"] == bpy.types.Object else context.object

        if obj and obj.animation_data and obj.animation_data.action and obj.animation_data.action.fcurves:
            fcurves.extend(obj.animation_data.action.fcurves)

        for mat_slot in getattr(obj, 'material_slots', []):
            material = mat_slot.material
            if material and material.node_tree and material.node_tree.animation_data and material.node_tree.animation_data.action:
                fcurves.extend(material.node_tree.animation_data.action.fcurves)

        return fcurves

    @classmethod
    def poll(cls, context):
        return bool(context.visible_fcurves and context.selected_visible_fcurves)

    def execute(self, context):
        bpy.ops.ed.undo_push(message="Overlap")
        o_fcurves = bpy.ordered_fcurves
        fcurves = [o_fcurve.get_fcurve() for o_fcurve in o_fcurves if o_fcurve.get_fcurve()]
        objects = bpy.ordered_objects

        if self.overlap_mode == "NAME":
            objects.sort(key=lambda x: x['name'])

        if self.overlap_mode == "OBJECT":
            # Process each object and its associated F-Curves for overlapping
            for index, obj in enumerate(objects):
                _fcurves = self.get_all_fcurves(context, obj)
                if _fcurves:
                    self.process_fcurves(_fcurves, object_index=index)
        elif self.overlap_mode == "NAME":
            # Overlap based on sorted name order
            for index, obj in enumerate(objects):
                _fcurves = self.get_all_fcurves(context, obj)
                if _fcurves:
                    self.process_fcurves(_fcurves, object_index=index)
        else:
            self.process_fcurves(fcurves)

        refresh()
        return {'FINISHED'}


def register():
    bpy.utils.register_class(GRAPHKIT_OT_overlap)


def unregister():
    bpy.utils.unregister_class(GRAPHKIT_OT_overlap)
