import bpy
from bpy.types import Operator
from bpy.props import EnumProperty,BoolProperty
from ..utils.utils import (refresh, align_handles, match_handle_lengths, get_selected_keys, get_prefs)

prefs = None
debug_mode = False
class GRAPHKIT_OT_perfect_loop(Operator):
    """Creates seamless loops for the selected F-Curves"""
    bl_idname = "graph.perfect_loop"
    bl_label = "Perfect Loop"
    bl_options = { "REGISTER", "UNDO" }

    selection_mode: EnumProperty(
        name= "Target",
        default="SELECTED_FCURVE",
        description= "Mode for creating the perfect loop with options "\
            "for entire F-Curve or selected keyframes",
        items= [
            ("SELECTED_FCURVE","Selected F-Curves","Loops all frames of selected F-Curves","IPO_EASE_IN_OUT",0),
            ("SELECTED_KEYS","Selected Keyframes","Loops selected keyframes.","KEYFRAME_HLT",1),
            ("PLAYBACK_RANGE", "Playback Range", "Loops the keyframes in the Playback Range.", "TIME", 2),
        ]
    )

    match_mode: EnumProperty(
        name= "Match",
        default="MATCH_START",
        description= "Keyframe to use for creating a loop",
        items= [("MATCH_START","Start","Uses the F-Curve's first keyframe's values to create a loop."),
                ("MATCH_END","End","Uses the F-Curve's last keyframe's values to create a loop.")
                ]
    )

    match_angles: BoolProperty(
        name= "Match Angles",
        default=True,
        description= "Adjust start and end keyframes' keyframe angles for smooth transition.",
    )

    match_lengths: BoolProperty(
        name= "Match Lengths",
        default=True,
        description= "Match start and end keyframes' keyframe lengths.",
    )

    @classmethod
    def poll(self,context):
        return context.selected_visible_fcurves

                
    def loop_keys(self, key1, key2):
        if self.match_mode == "MATCH_START":
            offset_left_y = key2.handle_left[1] - key2.co[1]
            offset_right_y = key2.handle_right[1] - key2.co[1]
            key2.co[1] = key1.co[1]
            key2.handle_left[1] = key2.co[1] + offset_left_y
            key2.handle_right[1] = key2.co[1] + offset_right_y
            if self.match_angles:
                align_handles(self.match_mode, key1, key2)
            if self.match_lengths:
                match_handle_lengths(key1, key2)
        elif self.match_mode == "MATCH_END":
            offset_left_y = key1.handle_left[1] - key1.co[1]
            offset_right_y = key1.handle_right[1] - key1.co[1]
            key1.co[1] = key2.co[1]
            key1.handle_left[1] = key1.co[1] + offset_left_y
            key1.handle_right[1] = key1.co[1] + offset_right_y
            if self.match_angles:
                align_handles(self.match_mode, key2, key1)
            if self.match_lengths:
                match_handle_lengths(key2, key1)

    def execute(self, context):
        fcurves = context.selected_editable_fcurves[:]
        global prefs,debug_mode
        prefs = get_prefs()
        debug_mode = True

        for fcurve in fcurves:
            selected_keys = get_selected_keys(fcurve,self.selection_mode)
            if debug_mode:
                print("first_key",selected_keys[0],
                "last_key",selected_keys[-1])
            self.loop_keys(key1 = selected_keys[0], key2 = selected_keys[-1])
        
        for fcurve in fcurves:
            fcurve.update()

        refresh()
        return {'FINISHED'}

def register():
    bpy.utils.register_class(GRAPHKIT_OT_perfect_loop)

def unregister():
    bpy.utils.unregister_class(GRAPHKIT_OT_perfect_loop)