import bpy, os, random
from bpy.types import Operator
from ..utils.utils import (get_selected_keys, 
    get_prefs, is_rotation, convert_to_auto)
from bpy.props import (BoolProperty,
                       FloatProperty,
                       IntProperty,
                       EnumProperty,)

prefs = None
debug_mode = False
class GRAPHKIT_OT_randomize(Operator):
    """Randomize selected keyframes' frames or values by a specified amount"""
    bl_idname = "graph.randomize"
    bl_label = "Randomize"
    bl_options = { "REGISTER", "UNDO" }

    selection_mode: EnumProperty(
        name= "Target",
        default="SELECTED_FCURVE",
        description= "Mode for assigning random values to keyframes.",
        items= [
            ("SELECTED_FCURVE","Selected F-Curves","Randomizes all keyframes of selected F-Curves.","IPO_EASE_IN_OUT",0),
            ("SELECTED_KEYS","Selected Keyframes","Randomizes selected keyframes.","KEYFRAME_HLT",1),
            ("PLAYBACK_RANGE", "Playback Range", "Randomizes the keyframes in the Playback Range.", "TIME", 2),
        ]
    )

    axis: EnumProperty(
        name= "Axis",
        default="Y_AXIS",
        description= "Randomize method",
        items= [
            ("X_AXIS","X","Randomize keyframes by frames."),
            ("Y_AXIS","Y","Randomize keyframes by values."),
            ("BOTH","Both","Randomize keyframes by both frames and values."),
        ]
    )

    randomize_by_fcurves: BoolProperty(
        name = "Randomize by F-Curves",
        description = "Applies random values to keyframes based on random F-Curve reed.",
        default = True
    )

    absolute: BoolProperty(
        name = "Absolute",
        description = "Use absolute values.",
        default = False
        )

    snap: BoolProperty(
        name = "Snap to Frame",
        description = "Snap randomized frames to whole frame values.",
        default = True
     )

    keep_sign: BoolProperty(
        name = "Keep Sign",
        description = "Prevent keyframe values from changing from positive"\
        "to negative or vice versa.",
        default = True
        )

    min: FloatProperty(
        name = "Min",
        description = "Minimum amount for randomized keyframes.",
        default = -10.0,
        soft_min = -1000, soft_max = 1000)

    max: FloatProperty(
        name = "Max",
        description = "Maximum amount for randomized keyframes.",
        default = 10.0,
        soft_min = -1000, soft_max = 1000)

    amount: FloatProperty(
        name = "Amount",
        description = "Amount of randomization to apply.",
        default = 5.0,
        soft_min = 0, soft_max = 300)

    random_seed: IntProperty(
        name = "Random Seed",
        description = "Random seed for random amounts.",
        default=0,
        soft_min = 0, soft_max = 1000)

    exclude_start_and_end: BoolProperty(
        name = "Exclude Start & End",
        description = "Excludes the start & end frame from randomization "\
            "in the case of looping F-Curve animations.",
        default = False
     )

    automatic: BoolProperty(
        name = "Convert to Automatic",
        description = "Converts selected keyframes' handles to automatic handles.",
        default = False
     )

    def get_curve_value_limits(self,fcurve):
        values = [ keyframe.co[1] for keyframe in fcurve.keyframe_points ]
        return min(values),max(values)

    def get_curve_frame_limits(self,fcurve):
        frames = [ keyframe.co[0] for keyframe in fcurve.keyframe_points ]
        return min(frames),max(frames)

    @classmethod
    def poll(self,context):
        return context.selected_visible_fcurves

    def execute(self, context):
        rotation_dps = ["rotation_euler","rotation_quaternion",
        "delta_rotation_quaternion","delta_rotation_euler"]

        fcurves = context.selected_editable_fcurves[:]

        global prefs,debug_mode
        prefs = get_prefs()
        debug_mode = prefs.debug_mode
        all_rotations = is_rotation(context)

        for fi,fcurve in enumerate(fcurves):
            selected_keys = get_selected_keys(fcurve,self.selection_mode,
                exclude_limits=self.exclude_start_and_end)
            #min_value,max_value = self.get_curve_value_limits(fcurve)
            #min_frame,max_frame = self.get_curve_frame_limits(fcurve)

            if debug_mode:
                print(f"F-Curve: {fcurve.data_path}[{fcurve.array_index}]")
                print(f"min_value: {self.min}, max_value: {self.max}")

            for i, sk in enumerate(selected_keys):
                fcurve_seed = 0
                if self.randomize_by_fcurves:
                    fcurve_seed = fi
                random.seed(self.random_seed*len(selected_keys) + i + fcurve_seed)
                x_amt = self.amount
                y_amt = self.amount/100
                if not all_rotations:
                    for rotation_data_path in rotation_dps:
                        if fcurve.data_path.endswith(rotation_data_path):
                            y_amt = self.amount/10000
                            break

                x_val = random.uniform(self.min, self.max) * x_amt
                y_val = random.uniform(self.min, self.max) * y_amt

                if debug_mode:
                    print("self.axis",self.axis,"x_val",x_val,"y_val",y_val)

                if self.axis == "X_AXIS" or self.axis == "BOTH":
                    if self.absolute:
                        sk.co[0] = x_val
                        sk.handle_left[0] = x_val
                        sk.handle_right[0] = x_val
                    else:
                        sk.co[0] += x_val
                        sk.handle_left[0] += x_val
                        sk.handle_right[0] += x_val
                    if self.snap == True:
                        sk.co[0] = round(sk.co[0])
                        sk.handle_left[0] = round(sk.handle_left[0])
                        sk.handle_right[0] = round(sk.handle_right[0])

                if self.axis == "Y_AXIS" or self.axis == "BOTH":
                    if (self.keep_sign):
                        sign = lambda a: (a>0) - (a<0)
                        key_sign = sign(sk.co[1])
                        sign_val = sign(y_val)
                        if key_sign != sign_val:
                            y_val *= -1

                    if self.absolute:
                        sk.co[1] = y_val
                        sk.handle_left[1] = y_val
                        sk.handle_right[1] = y_val
                    else:
                        sk.co[1] += y_val
                        sk.handle_left[1] += y_val
                        sk.handle_right[1] += y_val

        """
        if self.axis == "X_AXIS" or self.axis == "BOTH":
            bpy.ops.graph.clean(channels=False)
            if debug_mode:
                print("bpy.ops.graph.clean(channels=False) called.")
        """

        if self.automatic == True:
            convert_to_auto(fcurves,self.exclude_start_and_end)
            if debug_mode:
                print("convert_to_auto called.")


        return {'FINISHED'}

def register():
    bpy.utils.register_class(GRAPHKIT_OT_randomize)

def unregister():
    bpy.utils.unregister_class(GRAPHKIT_OT_randomize)
