import bpy
from bpy.types import Operator
from bpy.props import IntProperty, BoolProperty
from collections import Counter
from ..utils.utils import get_prefs

prefs = None
debug_mode = False
class GRAPHKIT_OT_range_to_loop(Operator):
    """Sets frame range based on loop length"""
    bl_idname = "graph.range_to_loop"
    bl_label = "Range to Loop"
    bl_options = { "REGISTER", "UNDO" }

    start_frame: IntProperty(
        name= "Start Frame",
        default= 1,
        description= "Start frame for looping animation",
        soft_min = 0, soft_max = 5000
    )

    end_frame: IntProperty(
        name= "End Frame",
        default= 1,
        options={'HIDDEN'},
        description= "End frame for looping animation",
    )

    loop_length: IntProperty(
        name= "Loop Length",
        default= 24,
        description= "Loop length for setting frame range",
        soft_min = 1, soft_max = 500,
    )

    loop_count: IntProperty(
        name= "Loop Count",
        default= 1,
        description= "Loop count for setting frame range",
        soft_min = 1, soft_max = 500,
    )

    preview_range: BoolProperty(
        name= "Preview Range",
        default= False,
        description= "Set Preview Frame Range to Loop"
    )

    scene_range: BoolProperty(
        name= "Scene Range",
        default= True,
        description = "Sets Scene Frame Range to Loop"
    )

    @staticmethod
    def most_frequent(List,debug_mode=False):
        occurence_count = Counter(List)
        return occurence_count.most_common(1)[0]
    
    @classmethod
    def poll(self,context):
        return context.visible_fcurves

    def get_loop(self,fcurves):
        global debug_mode

        first_frames = list()
        last_frames = list()

        for fcurve in fcurves:
            if debug_mode:
                print(f"F-Curve: {fcurve.data_path}[{fcurve.array_index}]")
            if len(fcurve.keyframe_points) <= 1:
                continue
            first_keyframe = fcurve.keyframe_points[0]
            fcurve_length = len(fcurve.keyframe_points)-1
            last_keyframe = fcurve.keyframe_points[fcurve_length]
            if first_keyframe.co[0] != last_keyframe.co[0]:
                first_frames.append(int(round(first_keyframe.co[0])))
                last_frames.append(int(round(last_keyframe.co[0])))
            if debug_mode:
                print("first_keyframe",first_keyframe.co[0],
                    "last_keyframe",last_keyframe.co[0],
                    "fcurve_length",fcurve_length)


        most_common_first = None
        most_common_last = None
        loop_length = None

        if len(first_frames):
            most_common_first = GRAPHKIT_OT_range_to_loop.most_frequent(first_frames,debug_mode)
        if len(last_frames):
            most_common_last = GRAPHKIT_OT_range_to_loop.most_frequent(last_frames,debug_mode)

        first = min(first_frames)
        last = max(last_frames)

        if most_common_last and most_common_first:

            if most_common_first[1] > 1:
                first = most_common_first[0]
            if most_common_last[1] > 1:
                last = most_common_last[0]

        loop_length = last-first

        for fcurve in fcurves:
            if len(fcurve.keyframe_points) <= 1:
                continue
            for keyframe in fcurve.keyframe_points:
                if keyframe.co[0] < first:
                    keyframe.select_control_point = False
                elif keyframe.co[0] > last:
                    keyframe.select_control_point = False

        if debug_mode:
            print("first",first,
                "last",last,
                "loop_length",loop_length)

        return first, last, loop_length

    def invoke(self, context, event):
        fcurves = context.visible_fcurves[:]
        if len(fcurves) == 0:
            return self.execute(context)

        most_common_first, most_common_last, loop_length = self.get_loop(fcurves)

        if most_common_first != None and loop_length != None:
            self.start_frame = most_common_first
            self.end_frame = most_common_last
            self.loop_length = loop_length

        return self.execute(context)

    def execute(self, context):
        global prefs,debug_mode
        prefs = get_prefs()
        debug_mode = prefs.debug_mode

        scene_data = bpy.data.scenes[context.scene.name]

        scene_data.use_preview_range = self.preview_range
        if self.preview_range:
            scene_data.frame_preview_start = self.start_frame
            scene_data.frame_preview_end = self.start_frame + (self.loop_length*self.loop_count) - 1
        if self.scene_range:
            scene_data.frame_start = self.start_frame
            scene_data.frame_end = self.start_frame + (self.loop_length*self.loop_count) - 1
        return {'FINISHED'}


def register():
    bpy.utils.register_class(GRAPHKIT_OT_range_to_loop)

def unregister():
    bpy.utils.unregister_class(GRAPHKIT_OT_range_to_loop)