import bpy
from bpy.types import Operator
from ..data_classes.animation_data import (Loop)
from ..utils.utils import (refresh, get_prefs,
    copy_key, get_keyframe_at_frame)

prefs = None
debug_mode = False
class GRAPHKIT_OT_repeat(Operator):
    """Loops the selected F-Curves keyframe range"""
    bl_idname = "graph.repeat"
    bl_label = "Repeat"
    bl_options = { "REGISTER", "UNDO" }

    selection_mode: bpy.props.EnumProperty(
        name= "Target",
        default="SELECTED_FCURVE",
        description= "Repeating mode to use before the first keyframe",
        items= [
            ("SELECTED_FCURVE","Selected F-Curves","Repeats all frames of selected F-Curves.","IPO_EASE_IN_OUT",0),
            ("SELECTED_KEYS","Selected Keyframes","Repeats selected keyframes.","KEYFRAME_HLT",1),
            ("PLAYBACK_RANGE", "Playback Range", "Repeats the keyframes in the Playback Range.", "TIME", 2)
        ]
    )

    before_mode: bpy.props.EnumProperty(
        name= "Before Mode",
        default="REPEAT_MOTION",
        description= "Repeating mode to use before the first keyframe",
        items= [("REPEAT_MOTION","Repeat Motion","Repeats keyframe range as-is."),
                ("REPEAT_OFFSET","Repeat with Offset","Repeats keyframe range "\
                "with continued rate of change from start & end keyframe."),
                ("REPEAT_MIRROR","Repeat Mirrored","Alternate between forward and "\
                "reverse playback of keyframe range."),
            ]
    )

    before_count: bpy.props.IntProperty(
        name = "Count",
        description = "Number of times to repeat the keyframe range",
        default = 0,
        soft_min = 0, soft_max = 100
    )

    after_mode: bpy.props.EnumProperty(
        name= "After Mode",
        default="REPEAT_MOTION",
        description= "Repeating mode to use before the first keyframe",
        items= [("REPEAT_MOTION","Repeat Motion","Repeats keyframe range as-is."),
                ("REPEAT_OFFSET","Repeat with Offset","Repeats keyframe range "\
                "with continued rate of change from start & end keyframe."),
                ("REPEAT_MIRROR","Repeat Mirrored","Alternate between forward and "\
                "reverse playback of keyframe range."),
            ]
    )

    after_count: bpy.props.IntProperty(
        name = "Count",
        description = "Number of times to repeat the keyframe range",
        default = 1,
        soft_min = 0, soft_max = 100
    )

    @classmethod
    def poll(self,context):
        return context.selected_visible_fcurves

    def build_loop(self,loop,mode,shift_frames_amt,shift_values_amt):
        new_loop = None
        new_loop = loop.duplicate()
        new_loop.shift_frames(shift_frames_amt)
        if mode == "REPEAT_OFFSET":
            new_loop.shift_values(shift_values_amt)
        return new_loop

    def add_loops(self,loop,before,after):
        before_loops = []
        after_loops = []

        for bc in range(0,before["count"]):
            shift_frames_amt = -loop.frame_length+(bc*-loop.frame_length)
            shift_values_amt = -loop.value_change+(bc*-loop.value_change)
            new_loop = None
            new_loop = self.build_loop(loop,before["mode"],shift_frames_amt,shift_values_amt)
            if before["mode"] == "REPEAT_MIRROR":
                if (bc % 2) == 0:
                    new_loop = new_loop.mirror()
            before_loops.append(new_loop)

        for ac in range(0,after["count"]):
            #shift_frames_amt = loop.frame_length+(ac*loop.frame_length)
            shift_frames_amt = (ac*loop.frame_length)
            shift_values_amt = loop.value_change+(ac*loop.value_change)
            new_loop = self.build_loop(loop,after["mode"],shift_frames_amt,shift_values_amt)
            if after["mode"] == "REPEAT_MIRROR":
                if (ac % 2) == 0:
                    new_loop = new_loop.mirror()
                    new_loop.shift_frames(loop.frame_length)
            after_loops.append(new_loop)
        
        return before_loops,after_loops

    def add_key(self,fcurve,keyframe,direction,loop_len=None,mirrored=False,mode=None):
        add_new_key = True
        if direction == "BEFORE":
            target_index = loop_len-1
            if mode == "REPEAT_MIRROR":
                if mirrored == True:
                    target_index = 0
                else:
                    target_index = loop_len-1

            if keyframe.index == target_index:
                old_key = get_keyframe_at_frame(fcurve,keyframe.co[0])
                if old_key:
                    old_key.handle_left_type = keyframe.handle_left_type
                    old_key.handle_left[0] = keyframe.handle_left[0]
                    old_key.handle_left[1] = keyframe.handle_left[1]
                    add_new_key = False
        elif direction == "AFTER":
            target_index = 0
            if mirrored == True:
                target_index = loop_len-1
            if keyframe.index == target_index:
                old_key = get_keyframe_at_frame(fcurve,keyframe.co[0])
                if old_key:
                    old_key.handle_right_type = keyframe.handle_right_type
                    old_key.handle_right[0] = keyframe.handle_right[0]
                    old_key.handle_right[1] = keyframe.handle_right[1]
                    add_new_key = False

        if add_new_key:
            # fcurve.keyframe_points.add(1)
            # new_key = fcurve.keyframe_points[-1]
            # new_key.co[0] = keyframe.co[0]
            # new_key.co[0] = keyframe.co[1]
            new_key = fcurve.keyframe_points.insert(keyframe.co[0], keyframe.co[1], options={"FAST"})
            new_key.interpolation = keyframe.interpolation
            new_key.amplitude = keyframe.amplitude
            new_key.back = keyframe.back
            new_key.easing = keyframe.easing
            new_key.period = keyframe.period
            new_key.type = keyframe.type
            new_key.handle_left_type = keyframe.handle_left_type
            new_key.handle_right_type = keyframe.handle_right_type
            new_key.handle_left[0] = keyframe.handle_left[0]
            new_key.handle_left[1] = keyframe.handle_left[1]
            new_key.handle_right[0] = keyframe.handle_right[0]
            new_key.handle_right[1] = keyframe.handle_right[1]

    def invoke(self, context, event):
        fcurves = context.selected_editable_fcurves[:]
        if len(fcurves) == 0:
            return self.execute(context)

        for fcurve in fcurves:
            if len(fcurve.keyframe_points) == 2 and \
                ((self.before_mode == "REPEAT_MOTION" and self.before_count > 0) or
                (self.after_mode == "REPEAT_MOTION" and self.after_count > 0)):
                self.before_mode = "REPEAT_MIRROR"
                self.after_mode = "REPEAT_MIRROR"
                return self.execute(context)

        return self.execute(context)

    def execute(self, context):
        fcurves = context.selected_editable_fcurves[:]

        global prefs,debug_mode
        prefs = get_prefs()
        debug_mode = prefs.debug_mode

        for fcurve in fcurves:
            if debug_mode:
                print("*" * 50, f"{fcurve.data_path}[{fcurve.array_index}]", "*" * 50)

            original_loop = []
            original_loop = Loop(fcurve=fcurve,selection_mode=self.selection_mode)

            if len(original_loop.keys) == 2:
                if (self.before_mode=="REPEAT_MOTION" and self.before_count > 0) or \
                    (self.after_mode=="REPEAT_MOTION" and self.after_count > 0):
                    self.report({"WARNING"}, f"For best results, 'Repeat Motion' should be "
                         f"used with looping animations. {fcurve.data_path}[{fcurve.array_index}] does not loop.")

            if debug_mode:
                print("#"*20,"Original Loop","#"*20)
                print(original_loop)
                print("#"*55)

            before_loops = []
            after_loops = []

            before_loops,after_loops = self.add_loops(original_loop,
                                                {"mode":self.before_mode,"count":self.before_count},
                                                {"mode":self.after_mode,"count":self.after_count})

            if debug_mode:
                print("*"*10,"before","*"*10)
                print("len(before_loops)",len(before_loops))
            for x in before_loops:

                if x.mirrored:
                    x.shift_frames(x.frame_length)

                for k in x.keys:
                    self.add_key(fcurve,k,"BEFORE",len(x.keys),
                        mirrored=x.mirrored,mode=self.before_mode)
                if debug_mode:
                    print(x)

            if debug_mode:
                print("*"*10,"after","*"*10)
                print("len(after_loops)",len(after_loops))
            for x in after_loops:
                x.shift_frames(x.frame_length)
                for k in x.keys:
                    self.add_key(fcurve,k,"AFTER",len(x.keys),
                        mirrored=x.mirrored,mode=self.after_mode)
                if debug_mode:
                    print(x)

        refresh()
        return {'FINISHED'}

def register():
    bpy.utils.register_class(GRAPHKIT_OT_repeat)

def unregister():
    bpy.utils.unregister_class(GRAPHKIT_OT_repeat)