import bpy
import random
from bpy.props import EnumProperty, FloatProperty, BoolProperty, IntProperty
from bpy.types import Operator, PoseBone
from ..utils.utils import get_prefs, get_object_for_fcurve, get_object

def is_type(obj, obj_type):
    if obj_type == "ALL":
        return True
    return obj.type == obj_type

obj_types = [
            ("ALL", "All", "All object types", "BLENDER", 0),
            ("MESH", "Meshes", "Mesh objects", "MESH_DATA", 1),
            ("CURVE", "Curves", "Curve objects", "CURVE_DATA", 2),
            ("SURFACE", "Surfaces", "Surface objects", "SURFACE_DATA", 3),
            ("META", "Metaballs", "Metaball objects", "META_DATA", 4),
            ("FONT", "Text", "Text objects", "FONT_DATA", 5),
            ("POINTCLOUD", "Point Clouds", "Point Cloud objects", "POINTCLOUD_DATA", 6),
            ("FIELD", "Force Fields", "Force Field objects", "OUTLINER_OB_FORCE_FIELD", 7),
            ("VOLUME", "Volumes", "Volume objects", "VOLUME_DATA", 8),
            ("GPENCIL", "Grease Pencil", "Grease Pencil objects", "GREASEPENCIL", 9),
            None,
            ("ARMATURE", "Armatures", "Armature objects", "ARMATURE_DATA", 10),
            ("LATTICE", "Lattices", "Lattice objects", "LATTICE_DATA", 11),
            None,
            ("EMPTY", "Empties", "Empty objects", "EMPTY_DATA", 12),
            None,
            ("LIGHT", "Lights", "Light objects", "LIGHT_DATA", 13),
            None,
            ("CAMERA", "Cameras", "Camera objects", "CAMERA_DATA", 14),
            None,
            ("SPEAKER", "Speakers", "Speaker objects", "SPEAKER", 15),
            None,
            ("MATERIAL", "Materials", "Material objects", "MATERIAL", 16)
]

def matches_side_convention(name, suffixes):
    for suffix in suffixes:
        if name.endswith(suffix) or name.startswith(suffix):
            return True
    return False

def select_fcurves(action, select_settings, scene, is_armature=False, side_suffixes=None):
    select_count = 0
    for fcurve in action.fcurves:
        # Determine whether to select this F-Curve based on the provided conditions
        should_select = False

        if is_armature and "pose.bones" in fcurve.data_path:
            bone_name = fcurve.data_path.split('"')[1]  # Extract bone name from data_path
            should_select = side_suffixes is None or matches_side_convention(bone_name, side_suffixes)
        elif not is_armature:
            # For non-armature objects, check the action name against side_suffixes
            object_name = get_object_for_fcurve(fcurve, scene)
            if object_name and (side_suffixes is None or matches_side_convention(object_name, side_suffixes)):
                should_select = True

        # If the F-Curve should be selected, apply the selection settings
        if should_select:
            fcurve.select = select_settings['select_fcurve']
            select_count += 1
            for keyframe in fcurve.keyframe_points:
                keyframe.select_control_point = select_settings['select_keyframe']
                keyframe.select_left_handle = select_settings['select_left_handle']
                keyframe.select_right_handle = select_settings['select_right_handle']
    return select_count

class GRAPHKIT_OT_select_by_type(Operator):
    """Select F-Curves by Object Type"""
    bl_idname = "graph.select_by_type"
    bl_label = "Select By Type"
    bl_options = {'REGISTER', 'UNDO'}

    selection_type: EnumProperty(
        name="Selection",
        default="NEW",
        items=[
            ("NEW","New","Deselect all F-Curves and create a new selection by type"),
            ("ADD","Add","Add selection by type to existing selection")
        ]
    )

    object_type: EnumProperty(
        name="Object Type",
        items=obj_types
    )

    select_fcurve: BoolProperty(
        name="F-Curves",
        description="Select the F-Curve",
        default=True
    )

    select_keyframe: BoolProperty(
        name="Keyframes",
        description="Select the Keyframe",
        default=True
    )

    select_left_handle: BoolProperty(
        name="Left Handles",
        description="Select the Keyframe's Left Handle",
        default=True
    )

    select_right_handle: BoolProperty(
        name="Right Handles",
        description="Select the Keyframe's Right Handle",
        default=True
    )

    @classmethod
    def poll(self,context):
        return context.visible_fcurves

    def execute(self, context):
        if self.selection_type == "NEW":
            bpy.ops.graph.select_all(action='DESELECT')

        obj_type = self.object_type

        select_settings = {
            'select_fcurve': self.select_fcurve,
            'select_keyframe': self.select_keyframe,
            'select_left_handle': self.select_left_handle,
            'select_right_handle': self.select_right_handle
        }

        select_count = 0
        # Select F-Curves for all objects if 'ALL' is selected
        if obj_type == "ALL":
            bpy.ops.graph.select_all(action='SELECT')
        elif obj_type == "MATERIAL":
            for mat in bpy.data.materials:
                if mat.node_tree and mat.node_tree.animation_data and mat.node_tree.animation_data.action:
                    select_count += select_fcurves(mat.node_tree.animation_data.action, select_settings, scene = context.scene)
        else:
            for obj in bpy.context.scene.objects:
                if is_type(obj, obj_type) and obj.animation_data and obj.animation_data.action:
                    select_count += select_fcurves(obj.animation_data.action, select_settings, scene = context.scene, is_armature=(obj.type == 'ARMATURE'))

        self.report({"INFO"}, f"{select_count} F-curves selected by type.")
        return {'FINISHED'}

class GRAPHKIT_OT_select_by_side(Operator):
    """Select F-Curves by Side and Object Type"""
    bl_idname = "graph.select_by_side"
    bl_label = "Select By Side"
    bl_options = {'REGISTER', 'UNDO'}

    selection_type: EnumProperty(
        name="Selection",
        default="NEW",
        items=[
            ("NEW","New","Deselect all F-Curves and create a new selection by type"),
            ("ADD","Add","Add selection by type to existing selection")
        ]
    )

    side: EnumProperty(
        name="Side",
        items=[
            ("LEFT", "Left", "Select left side"),
            ("RIGHT", "Right", "Select right side")
        ]
    )

    object_type: EnumProperty(
        name="Object Type",
        items=obj_types
    )

    select_fcurve: BoolProperty(
        name="F-Curves",
        description="Select the F-Curve",
        default=True
    )

    select_keyframe: BoolProperty(
        name="Keyframes",
        description="Select the Keyframe",
        default=True
    )

    select_left_handle: BoolProperty(
        name="Left Handles",
        description="Select the Keyframe's Left Handle",
        default=True
    )

    select_right_handle: BoolProperty(
        name="Right Handles",
        description="Select the Keyframe's Right Handle",
        default=True
    )

    @classmethod
    def poll(self,context):
        return context.visible_fcurves

    def execute(self, context):
        if self.selection_type == "NEW":
            bpy.ops.graph.select_all(action='DESELECT')

        side = self.side
        obj_type = self.object_type

        side_suffixes = {
            "LEFT": ["_L", ".l", "-l", " LEFT", "Left", "L"],
            "RIGHT": ["_R", ".r", "-r", " RIGHT", "Right", "R"]
        }

        select_settings = {
            'select_fcurve': self.select_fcurve,
            'select_keyframe': self.select_keyframe,
            'select_left_handle': self.select_left_handle,
            'select_right_handle': self.select_right_handle
        }

        select_count = 0
        if obj_type == "MATERIAL":
            for mat in bpy.data.materials:
                if mat.node_tree and mat.node_tree.animation_data and mat.node_tree.animation_data.action:
                    select_count += select_fcurves(mat.node_tree.animation_data.action, select_settings, scene = context.scene, side_suffixes=side_suffixes[side])
        else:
            for obj in bpy.context.scene.objects:
                if is_type(obj, obj_type):
                    if obj.animation_data and obj.animation_data.action:
                        select_count += select_fcurves(obj.animation_data.action, select_settings, scene = context.scene, is_armature=(obj.type == 'ARMATURE'), side_suffixes=side_suffixes[side])

        self.report({"INFO"}, f"{select_count} F-curves selected by side.")
        return {'FINISHED'}

class GRAPHKIT_OT_select_random_fcurves(Operator):
    """Select Random F-Curves of visible objects or F-Curves in the active View Layer"""
    bl_idname = "graph.select_random_fcurves"
    bl_label = "Select Random F-Curves"
    bl_options = {'REGISTER', 'UNDO'}

    selection_type: EnumProperty(
        name="Selection",
        default="NEW",
        items=[
            ("NEW","New","Deselect all F-Curves and create a new selection by type"),
            ("ADD","Add","Add selection by type to existing selection")
        ]
    )

    random_modes = [
        ("OBJECTS", "Visible Objects", "Randomize by visible objects", "OBJECT_DATA", 0),
        ("FCURVES", "Visible F-Curves", "Randomize by visible F-curves", "IPO_EASE_IN_OUT", 1),
    ]

    mode: EnumProperty(
        name="Mode",
        description="Select the randomization mode",
        items=random_modes,
        default='OBJECTS'
    )

    object_type: EnumProperty(
        name="Object Type",
        items=obj_types
    )

    percentage: FloatProperty(
        name="Percentage",
        description="Percentage of F-Curves to select",
        default=50.0,
        min=0.0,
        max=100.0
    )

    random_seed: IntProperty(
        name="Random Seed",
        description="Seed value for the random selection",
        default=0
    )

    select_fcurve: BoolProperty(
        name="F-Curves",
        description="Select the F-Curve",
        default=True
    )

    select_keyframe: BoolProperty(
        name="Keyframes",
        description="Select the Keyframe",
        default=True
    )

    select_left_handle: BoolProperty(
        name="Left Handles",
        description="Select the Keyframe's Left Handle",
        default=True
    )

    select_right_handle: BoolProperty(
        name="Right Handles",
        description="Select the Keyframe's Right Handle",
        default=True
    )

    @classmethod
    def poll(self,context):
        return context.visible_fcurves

    def select_random_fcurves(self, fcurves):
        random.shuffle(fcurves)
        num_to_select = int(len(fcurves) * (self.percentage / 100.0))
        selected_fcurves = fcurves[:num_to_select]

        select_count = 0

        # Deselect all F-Curves first
        for fcurve in fcurves:
            fcurve.select = False
            for keyframe in fcurve.keyframe_points:
                keyframe.select_control_point = False
                keyframe.select_left_handle = False
                keyframe.select_right_handle = False

        # Select the chosen F-Curves
        for fcurve in selected_fcurves:
            fcurve.select = self.select_fcurve
            select_count += 1
            for keyframe in fcurve.keyframe_points:
                keyframe.select_control_point = self.select_keyframe
                keyframe.select_left_handle = self.select_left_handle
                keyframe.select_right_handle = self.select_right_handle
        return select_count

    def execute(self, context):
        if context.area.type != 'GRAPH_EDITOR':
            self.report({'WARNING'}, "Graph Editor not active")
            return {'CANCELLED'}

        if self.selection_type == "NEW":
            bpy.ops.graph.select_all(action='DESELECT')

        random.seed(self.random_seed)

        select_count = 0

        if self.object_type == "MATERIAL":
            all_material_fcurves = []
            for mat in bpy.data.materials:
                if mat.animation_data and mat.animation_data.action:
                    all_material_fcurves.extend(mat.animation_data.action.fcurves)
            select_count += self.select_random_fcurves(all_material_fcurves)
        else:
            if self.mode == 'OBJECTS':
                filtered_objects = [obj for obj in context.view_layer.objects if obj.visible_get() and is_type(obj, self.object_type) and obj.animation_data and obj.animation_data.action]
                all_object_fcurves = [fcurve for obj in filtered_objects for fcurve in obj.animation_data.action.fcurves]
                select_count += self.select_random_fcurves(all_object_fcurves)
            elif self.mode == 'FCURVES':
                all_fcurves = []
                for obj in context.view_layer.objects:
                    if obj.visible_get() and is_type(obj, self.object_type) and obj.animation_data and obj.animation_data.action:
                        all_fcurves.extend(obj.animation_data.action.fcurves)
                select_count += self.select_random_fcurves(all_fcurves)

        self.report({"INFO"}, f"{select_count} F-curves selected randomly.")
        return {'FINISHED'}

class GRAPHKIT_OT_select_objects_from_fcurves(Operator):
    """Select objects using selected F-Curves in the Graph Editor"""
    bl_idname = "graph.select_objects_from_fcurves"
    bl_label = "Select Objects/Pose Bones from Selected F-Curves"

    def execute(self, context):
        scene = context.scene
        selected_objects = 0
        selected_pose_bones = 0

        for fcurve in context.selected_editable_fcurves:
            obj = get_object(fcurve, scene)
            if obj:
                if isinstance(obj, PoseBone):
                    obj.bone.select = True
                    selected_pose_bones += 1
                else:
                    obj.select_set(True)
                    selected_objects += 1

        report_message = f"{selected_objects} objects"
        if context.selected_pose_bones and selected_pose_bones:
            report_message += f" and {selected_pose_bones} bones"
        report_message += " selected from F-curves."
        
        self.report({"INFO"}, report_message)
        return {'FINISHED'}


classes = (GRAPHKIT_OT_select_by_type,
            GRAPHKIT_OT_select_by_side,
            GRAPHKIT_OT_select_random_fcurves,
            GRAPHKIT_OT_select_objects_from_fcurves)

def register():
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)

def unregister():
    from bpy.utils import unregister_class
    for cls in reversed(classes):
        unregister_class(cls)

if __name__ == "__main__":
    register()