import bpy
from ..utils.utils import (refresh_ui_keyframes, get_keyframe_at_frame,
    get_prefs, get_object_for_fcurve, get_object)
from bpy.types import (Operator,PropertyGroup)
from bpy.props import (FloatProperty,BoolProperty,PointerProperty,EnumProperty)
debug_mode = False
bpy.tweens = []

class DRAGGABLEPROP_OT_subscribe(Operator):
    bl_idname  = "draggableprop.subscribe"
    bl_label   = ""
    stop: BoolProperty()  # This is used so we don't end up in an infinite loop because we blocked the release event

    def modal(self, context, event):
        prefs = get_prefs()
        global debug_mode
        debug_mode = prefs.debug_mode

        if self.stop:
            context.scene.tween_props.is_dragging = False
            if debug_mode: print("End Drag")
            self.stop = True
            return {'FINISHED'}
        if event.value == 'RELEASE':  # Stop the modal on next frame. Don't block the event since we want to exit the field dragging
            if debug_mode: print("RELEASE")
            context.scene.tween_props.is_dragging = False
            self.stop = True
        return {'PASS_THROUGH'}

    def invoke(self, context, event):
        prefs = get_prefs()
        global debug_mode
        debug_mode = prefs.debug_mode
        fcurves = bpy.context.selected_editable_fcurves
        bpy.tweens = [fcurve.evaluate(bpy.context.scene.frame_current) for fcurve in fcurves]  # Store evaluated values
        if debug_mode:
            print("invoke")
            print(bpy.tweens)
        self.stop = False
        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}

def linear_interpolation(prev_value, next_value, amount):
    return prev_value + (next_value - prev_value) * amount

def find_next_keyframe(fcurve, current_frame):
    next_keyframe = None
    for keyframe in fcurve.keyframe_points:
        if keyframe.co.x > current_frame:
            if next_keyframe is None or keyframe.co.x < next_keyframe.co.x:
                next_keyframe = keyframe
    return next_keyframe

def get_default_value(obj,fcurve):
    if debug_mode:
        print(obj)

    dpath = None
    if obj:
        if obj.rna_type.name == "Pose Bone":
            names = fcurve.data_path.split('"')[1::2]
            no_name = ''.join(fcurve.data_path.split('"')[0::2])
            bone_data = no_name.replace('][', '].[').split('.')

            for id, i in enumerate(bone_data):
                if i.find('[]') > 0:
                    bone_data[id] = i.replace('[]', '["' + names.pop(0) + '"]')
            boneName = bone_data[1].split('"')[1::2]

            if boneName:
                if bone_data[2] != "[]":
                    dpath = obj.bl_rna.properties[bone_data[2]]
        else:
            dpath = obj.bl_rna.properties[fcurve.data_path]

    if dpath != None:
        if dpath.default_array:
            default = dpath.default_array[fcurve.array_index]
        else:
            default = dpath.default
        return default
    else:
        return 0  # Default value if none found

def set_track_values(self,tween_amount):
    context = bpy.context
    scene = context.scene
    fcurves = context.selected_editable_fcurves[:]

    if not len(bpy.tweens):
        bpy.tweens = [fcurve.evaluate(bpy.context.scene.frame_current) for fcurve in fcurves]  # Store evaluated values

    for i,fcurve in enumerate(fcurves):
        keyframes = fcurve.keyframe_points
        key_count = len(keyframes)
        if key_count < 2:
            continue
        current_frame = context.scene.frame_current

        inter = "BEZIER"

        prev_keyframe = None
        for keyframe in keyframes:
            if keyframe.co.x < current_frame:
                prev_keyframe = keyframe
                inter = prev_keyframe.interpolation
            else:
                break

        # Interpolate values
        if prev_keyframe:
            prev_value = prev_keyframe.co.y
            next_keyframe = find_next_keyframe(fcurve, current_frame)

            if next_keyframe:
                next_value = next_keyframe.co.y
                tween_type = context.scene.tween_props.tween_type
                if tween_type == "BREAKDOWN":
                    tween_amount_normalized = (tween_amount + 1) / 2  # Normalize -1 to 1 to 0 to 1
                    new_value = linear_interpolation(prev_value, next_value, tween_amount_normalized)
                elif tween_type == "BLEND_TO_NEIGHBOR":
                    original_value = bpy.tweens[i]
                    if tween_amount < 0:
                        delta = original_value - prev_value
                    else:
                        delta = next_value - original_value
                    new_value = original_value + delta * tween_amount
                elif tween_type == "BLEND_TO_DEFAULT":
                    original_value = bpy.tweens[i]
                    obj = None
                    if fcurve.id_data.id_root == "OBJECT":
                        obj = get_object(fcurve,scene)
                    default_value = get_default_value(obj,fcurve)
                    new_value = original_value + (default_value - original_value) * tween_amount

                # Check if keyframe exists at the current frame
                key = get_keyframe_at_frame(fcurve,current_frame)

                if key:
                    key.co.y = new_value
                else:
                    key = keyframes.insert(current_frame, new_value)
                key.interpolation = inter
                if inter == "BEZIER":
                    key.handle_left_type = "AUTO_CLAMPED"
                    key.handle_right_type = "AUTO_CLAMPED"
                key.type = "BREAKDOWN"

        if tween_amount > 1.0 or tween_amount < -1.0:
            bpy.context.workspace.status_text_set(text=f"{round(tween_amount, 5)} ** Overshooting **")
        else:
            bpy.context.workspace.status_text_set(text=f"{round(tween_amount, 5)}")
        fcurve.update()
    refresh_ui_keyframes()

def update_prop(self, value):
    # global debug_mode

    #if debug_mode: print("self.is_dragging",self.is_dragging)
    if self.is_dragging == True:
        if debug_mode:
            print("Dragging")
        pass
    else:
        if debug_mode:
            print("Drag start")
        self.is_dragging = True
        bpy.ops.draggableprop.subscribe('INVOKE_DEFAULT')

class Tween_Props(PropertyGroup):
    tween_amt: FloatProperty(
        name = "Tween",
        description = "Create a keyframe between two keyframes by a "\
        "percentage of the difference",
        set=set_track_values,
        update=update_prop,
        soft_min = -3, soft_max = 3
    )
    show_presets: BoolProperty(default=False)
    is_dragging: BoolProperty(
        default=False
    )
    tween_type: EnumProperty(
        name = "Type",
        default="BREAKDOWN",
        items = [
            ("BREAKDOWN","Breakdown","Breakdown"),
            ("BLEND_TO_NEIGHBOR","Blend to Neighbor","Blend to Neighbor"),
            ("BLEND_TO_DEFAULT","Blend to Default","Blend to Default"),
        ]
    )

def register():
    bpy.utils.register_class(DRAGGABLEPROP_OT_subscribe)  # Register modal operator
    bpy.utils.register_class(Tween_Props)
    bpy.types.Scene.tween_props = PointerProperty(type=Tween_Props)

def unregister():
    bpy.utils.unregister_class(DRAGGABLEPROP_OT_subscribe)
    del bpy.types.Scene.tween_props
    bpy.utils.unregister_class(Tween_Props)
