import bpy
from bpy.props import (
    BoolProperty,
    FloatProperty,
    StringProperty,
    IntProperty,
    EnumProperty,
    PointerProperty
)
from .icons.custom_icons import get_icon
from .properties.props import GRAPHKIT_PG_graph_editor_props
from .utils.utils import (refresh,
                          get_prefs)
from bpy.utils import (register_class,
                       unregister_class)

def menu_func(self, context):
    layout = self.layout
    layout.separator()
    props = context.scene.graph_props
    row = layout.row(align=True)

    prefs = get_prefs()

    if not props or not hasattr(props,"pin") or (prefs and not prefs.show_toggles):
        return

    if prefs:
        if prefs.show_pin:
            if props.pin:
                row.prop(props,"pin", text="", icon="PINNED")
            else:
                row.prop(props,"pin", text="", icon="UNPINNED") 

        if prefs.show_hide:
            if props.hide:
                row.prop(props,"hide", text="", icon="HIDE_OFF")
            else:
                row.prop(props,"hide", text="", icon="HIDE_ON") 

        if prefs.show_modifiers:
            if props.modifiers:
                row.prop(props,"modifiers", text="", icon="MODIFIER_ON")
            else:
                row.prop(props,"modifiers", text="", icon="MODIFIER_OFF")

        if prefs.show_mute:
            if props.mute:
                row.prop(props,"mute", text="", icon="CHECKBOX_HLT")
            else:
                row.prop(props,"mute", text="", icon="CHECKBOX_DEHLT")

        if prefs.show_lock:
            if props.lock:
                row.prop(props,"lock", text="", icon="LOCKED")
            else:
                row.prop(props,"lock", text="", icon="UNLOCKED")
        
        if prefs.show_refresh:
            row.operator("wm.refresh_graphkit", text="", icon="FILE_REFRESH")

    layout.separator()

def update_toggles(self=None, context=bpy.context):
    classes = (GRAPHKIT_PG_graph_editor_props,)

    for cls in classes:
        try:
            del bpy.types.Scene.graph_props
            bpy.types.GRAPH_MT_editor_menus.remove(menu_func)
            bpy.types.GRAPH_HT_header.remove(menu_func)
            unregister_class(cls)
        except:
            pass
        register_class(cls)
        bpy.types.Scene.graph_props = PointerProperty(type=GRAPHKIT_PG_graph_editor_props)
        if self.show_toggles:
            if self.toggle_location == "MENU_LEFT":
                bpy.types.GRAPH_MT_editor_menus.prepend(menu_func)
            elif self.toggle_location == "MENU_RIGHT":
                bpy.types.GRAPH_MT_editor_menus.append(menu_func)
            elif self.toggle_location == "HEADER_RIGHT":
                bpy.types.GRAPH_HT_header.append(menu_func)

def refresh_screens(self,context):
    refresh()

class GRAPHKIT_preferences(bpy.types.AddonPreferences):
    bl_idname = __package__

    show_toggles: BoolProperty(name="Show Toggles", default=True, update=update_toggles)
    toggle_location: EnumProperty(name="Toggle Location",
                                  default="MENU_LEFT",
                                  items=[
                                      ("MENU_LEFT","Menu: Left","Shows toggle buttons to the Left of the Graph Editor Menu"),
                                      ("MENU_RIGHT","Menu: Right","Shows toggle buttons to the Right of the Graph Editor Menu"),
                                      ("HEADER_RIGHT","Header: Right","Shows toggle buttons on the right of the Graph Editor Header"),
                                      ("GRAPHKIT_PANEL_TOP","GraphKit Panel: Top","Shows toggle buttons in the GraphKit Panel at the top"),
                                      ("GRAPHKIT_PANEL_BOTTOM","GraphKit Panel: Bottom","Shows toggle buttons in the GraphKit Panel at the bottom"),
                                  ],
                                  update=update_toggles
                                  )
    show_pin: BoolProperty(name="Pin", default=True, update=refresh_screens)
    show_hide: BoolProperty(name="Hide", default=True, update=refresh_screens)
    show_modifiers: BoolProperty(name="Modifiers", default=True, update=refresh_screens)
    show_mute: BoolProperty(name="Mute", default=True, update=refresh_screens)
    show_lock: BoolProperty(name="Lock", default=True, update=refresh_screens)
    show_refresh: BoolProperty(name="Refresh Icons", default=True, update=refresh_screens)

    show_alerts: BoolProperty(name="Colorize Tween Overshoots", default=False)
    update: BoolProperty(name="Update", default=False)
    version: StringProperty(name="Version", default="")
    autocheck: BoolProperty(name="Auto-check for Update", default=True)
    debug_mode: BoolProperty(name="Debug Mode", default=False)
    icons_only: BoolProperty(name="Icons Only", default=False)
    scale_y: FloatProperty(name="Button Scale", default=1)
    weeks: IntProperty(
        min = 1,
        max = 52,
        soft_min = 1,
        soft_max = 52,
        name="Frequency (weeks)",
        default=2
    )

    def draw(self, context):
        layout = self.layout
        row = layout.row()
        row.prop(self, "icons_only")
        row.prop(self, "scale_y")
        layout.separator()
        layout.prop(self, "show_toggles")
        if self.show_toggles:
            col = layout.column()
            col.prop(self, "toggle_location")
            box = col.box()
            col = box.column(align=True)
            if self.show_pin:
                col.prop(self, "show_pin", icon="PINNED")
            else:
                col.prop(self, "show_pin", icon="UNPINNED")                

            if self.show_hide:
                col.prop(self, "show_hide", icon="HIDE_OFF")
            else:
                col.prop(self, "show_hide", icon="HIDE_ON")   

            if self.show_modifiers:
                col.prop(self, "show_modifiers", icon="MODIFIER_ON")
            else:
                col.prop(self, "show_modifiers", icon="MODIFIER_OFF")   

            if self.show_mute:
                col.prop(self, "show_mute", icon="CHECKBOX_HLT")
            else:
                col.prop(self, "show_mute", icon="CHECKBOX_DEHLT")   

            if self.show_lock:
                col.prop(self, "show_lock", icon="LOCKED")
            else:
                col.prop(self, "show_lock", icon="UNLOCKED")

            col.prop(self, "show_refresh", text="Refresh Icons", icon="FILE_REFRESH", toggle=True)

        col = layout.column()
        col.prop(self, "show_alerts", icon="ERROR") 

        layout.separator()
        layout.prop(self, "debug_mode")
        layout.prop(self, "autocheck")
        row = layout.row()
        row.enabled = self.autocheck
        box = row.box()
        box.use_property_split = True
        box.use_property_decorate = False
        box.prop(self, "weeks")
        if bpy.app.version >= (4, 2, 0):
            product_name = __package__.split(".")[2]
        else:
            product_name = __package__.split(".")[0]
        layout.operator(f"{product_name.lower()}.check_updates", text="Check For Updates", icon="FILE_REFRESH")
        if self.update:
            layout.alert = True
            layout.label(text=f"Update Available. {str(self.version)}. Please download from the store where you purchased this Add-on.")
        elif self.version != "":
            layout.label(text="You have the latest version.")
        layout.alert = False
        row = layout.row(align=True)
        row.scale_y = 1.3
        row.operator("wm.url_open", text="Blastframe", icon_value=get_icon("Blastframe_Logo_White")).url = "https://blastframe.com"
        row.operator("wm.url_open", text="YouTube", icon_value=get_icon("YouTube")).url = "https://www.youtube.com/@blastframe"
        row.operator("wm.url_open", text="Blender Market", icon_value=get_icon("Blender_Market")).url = "https://blendermarket.com/creators/blastframe"

def register():
    from bpy.utils import register_class, unregister_class
    try:
        unregister_class(GRAPHKIT_preferences)
    except RuntimeError:
        pass
    register_class(GRAPHKIT_preferences)

def unregister():
    try:
        bpy.types.GRAPH_MT_editor_menus.remove(menu_func)
    except ValueError as e:
        print(f"Error removing menu_func from GRAPH_MT_editor_menus: {e}")

    try:
        bpy.types.GRAPH_HT_header.remove(menu_func)
    except ValueError as e:
        print(f"Error removing menu_func from GRAPH_HT_header: {e}")

    from bpy.utils import unregister_class
    try:
        unregister_class(GRAPHKIT_preferences)
    except RuntimeError as e:
        print(f"Error unregistering GRAPHKIT_preferences: {e}")