import bpy, math, os
from pathlib import Path
from ..data_classes.animation_data import (FCurve,Keyframe)

def refresh():
    #Updates all animation related UI
    for area in bpy.context.screen.areas:
        if area.type == 'VIEW_3D':
            for region in area.regions:
                region.tag_redraw()
        elif area.type == 'DOPESHEET_EDITOR':
            for region in area.regions:
                region.tag_redraw()
        elif area.type == 'GRAPH_EDITOR':
            for region in area.regions:
                region.tag_redraw()
        elif area.type == 'NLA_EDITOR':
            for region in area.regions:
                region.tag_redraw()
        elif area.type == 'OUTLINER':
            for region in area.regions:
                region.tag_redraw()
        elif area.type == 'PROPERTIES':
            for region in area.regions:
                region.tag_redraw()

def load_icons():
    preview_collections = {}
    icons = [
        {"name":"amplify_icon","file":"Amplify.png"},
        {"name":"distribute_icon","file":"Distribute.png"},
        {"name":"fade_icon","file":"Fade.png"},
        {"name":"flatten_holds_icon","file":"Flatten_Holds.png"},
        {"name":"randomize_icon","file":"Randomize.png"},
        {"name":"repeat_icon","file":"Repeat.png"}
    ]

    pcoll = bpy.utils.previews.new()
    path = Path(os.path.dirname(__file__))
    parent = path.parent.absolute()
    icon_dir = os.path.join(parent, "icons")

    for icon in icons:
        icon_path = os.path.join(icon_dir, icon["file"])
        pcoll.load(icon["name"], icon_path, 'IMAGE')

    preview_collections["main"] = pcoll
    return preview_collections

def is_rotation(context):
    fcurves = context.selected_editable_fcurves[:]
    data_types = [fcurve.data_path.split('.')[-1] for fcurve in fcurves]
    rotations = ["rotation_euler", "rotation_quaternion", "rotation_axis_angle", 
        "delta_rotation_euler", "delta_rotation_quaternion",]    
    rotation_data_paths = [dp in rotations for dp in data_types]
    if sum(rotation_data_paths) == len(data_types):
        return True

def update_fcurves():
    for fcurve in bpy.context.visible_fcurves[:]:
        for keyframe in fcurve.keyframe_points:
            if keyframe.select_control_point:
                fcurve.update()
                break

def get_keyframe_at_frame(fcurve,frame):
    for k in fcurve.keyframe_points:
        if k.co[0] == frame:
            return k

def get_object_for_fcurve(fcurve, scene):
    for obj in scene.objects:
        if obj.animation_data and obj.animation_data.action == fcurve.id_data:
            return obj.name
    return None

def get_object(fcurve,scene):
    prefs = get_prefs()
    debug_mode = None
    if prefs:
        debug_mode = get_prefs().debug_mode
    else:
        debug_mode = True

    if debug_mode:
        print(f"get_object data_path: {fcurve.data_path}")
    obj_name = get_object_for_fcurve(fcurve, scene)
    if debug_mode:
        print(f"get_object: {obj_name}")
    obj = scene.objects.get(obj_name)
    if obj:
        if obj.type == "ARMATURE" and bpy.context.mode == "POSE":
            data_path = fcurve.data_path
            if data_path.startswith("pose.bones["):
                bone_name = data_path.split('"')[1]
                pose_bone = obj.pose.bones.get(bone_name)
                return pose_bone
            else:
                return obj
        else:
            return obj

def copy_key(keyframe,key_index=None):
    new_key = Keyframe([keyframe.co[0],keyframe.co[1]],
                        [keyframe.handle_left[0],keyframe.handle_left[1]],
                        [keyframe.handle_right[0],keyframe.handle_right[1]],key_index=key_index)
    new_key.interpolation = keyframe.interpolation
    new_key.amplitude = keyframe.amplitude
    new_key.back = keyframe.back
    new_key.easing = keyframe.easing
    new_key.period = keyframe.period
    new_key.type = keyframe.type
    new_key.handle_left_type = keyframe.handle_left_type
    new_key.handle_right_type = keyframe.handle_right_type
    return new_key

def get_prefs():
    # Attempt to determine the add-on name dynamically from available addons
    addon_name = 'graphkit'  # Replace with the unique part of your add-on's name
    addons = bpy.context.preferences.addons
    
    # Loop through all add-ons to find an exact match with the last part of each key
    for key in addons.keys():
        # Split the key by '.' and check the last element
        if key.split('.')[-1].replace(' ', '_').lower() == addon_name.lower():
            return addons[key].preferences  # Return preferences if exact match is found
    
    return None  # Return None if no exact match is found

def get_fcurve_selected_keys(fcurve, no_limits=False):
    selected_keys = []

    for keyframe in fcurve.keyframe_points:
        if keyframe.select_control_point:
            selected_keys.append(keyframe)

    if no_limits == True:
        selected_keys = selected_keys[1:-1]

    return selected_keys

def get_fcurve_unselected_keys(fcurve, no_limits=False):
    unselected_keys = []

    for keyframe in fcurve.keyframe_points:
        if not keyframe.select_control_point:
            unselected_keys.append(keyframe)

    if no_limits == True:
        unselected_keys = unselected_keys[1:-1]

    return unselected_keys

def get_all_selected_keys():
    selected_keys = []

    fcurves = bpy.context.selected_editable_fcurves[:]
    for fcurve in fcurves:
        #print(fcurve.id_data)
        for keyframe in fcurve.keyframe_points:
            if keyframe.select_control_point:
                selected_keys.append(keyframe)

    return selected_keys

def nice_fcurve_name(fcurve):
    axis_name = [ "X", "Y", "Z" ]
    quat_name = [ "W", "X", "Y", "Z" ]
    data_path = fcurve.data_path
    array_index = fcurve.array_index

    if data_path in ["location", "rotation_euler", "rotation_quaternion", "scale"]:
        if "quaternion" in data_path:
            return f"{quat_name[array_index]} {data_path}"
        else:
            return f"{axis_name[array_index]} {data_path}"
    else:
        return data_path

def get_data_fcurve(fcurve):
    return FCurve(fcurve.data_path,fcurve.array_index,action=fcurve.id_data.name)

def get_selected_keys(fcurve,mode,exclude_limits=False):
    selection = list()

    if mode == "SELECTED_FCURVE":
        selection = fcurve.keyframe_points
    elif mode == "SELECTED_KEYS":
        selection = get_fcurve_selected_keys(fcurve)
    elif mode == "PLAYBACK_RANGE":
        context = bpy.context
        scene = context.scene

        # Set the keyframe at scene.frame_start
        value_start = fcurve.evaluate(scene.frame_start)
        fcurve.keyframe_points.insert(scene.frame_start, value_start)

        # Set the keyframe at scene.frame_end
        value_end = fcurve.evaluate(scene.frame_end)
        fcurve.keyframe_points.insert(scene.frame_end, value_end)

        # Iterate over the keyframe points in the F-Curve
        for keyframe in fcurve.keyframe_points:
            # Check if the keyframe is within the playback frame range
            if scene.frame_start <= keyframe.co.x <= scene.frame_end:
                selection.append(keyframe)

    if exclude_limits == True:
        selection = selection[1:-1]

    return selection


def get_keyframe(fcurve,frame):
    found = None
    for keyframe in fcurve.keyframe_points:
        if keyframe.co[0] == frame:
            found = keyframe
            break
    return found

def get_keyframe_index(fcurve,frame):
    foundIndex = None
    for i,keyframe in enumerate(fcurve.keyframe_points):
        if keyframe.co[0] == frame:
            foundIndex = i
            break
    return foundIndex

def refresh_ui_keyframes():
    try:
        for area in bpy.context.screen.areas:
            if area.type in ('TIMELINE', 'GRAPH_EDITOR', 'DOPESHEET_EDITOR'):
                area.tag_redraw()
    except:
        pass

def get_linear_x_value(x1,y1,x2,y2,y_value):
    denom = (x1-x2)
    if denom == 0:
        denom = 0.0001
    line_slope = (y1-y2)/denom
    yintercept = (x1*y2 - x2*y1)/denom
    try:
        frame = (y_value - yintercept)/ line_slope
        return frame
    except ZeroDivisionError:
        return 0

def get_linear_y_value(x1,y1,x2,y2,frame):
    denom = (x1-x2)
    if denom == 0:
        denom = 0.0001
    line_slope = (y1-y2)/denom
    yintercept = (x1*y2 - x2*y1)/denom
    value = line_slope * frame + yintercept
    return value

def convert_to_auto(fcurves,exclude_start_and_end):
    for fcurve in fcurves:
        selected_keys = get_fcurve_selected_keys(fcurve,
            no_limits=exclude_start_and_end)
        for keyframe in selected_keys:
            keyframe.handle_left_type = "AUTO"
            keyframe.handle_right_type = "AUTO"
        fcurve.update()

def convert_to_vector(keyframe):
    keyframe.handle_left_type = "VECTOR"
    keyframe.handle_right_type = "VECTOR"

def convert_to_aligned(keyframe):
    keyframe.handle_left_type = "ALIGNED"
    keyframe.handle_right_type = "ALIGNED"

def magnitude(p):
  return math.sqrt(p[0]**2 + p[1]**2)

def normalize(p):
  m = magnitude(p)
  if (m > 0):
    return [p[0]/m,p[1]/m]

def line_length(pt1,pt2):
    a = pt2[0] - pt1[0]
    b = pt2[1] - pt1[1]
    return math.sqrt(a**2 + b**2)

def dot(pt1, pt2):
    return pt1[0]*pt2[0] + pt1[1]*pt2[1]

def diff(pt1,pt2):
  return [pt1[0]-pt2[0],pt1[1]-pt2[1]]

def sqr(pt):
  return [pt[0]**2,pt[1]**2]

def prod(pt1,pt2):
  return [pt1[0]*pt2[0],pt1[1]*pt2[1]]

def mult(pt,f):
  return [pt[0]*f,pt[1]*f]

def add(pt1,pt2):
  return [pt1[0]+pt2[0],pt1[1]+pt2[1]]

def div(pt,q):
  return [pt[0]/q,pt[1]/q]

def get_handle_length(x1, x2, y1, y2):
    a = x1 - x2
    b = y1 - y2

    c = math.sqrt( a*a + b*b )
    return c

def sub_v2_v2v2(a, b):
    r = [0,0]
    r[0] = a[0] - b[0]
    r[1] = a[1] - b[1]
    return r

def dot_v2v2(a, b):
    return a[0] * b[0] + a[1] * b[1]

def mul_v2_v2fl(r, a, f):
    r[0] = a[0] * f
    r[1] = a[1] * f

def zero_v2(r):
    r[0] = 0.0
    r[1] = 0.0

def normalize_v2_v2_length(r, a, unit_length):
    d = dot_v2v2(a, a)
    if (d > 1.0e-35):
        d = math.sqrt(d)
        mul_v2_v2fl(r, a, unit_length / d)
    else:
        # Either the vector is small or one of it's values contained `nan`.
        zero_v2(r)
        d = 0.0
    return d

def normalize_v2_length(n, unit_length):
    return normalize_v2_v2_length(n, n, unit_length)

def add_v2_v2v2(r, a, b):
    r[0] = a[0] + b[0]
    r[1] = a[1] + b[1]
    return r

def align_handles(match_mode, keyframe1, keyframe2):
    key1_left_handle_direction = normalize(sub_v2_v2v2(keyframe1.handle_left, keyframe1.co))
    key1_right_handle_direction = normalize(sub_v2_v2v2(keyframe1.handle_right, keyframe1.co))

    key1_left_handle_length = get_handle_length(keyframe1.co[0], keyframe1.handle_left[0],
                                                keyframe1.co[1], keyframe1.handle_left[1])
    key1_right_handle_length = get_handle_length(keyframe1.co[0], keyframe1.handle_right[0],
                                                 keyframe1.co[1], keyframe1.handle_right[1])

    key2_left_handle_length = get_handle_length(keyframe2.co[0], keyframe2.handle_left[0],
                                                keyframe2.co[1], keyframe2.handle_left[1])
    key2_right_handle_length = get_handle_length(keyframe2.co[0], keyframe2.handle_right[0],
                                                 keyframe2.co[1], keyframe2.handle_right[1])

    if match_mode == "MATCH_START":
        # matches left handle to right handle
        keyframe1.handle_left = add(keyframe1.co, mult([-key1_right_handle_direction[0], -key1_right_handle_direction[1]], key1_left_handle_length))
        keyframe2.handle_left = add(keyframe2.co, mult([-key1_right_handle_direction[0], -key1_right_handle_direction[1]], key2_left_handle_length))
        keyframe2.handle_right = add(keyframe2.co, mult([key1_right_handle_direction[0], key1_right_handle_direction[1]], key2_right_handle_length))

    elif match_mode == "MATCH_END":
        # matches right handle to left handle
        keyframe1.handle_right = add(keyframe1.co, mult([-key1_left_handle_direction[0], -key1_left_handle_direction[1]], key1_right_handle_length))
        keyframe2.handle_right = add(keyframe2.co, mult([-key1_left_handle_direction[0], -key1_left_handle_direction[1]], key2_right_handle_length))
        keyframe2.handle_left = add(keyframe2.co, mult([key1_left_handle_direction[0], key1_left_handle_direction[1]], key2_left_handle_length))


def match_handle_lengths(keyframe1, keyframe2):
    key1_left_handle_direction = normalize(sub_v2_v2v2(keyframe1.handle_left, keyframe1.co))
    key1_right_handle_direction = normalize(sub_v2_v2v2(keyframe1.handle_right, keyframe1.co))

    key2_left_handle_direction = normalize(sub_v2_v2v2(keyframe2.handle_left, keyframe2.co))
    key2_right_handle_direction = normalize(sub_v2_v2v2(keyframe2.handle_right, keyframe2.co))

    key1_left_handle_length = get_handle_length(keyframe1.co[0], keyframe1.handle_left[0],
                                                keyframe1.co[1], keyframe1.handle_left[1])
    key1_right_handle_length = get_handle_length(keyframe1.co[0], keyframe1.handle_right[0],
                                                 keyframe1.co[1], keyframe1.handle_right[1])

    keyframe2.handle_left = add(keyframe2.co, mult(key2_left_handle_direction, key1_left_handle_length))
    keyframe2.handle_right = add(keyframe2.co, mult(key2_right_handle_direction, key1_right_handle_length))


def snap(keyframe,handles=False):
    original_amt = float(keyframe.co[0])
    keyframe.co[0] = round(keyframe.co[0])
    diff = original_amt-keyframe.co[0]
    keyframe.handle_left[0] -= diff
    keyframe.handle_right[0] -= diff
    if handles == True:
        keyframe.handle_left[0] = round(keyframe.handle_left[0])
        keyframe.handle_right[0] = round(keyframe.handle_right[0])

def get_groups():
    fcurves = bpy.context.visible_fcurves
    groups = []
    if len(fcurves) == 0:
        groups = []
        for action in bpy.data.actions:
            for group in action.groups:
                groups.append(group)
    else:
        for fcurve in fcurves:
            if hasattr(fcurve,"group") and fcurve.group:
                groups.append(fcurve.group)
    return groups

def get_fcurves():
    fcurves = bpy.context.visible_fcurves
    # if len(fcurves) == 0:
    #     fcurves = []
    #     for action in bpy.data.actions:
    #         for fcurve in action.fcurves:
    #             if not fcurve.hide:
    #                 fcurves.append(fcurve)
    return fcurves

def get_modifiers():
    modifiers = []
    for fcurve in get_fcurves():
        for modifier in fcurve.modifiers:
            modifiers.append(modifier)
    return modifiers