import bpy, re
from bpy.types import Operator,Menu
from bpy.props import EnumProperty
from ..utils.utils import refresh, get_prefs, show_error

def remove_numeric_contents(string):
    pattern = r'\[\d+\]$'
    result = re.sub(pattern, '', string)
    return result

def replace_name_in_string(string, replacement):
    pattern = r'pose\.bones\["([^"]+)"\]'
    result = re.sub(pattern, r'pose.bones["' + replacement + r'"]', string)
    return result

def get_data_path(data_path,kind="SINGLE_ADD"):
    data_path = data_path.split(".")
    array_index = -1
    if data_path[-1].endswith("]"):
        if kind == "SINGLE_ADD":
            data_array = data_path[-1].split("[")[1].strip("]")
            if data_array.isnumeric():
                array_index = int(data_array)
        else:
            data_path[-1] = remove_numeric_contents(data_path[-1])
    return ".".join(data_path[3:]),array_index

class OBJECT_OT_Selected_Objects_Keying_Set(Operator):
    """Add Property of All Selected Objects to Keying Set"""
    bl_idname = "object.add_to_keying_set"
    bl_label = "Add Property of Selected Objects to Keying Set"

    kind: EnumProperty(
        name="Kind",
        items=[
            ("ARRAY_ADD","Array Add","Add an array property"),
            ("REMOVE","Array Remove","Remove an array property"),
            ("SINGLE_ADD","Single Add","Add a single property"),
        ]
    )

    @classmethod
    def poll(cls, context):
        return len(context.scene.keying_sets)

    def execute(self, context):

        if hasattr(context, 'button_prop'):
            debug_mode = get_prefs().debug_mode

            if debug_mode:
                print("#"*90)

            obj = context.object
            scene = context.scene

            if context.button_prop.is_animatable == False:
                self.report({"WARNING"}, "Property is not animatable.")
                #show_error(layout,"Property is not animatable.")

            d_path = None
            try:
                bpy.ops.ui.copy_data_path_button(full_path=True)
                d_path = context.window_manager.clipboard

            except Exception as e:
                self.report({"ERROR"}, f'Could not copy the button data path\n{e}')
                return {'FINISHED'}

            if debug_mode:
                print("*"*10,"OBJECT_OT_Selected_Objects_Keying_Set","*"*10)
                print("kind",self.kind)
                print("d_path",d_path)
            
            if obj.type == "ARMATURE" and obj.mode == "POSE":
                for pose_bone in bpy.context.selected_pose_bones:
                    dp,array_index = get_data_path(d_path,self.kind)
                    dp = replace_name_in_string(dp, pose_bone.name)
                    if self.kind in ["ARRAY_ADD","SINGLE_ADD"]:
                        scene.keying_sets.active.paths.add(pose_bone.id_data, dp, index=array_index)   
                    else:
                        for key_path in scene.keying_sets.active.paths:
                            if debug_mode:
                                print("key_path.id",key_path.id,"pose_bone.id_data",pose_bone.id_data,
                                      "key_path.data_path",key_path.data_path,"dp",dp)
                            kp = remove_numeric_contents(key_path.data_path)
                            if key_path.id == pose_bone.id_data and kp == dp:
                                scene.keying_sets.active.paths.remove(key_path)
            else:
                start = d_path.find('objects') + len('objects')
                end = d_path.find(']', start)
                for obj in context.selected_objects:
                    data_path = d_path.replace(d_path[start+2:end-1], "#")
                    data_path = data_path.split('bpy.data.objects[\"#\"]')[-1].strip('.')
                    prop = data_path
                    array_index = -1
                    if data_path.endswith("]"):
                        prop = data_path.split("[")[0]
                        if self.kind == "SINGLE_ADD":
                            data_array = data_path.split("[")[1].strip("]")
                            if data_array.isnumeric():
                                array_index = int(data_array)
                    if self.kind in ["ARRAY_ADD","SINGLE_ADD"]:
                        p = scene.keying_sets.active.paths.add(obj.id_data, prop, index=array_index)
                    else:
                        for key_path in scene.keying_sets.active.paths:
                            if key_path.id == obj.id_data and key_path.data_path == prop:
                                scene.keying_sets.active.paths.remove(key_path)

            refresh()
        return {'FINISHED'}

def keying_set_menu_func(self, context):
    if context.area.type == 'SEQUENCE_EDITOR':
        pass
    else:
        if hasattr(context.object,"mode"):
            current_mode = context.object.mode
        else:
            current_mode = None

        objs = None

        if current_mode == "POSE":
            objs = context.selected_pose_bones
        else:
            objs = context.selected_objects

        if not hasattr(context,"button_prop"):
            return

        viewports = context.button_prop.identifier == "hide_viewport"

        layout = self.layout
        layout.separator()

        if (objs == None or len(objs) == 0) and viewports == False:
            show_error(layout,"No objects are selected in the 3D Viewport.")
            return

        MULTIKEYINGSET_TXT = "Add Property of Selected Objects to Keying Set"
        SINGLEKEYINGSET_TXT = "Add Single Property of Selected Objects to Keying Set"
        MULTIDELETE_KEYING_TXT = "Remove Property of Selected Objects from Keying Set"

        if not context.button_prop.is_hidden and context.object and context.object.type in ["ARMATURE","MESH"]:
            if hasattr(context.button_prop, "is_array"):
                if context.button_prop.is_array:
                    layout.separator()
                    layout.operator(OBJECT_OT_Selected_Objects_Keying_Set.bl_idname, text=MULTIKEYINGSET_TXT, icon="KEYINGSET").kind = "ARRAY_ADD"
                    layout.operator(OBJECT_OT_Selected_Objects_Keying_Set.bl_idname, text=SINGLEKEYINGSET_TXT).kind = "SINGLE_ADD"
                    layout.operator(OBJECT_OT_Selected_Objects_Keying_Set.bl_idname, text=MULTIDELETE_KEYING_TXT).kind = "REMOVE"
                else:
                    layout.operator(OBJECT_OT_Selected_Objects_Keying_Set.bl_idname, text=SINGLEKEYINGSET_TXT).kind = "SINGLE_ADD"
                    layout.operator(OBJECT_OT_Selected_Objects_Keying_Set.bl_idname, text=MULTIDELETE_KEYING_TXT).kind = "REMOVE"
            else:
                layout.operator(OBJECT_OT_Selected_Objects_Keying_Set.bl_idname, text=SINGLEKEYINGSET_TXT).kind = "SINGLE_ADD"
                layout.operator(OBJECT_OT_Selected_Objects_Keying_Set.bl_idname, text=MULTIDELETE_KEYING_TXT).kind = "REMOVE"

class WM_MT_button_context(Menu):
    bl_label = "Insert Keys for Selected Objects"

    def draw(self, context):
        pass

if (3, 3, 0) > bpy.app.version:
    classes = (OBJECT_OT_Selected_Objects_Keying_Set, WM_MT_button_context,)
else:
    classes = (OBJECT_OT_Selected_Objects_Keying_Set,)    

def register():
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)

    if (3, 3, 0) > bpy.app.version:
        rcmenu = getattr(bpy.types, "WM_MT_button_context", None)
        if rcmenu is None:
            bpy.utils.register_class(WM_MT_button_context)
            rcmenu = WM_MT_button_context
        draw_funcs = rcmenu._dyn_ui_initialize()
        draw_funcs.append(keying_set_menu_func)
    else:
        bpy.types.UI_MT_button_context_menu.prepend(keying_set_menu_func)

def unregister():
    if (3, 3, 0) > bpy.app.version:
        bpy.utils.unregister_class(WM_MT_button_context)
    else:
        bpy.types.UI_MT_button_context_menu.remove(keying_set_menu_func)

    from bpy.utils import unregister_class
    for cls in reversed(classes):
        unregister_class(cls)

if __name__ == "__main__":
    register()