import bpy, re
from bpy.types import (Menu,Operator)
from bpy.props import (BoolProperty,StringProperty,EnumProperty)
from ..utils.utils import refresh,use_temp_override,get_prefs,show_error
from .key_selected_clips import OBJECT_OT_Key_Selected_Clips

#https://blender.stackexchange.com/questions/64951/add-menu-entry-to-generic-right-click-menu/76953

def find_last_nonempty_part(s, sep):
    parts = s.split(sep)
    for i in range(len(parts) - 1, -1, -1):
        if parts[i] != '':
          return parts[i]
        else:
          return "%s]"%parts[i-1]
    return None

def split_path(data_path):
    if not data_path:
        return []
    res = find_last_nonempty_part(data_path,"]").strip(".")
    return res

def split_shape_key_path(shape_key_name, data_path):
    if not data_path:
        return [], []

    # Define the pattern to split the data path
    PATTERN = re.compile(r'''((?:[^."']|"[^"]*"|'[^']*')+)''')
    data = PATTERN.split(data_path)[1::2]
    prop = []

    last_element_with_bracket = None
    for i, dc in enumerate(data):
        if "[" in dc and "]" in dc and i != len(data) - 1:
            last_element_with_bracket = i

    if last_element_with_bracket:
        prop = data[last_element_with_bracket + 1:]
        prop = '.'.join(prop)
        data = data[:last_element_with_bracket + 1]

    for i, dc in enumerate(data):
        if "[\"" in dc and "\"]" in dc:
            data[i] = re.sub(r'".*"', "\"%s\"" % shape_key_name, dc)
            break

    res = '.'.join(data)

    return res, prop

def split_material_path(mat_name,data_path):
    if not data_path:
        return []

    PATTERN = re.compile(r'''((?:[^."']|"[^"]*"|'[^']*')+)''')
    data = PATTERN.split(data_path)[1::2]
    prop = []

    if get_prefs().debug_mode:
        print(mat_name,data)

    last_element_with_bracket = None
    for i,dc in enumerate(data):
        if "[" in dc and "]" in dc and i != len(data)-1:
            last_element_with_bracket = i

    if last_element_with_bracket:
        prop = data[last_element_with_bracket+1:]
        prop = '.'.join(prop)
        data = data[:last_element_with_bracket+1]

    for i,dc in enumerate(data):
        if "[\"" in dc and "\"]" in dc:
            data[i] = re.sub(r'".*"', "\"%s\""%mat_name, dc)
            break
    res = '.'.join(data)

    return res,prop

def split_special_path(stype,obj_name,mod_name,data_path,current_mode=None):
    debug_mode = get_prefs().debug_mode
    if debug_mode:
        print("** split_special_path **")
        print("stype",stype)
        print("obj_name",obj_name)
        print("mod_name",mod_name)
        print("data_path",data_path)

    if not data_path:
        return []

    d_mod_name = '%s["%s"]'%(stype,mod_name)
    
    modifier = data_path.find(d_mod_name)
    if modifier == -1:
        return [None,None]

    if debug_mode:
        print(data_path)

    res = data_path[:(modifier+len(d_mod_name))]
    res = res.rstrip('.')

    if current_mode == "POSE":
        d_obj_name = 'bones["%s"]'%obj_name
        res = re.sub("bones\[\".*?\"\]", d_obj_name, res)
    elif "grease_pencils" in data_path:
        d_obj_name = 'grease_pencils["%s"]'%obj_name
        res = re.sub("grease_pencils\[\".*?\"\]", d_obj_name, res)  
    else:
        d_obj_name = 'objects["%s"]'%obj_name
        res = re.sub("objects\[\".*?\"\]", d_obj_name, res)

    prop = data_path[(modifier+len(d_mod_name)):]
    prop = prop.lstrip('.')
    
    return res,prop

def get_objs_from_outliner_selections(context):
    objs = []
    for item in context.scene.outliner_selections:
        ob = None
        if item.identifier == "Collection":
            ob = bpy.data.collections.get(item.name)
        else:
            ob = context.scene.objects[item.name]
        objs.append(ob)
    return objs

class OBJECT_OT_Key_Selected_Objects(Operator):
    """Key All Selected Objects"""
    bl_idname = "object.key_all_objects"
    bl_label = "Keyframe All Selected Objects"

    viewports: BoolProperty(
        name = "Show in Viewports",
        description = "Show in Viewports",
        default = False)

    multiple: BoolProperty(
        name = "Multiple Properties",
        description = "Keyframe multiple properties",
        default = False)

    delete: BoolProperty(
        name = "Delete Keyframes",
        description = "Delete keyframes for multiple properties",
        default = False)

    @classmethod
    def poll(cls, context):
        return context.active_object is not None

    def set_outliner_selections(self):
        win = bpy.context.window
        scr = win.screen
        areas3d = [area for area in scr.areas if area.type == 'OUTLINER']
        if not len(areas3d):
            print("Please open an Outliner space to keyframe this property.")
            self.report({"WARNING"},"Please open an Outliner space to keyframe this property.")
            return False
        regions = [region for region in areas3d[0].regions if region.type == 'WINDOW']

        if use_temp_override():
            with bpy.context.temp_override(window=win, area=areas3d[0], region=regions[0],
                screen=scr):
                bpy.ops.outliner.selected_objects()
        else:
            override = {
                'window': win,
                'screen': scr,
                'area': areas3d[0],
                'region': regions[0],
            }
            bpy.ops.outliner.selected_objects(override)
        return True

    def special_key_handler(self, obj_name, special_elem, special_elem_name, stype, d_path, delete, current_mode):
        res,last_data = split_special_path(stype, obj_name, special_elem_name, d_path, current_mode)
        #bpy.data.objects["Armature"].pose.bones["Bone"].constraints["Child Of"].influence        
        #bpy.data.objects["Cube"].modifiers["GeometryNodes"].keyframe_insert(data_path="[\"Input_2\"]")
        if not res or not last_data:
            return
        data_name = None
        data_index = None
        debug_mode = get_prefs().debug_mode

        if debug_mode:
            print("obj_name:",obj_name)
            print("special_elem_name",special_elem_name)
            print("d_path",d_path)
            print("res:",res)
            print("last_data",last_data)

        try:
            special_elem = eval(res)
        except KeyError as e:
            self.report({"ERROR"}, f'Please make sure your {stype} match.')
            print(e)
            return  # Exit early on error

        if debug_mode:
            print("special_elem 2:",special_elem)

        if last_data.endswith("]") and not last_data.startswith("["):
            last_data = last_data.split('[')
            data_name = last_data[0]
            end = last_data[-1][:-1]
            if debug_mode:
                print("data_name",data_name,"end",end)
            if end.isnumeric():
                data_index = int(end)
        else:
            data_name = last_data

        try:
            if "settings" in last_data:
                if debug_mode:
                    print("settings in last_data")
                data_name = data_name.replace("settings.", "")
                if not delete:
                    special_elem.settings.keyframe_insert(data_path=data_name)
                else:
                    special_elem.settings.keyframe_delete(data_path=data_name)
            else:
                if not delete:
                    if data_index is not None and not self.multiple:
                        special_elem.keyframe_insert(data_path=data_name, index=data_index)
                    else:
                        special_elem.keyframe_insert(data_path=data_name)
                else:
                    if data_index is not None and not self.multiple:
                        special_elem.keyframe_delete(data_path=data_name, index=data_index)
                    else:
                        special_elem.keyframe_delete(data_path=data_name)
        except Exception as e:
            if not delete:
                print('Could not insert a keyframe in %s.\n'%stype,e,'\n')
            else:
                print('Could not delete a keyframe in %s.\n'%stype,e,'\n')

    def shape_key_handler(self, obj, shape_key_name, key_block, d_path, delete):
        res,last_data = split_shape_key_path(shape_key_name,d_path)
        data_name = None
        data_index = None
        debug_mode = get_prefs().debug_mode

        if debug_mode:
            print(res,last_data)

        try:
            shape_key = eval(res)
        except KeyError as e:
            print('Please make sure your shape keys match.\n',e,'\n')

        if last_data.endswith("]"):
            last_data = last_data.split('[')
            data_name = last_data[0]
            data_index = int(last_data[-1][:-1])
        else:
            data_name = last_data

        try:
            if not delete:
                if data_index is not None and not self.multiple:
                    key_block.keyframe_insert(data_path=data_name, index=data_index)
                else:
                    key_block.keyframe_insert(data_path=data_name)
            else:
                if data_index is not None and not self.multiple:
                    key_block.keyframe_delete(data_path=data_name, index=data_index)
                else:
                    key_block.keyframe_delete(data_path=data_name)
        except Exception as e:
            if not delete:
                print('Could not insert a keyframe in material.\n',e,'\n')
            else:
                print('Could not delete a keyframe in material.\n',e,'\n')

    def mat_key_handler(self, mat, d_path, delete):
        res,last_data = split_material_path(mat.name,d_path)
        data_name = None
        data_index = None
        debug_mode = get_prefs().debug_mode

        if debug_mode:
            print(res,last_data)

        try:
            mat = eval(res)
        except KeyError as e:
            print('Please make sure your surface shaders match.\n',e,'\n')

        if last_data.endswith("]"):
            last_data = last_data.split('[')
            data_name = last_data[0]
            data_index = int(last_data[-1][:-1])
        else:
            data_name = last_data

        try:
            if not delete:
                if data_index is not None and not self.multiple:
                    mat.keyframe_insert(data_path=data_name, index=data_index)
                else:
                    mat.keyframe_insert(data_path=data_name)
            else:
                if data_index is not None and not self.multiple:
                    mat.keyframe_delete(data_path=data_name, index=data_index)
                else:
                    mat.keyframe_delete(data_path=data_name)
        except Exception as e:
            if not delete:
                print('Could not insert a keyframe in material.\n',e,'\n')
            else:
                print('Could not delete a keyframe in material.\n',e,'\n')

    def key_handler(self, obj, data_name, delete, grp=None, data_index=None):
        debug_mode = get_prefs().debug_mode
        data_types = {
            "OBJECTS", "LIGHTS", "CAMERAS", "MESHES", "ARMATURES", "CURVES", "LATTICES", "METABALLS", "MATERIALS", 
            "TEXTURES", "IMAGES", "NODE_GROUPS", "WORLDS", "COLLECTIONS", "ACTIONS", "BRUSHES", "CACHE_FILES", 
            "FONTS", "GREASE_PENCILS", "HAIR_CURVES", "LIBRARIES", "LIGHTPROBES", "LINESTYLES", "MASKS", "MOVIECLIPS", 
            "PAINT_CURVES", "PALETTES", "PARTICLES", "POINTCLOUDS", "SCENES", "SCREENS", "SHAPE_KEYS", "SOUNDS", 
            "SPEAKERS", "TEXTS", "VOLUMES", "WINDOW_MANAGERS", "WORKSPACES"
        }

        try:

            if not delete:
                print("*"*30,"ADD","*"*30)
                print("obj: %s\ndata_name: %s\ndelete: %s\ngrp: "\
                    "%s\ndata_index: %s"%(obj, data_name, delete, grp, data_index))

                if hasattr(obj, "type") and obj.type in data_types:
                    obj.data.keyframe_insert(data_path=data_name, index=data_index)

                if grp and data_index is not None:
                    if debug_mode: print(f"key_handler: add 1")
                    obj.keyframe_insert(data_path=data_name, index=data_index, group=grp)
                elif grp:
                    if debug_mode: print(f"key_handler: add 2")
                    obj.keyframe_insert(data_path=data_name, group=grp)
                elif data_index:
                    if debug_mode: print(f"key_handler: add 3")
                    obj.keyframe_insert(data_path=data_name, index=data_index)
                else:
                    if debug_mode: print(f"key_handler: add 4")
                    obj.keyframe_insert(data_path=data_name)
            else:
                print("*"*30,"DELETE","*"*30)
                print("obj: %s\ndata_name: %s\ndelete: %s\ngrp: "\
                    "%s\ndata_index: %s"%(obj, data_name, delete, grp, data_index))

                if hasattr(obj, "type") and obj.type in data_types:
                    obj.data.keyframe_delete(data_path=data_name, index=data_index)

                if grp and data_index is not None:
                    if debug_mode: print(f"key_handler: delete 2")
                    obj.keyframe_delete(data_path=data_name, index=data_index, group=grp)
                elif grp:
                    if debug_mode: print(f"key_handler: delete 3")
                    obj.keyframe_delete(data_path=data_name, group=grp)
                elif data_index:
                    if debug_mode: print(f"key_handler: delete 4")
                    obj.keyframe_delete(data_path=data_name, index=data_index)
                else:
                    if debug_mode: print(f"key_handler: delete 5")
                    obj.keyframe_delete(data_path=data_name)
        except Exception as e:
            if not delete:
                print('Could not insert a keyframe in object.\n',e,'\n')
            else:
                print('Could not delete a keyframe in object.\n',e,'\n')

            try:
                if not delete:
                    print("*"*30,"ADD","*"*30)
                    print("obj: %s\ndata_name: %s\ndelete: %s\ngrp: "\
                        "%s\ndata_index: %s"%(obj, data_name, delete, grp, data_index))
                    if grp and data_index is not None:
                        if debug_mode: print(f"key_handler data: add 1")
                        obj.data.keyframe_insert(data_path=data_name, index=data_index, group=grp)
                    elif grp:
                        if debug_mode: print(f"key_handler data: add 2")
                        obj.data.keyframe_insert(data_path=data_name, group=grp)
                    elif data_index:
                        if debug_mode: print(f"key_handler data: add 3")
                        obj.data.keyframe_insert(data_path=data_name, index=data_index)
                    else:
                        if debug_mode:
                            print(f"key_handler data: add 4")
                            print(obj,data_name)
                        obj.data.keyframe_insert(data_path=data_name)
                else:
                    if debug_mode:
                        print("*"*30,"DELETE","*"*30)
                        print("obj: %s\ndata_name: %s\ndelete: %s\ngrp: "\
                            "%s\ndata_index: %s"%(obj, data_name, delete, grp, data_index))
                    if grp and data_index is not None:
                        if debug_mode: print(f"key_handler data: delete 1")
                        obj.data.keyframe_delete(data_path=data_name, index=data_index, group=grp)
                    elif grp:
                        if debug_mode: print(f"key_handler data: delete 2")
                        obj.data.keyframe_delete(data_path=data_name, group=grp)
                    elif data_index:
                        if debug_mode: print(f"key_handler data: delete 3")
                        obj.data.keyframe_delete(data_path=data_name, index=data_index)
                    else:
                        if debug_mode: print(f"key_handler data: delete 4")
                        obj.data.keyframe_delete(data_path=data_name)
            except Exception as e:
                if not delete:
                    print('Could not insert a keyframe in object or object data.\n',e,'\n')
                else:
                    print('Could not delete a keyframe in object or object data.\n',e,'\n')

    def get_name(self,obj_type,data_path):
        #bpy.data.objects["Body - Bottom"].modifiers["Screw"].angle
        #bpy.data.objects["Armature"].pose.bones["Bone.001"].constraints["Child Of"].influence
        result = data_path.partition(obj_type)[2]
        #["Screw"].angle
        startIndex = result.find('\"')
        if startIndex != -1:
            endIndex = result.find('\"', startIndex + 1)
            if startIndex != -1 and endIndex != -1: #i.e. both quotes were found
                result = result[startIndex+1:endIndex]
        return result        

    def execute(self, context):
        debug_mode = get_prefs().debug_mode

        if hasattr(context, 'button_prop'):
            if context.button_prop.is_animatable == False:
                self.report({"WARNING"}, "Property is not animatable.")
                #show_error(layout,"Property is not animatable.")

            d_path = None
            try:
                bpy.ops.ui.copy_data_path_button(full_path=True)
                d_path = context.window_manager.clipboard

            except Exception as e:
                self.report({"ERROR"}, f'Could not copy the button data path\n{e}')
                return {'FINISHED'}

            objs = None
            mats = []
            types = ["layers","particle_systems","modifiers","constraints"]
            if hasattr(context.object,"mode"):
                current_mode = context.object.mode
            else:
                current_mode = context.mode                

            if d_path.startswith("bpy.data.materials"):
                objects = context.selected_objects
                mats.extend(obj.active_material for obj in objects)
                for mat in mats:
                    self.mat_key_handler(mat, d_path, self.delete)

                refresh()
                return {'FINISHED'}
            elif d_path.startswith("bpy.data.shape_keys"):
                objects = context.selected_objects
                if debug_mode:
                    print(f"shape_keys: {d_path}")
                def extract_key_block_name(d_path):
                    # Define the pattern to match the key_block name
                    pattern = r'key_blocks\["(.*?)"\]'
                    # Use regular expressions to find the key_block name
                    match = re.search(pattern, d_path)
                    if match:
                        # Return the matched key_block name
                        return match.group(1)
                    else:
                        # Return None if no match is found
                        return None
                for obj in objects:
                    bpy.data.objects["Cube.002"].active_shape_key_index
                    if obj.data.shape_keys:
                        shape_key_name = obj.data.shape_keys.name
                        shape_keys = obj.data.shape_keys
                        key_block_name = extract_key_block_name(d_path)
                        key_block = shape_keys.key_blocks.get(key_block_name)
                        self.shape_key_handler(obj, shape_key_name, key_block, d_path, self.delete)
                refresh()
                return {'FINISHED'}
            else:
                for o_type in types:
                    if debug_mode:
                        print("Type: ",o_type,"Data Path: ",d_path)
                    if o_type in d_path:
                        o_name = self.get_name(o_type,d_path)
                        if debug_mode:
                            print("*"*25,o_name,"*"*25)
                        if current_mode == "POSE":
                            objects = context.selected_pose_bones
                        else:
                            objects = context.selected_objects
                        for obj in objects:
                            if current_mode == "POSE":
                                my_type_name = "bpy.data.objects[\"%s\"].pose.bones[\"%s\"].%s"%(obj.id_data.name,obj.name,o_type)
                                my_types = eval(my_type_name)
                            else:
                                my_type_name = "obj.%s"%o_type     
                                try:
                                    my_types = eval(my_type_name)
                                except AttributeError:
                                    my_types = eval("obj.data.%s"%o_type)
                            if debug_mode:
                                print("my_type_name",my_type_name)
                                print("my_types",my_types)
                            for o in my_types:
                                ob_name = ""
                                if type(o).__name__ == "GPencilLayer":
                                    ob_name = o.info
                                    print("%s (o_name) in %s (o.name)"%(o_name,ob_name))
                                else:
                                    ob_name = o.name
                                    print("%s (o_name) in %s (o.name)"%(o_name,ob_name))
                                if o_name in ob_name:
                                    self.special_key_handler(obj.name, o, ob_name, o_type, d_path, self.delete, current_mode)
                        refresh()
                        return {'FINISHED'}

            object_transforms = ["location","rotation_euler","rotation_quaternion","scale",
                "delta_location","delta_rotation_euler","delta_rotation_quaternion",
                "delta_scale"]

            if current_mode == "POSE":
                objs = context.selected_pose_bones
            else:
                objs = None
                if self.viewports == True:
                    if self.set_outliner_selections():
                        objs = get_objs_from_outliner_selections(context)
                    else:
                        return {'CANCELLED'}
                else:
                    objs = context.selected_objects

            last_data = split_path(d_path)
            data_index = None
            data_name = None

            if debug_mode:
                print("d_path: %s"%d_path)
                print("last_data 1: %s"%last_data)

            if not last_data:
                return {'FINISHED'}

            if last_data.endswith("]"):
                data_str = last_data.split('[')
                data_name = data_str[0]
                end = data_str[-1][:-1]
                if debug_mode:
                    print("data_name",data_name,"end",end)
                if end.isnumeric():
                    data_index = int(end)
                else:
                    data_name = last_data
            else:
                data_name = last_data

            if debug_mode:
                print("data_name: %s"%data_name)
                print("data_index: %s"%data_index)

            grp = None

            if debug_mode: print("self.multiple: %s"%self.multiple)

            for i,obj in enumerate(objs):
                if current_mode == "POSE":
                    grp=obj.bone.name
                else:
                    if data_name in object_transforms:
                        grp = "Object Transforms"

                if not self.multiple:
                    self.key_handler(obj, data_name, self.delete, grp, data_index)
                else:
                    self.key_handler(obj, data_name, self.delete, grp)
            # Update the dependency graph after performing keyframe operations
            bpy.context.view_layer.update()
            refresh()
        return {'FINISHED'}


class WM_MT_button_context(Menu):
    bl_label = "Insert Keys for Selected Objects"

    def draw(self, context):
        pass

def key_selected_menu_func(self, context):
    if context.area.type == 'SEQUENCE_EDITOR':
        # VSE specific code
        scene = context.scene
        sequencer = scene.sequence_editor
        sequences = context.selected_sequences

        if not sequences:
            show_error(layout,"No clips are selected in the 3D Viewport.")
            return

        layout = self.layout
        layout.separator()

        # Change the text dynamically based on the VSE context
        MULTIKEY_TXT = "Keyframe All Selected Clips"
        SINGLEKEY_TXT = "Single Keyframe for All Selected Clips"
        MULTIDELETE_TXT = "Delete Keyframes for All Selected Clips"
        SINGLEDELETE_TXT = "Delete Single Keyframe for All Selected Clips"

        if len(sequences):
            if not context.button_prop.is_hidden and context.button_prop.is_animatable:
                single = layout.operator(OBJECT_OT_Key_Selected_Clips.bl_idname, text=SINGLEKEY_TXT, icon='KEY_HLT')
                single.delete = False
                del_single = layout.operator(OBJECT_OT_Key_Selected_Clips.bl_idname, text=SINGLEDELETE_TXT, icon='KEY_DEHLT')
                del_single.delete = True
    else:
        if hasattr(context.object,"mode"):
            current_mode = context.object.mode
        else:
            current_mode = None

        objs = None

        if current_mode == "POSE":
            objs = context.selected_pose_bones
        else:
            objs = context.selected_objects

        if not hasattr(context,"button_prop"):
            return

        viewports = context.button_prop.identifier == "hide_viewport"

        layout = self.layout
        layout.separator()

        if (objs == None or len(objs) == 0) and viewports == False:
            print(f"context.area.type: {context.area.type}")
            show_error(layout,"No objects are selected in the 3D Viewport.")
            return

        MULTIKEY_TXT = "Keyframe All Selected Objects"
        SINGLEKEY_TXT = "Single Keyframe for All Selected Objects"
        MULTIDELETE_TXT = "Delete Keyframes for All Selected Objects"
        SINGLEDELETE_TXT = "Delete Single Keyframe for All Selected Objects"
        
        if len(objs) > 1 or viewports:
            """
            if context.button_prop.is_animatable == False:
                show_error(layout,"Property is not animatable.")
                return
            if context.button_prop.is_animatable and not context.button_prop.is_hidden:
            """
            if not context.button_prop.is_hidden:
                if hasattr(context.button_prop, "is_array"):
                    if context.button_prop.is_array:
                        multi = layout.operator(OBJECT_OT_Key_Selected_Objects.bl_idname, text=MULTIKEY_TXT, icon='KEY_HLT')
                        multi.multiple = True
                        multi.delete = False
                        single = layout.operator(OBJECT_OT_Key_Selected_Objects.bl_idname, text=SINGLEKEY_TXT)
                        single.multiple = False
                        single.delete = False
                        del_multi = layout.operator(OBJECT_OT_Key_Selected_Objects.bl_idname, text=MULTIDELETE_TXT, icon='KEY_DEHLT')
                        del_multi.multiple = True
                        del_multi.delete = True
                        del_single = layout.operator(OBJECT_OT_Key_Selected_Objects.bl_idname, text=SINGLEDELETE_TXT)
                        del_single.multiple = False
                        del_single.delete = True
                    else:
                        single = layout.operator(OBJECT_OT_Key_Selected_Objects.bl_idname, text=MULTIKEY_TXT, icon='KEY_HLT')
                        single.multiple = False
                        single.delete = False
                        single.viewports = viewports
                        del_single = layout.operator(OBJECT_OT_Key_Selected_Objects.bl_idname, text=MULTIDELETE_TXT, icon='KEY_DEHLT')
                        del_single.multiple = False
                        del_single.delete = True
                        del_single.viewports = viewports
                else:
                    single = layout.operator(OBJECT_OT_Key_Selected_Objects.bl_idname, text=MULTIKEY_TXT, icon='KEY_HLT')
                    single.multiple = False
                    single.delete = False
                    del_single = layout.operator(OBJECT_OT_Key_Selected_Objects.bl_idname, text=MULTIDELETE_TXT, icon='KEY_DEHLT')
                    del_single.multiple = False
                    del_single.delete = True

if (3, 3, 0) > bpy.app.version:
    classes = (OBJECT_OT_Key_Selected_Objects, WM_MT_button_context,)
else:
    classes = (OBJECT_OT_Key_Selected_Objects,)

def register():
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)

    if (3, 3, 0) > bpy.app.version:
        rcmenu = getattr(bpy.types, "WM_MT_button_context", None)
        if rcmenu is None:
            bpy.utils.register_class(WM_MT_button_context)
            rcmenu = WM_MT_button_context
        draw_funcs = rcmenu._dyn_ui_initialize()
        draw_funcs.append(key_selected_menu_func)
    else:
        bpy.types.UI_MT_button_context_menu.prepend(key_selected_menu_func)

def unregister():
    if (3, 3, 0) > bpy.app.version:
        bpy.utils.unregister_class(WM_MT_button_context)
    else:
        bpy.types.UI_MT_button_context_menu.remove(key_selected_menu_func)

    from bpy.utils import unregister_class
    for cls in reversed(classes):
        unregister_class(cls)